Require Import Coqlib.
Require Import Maps.
Require Import VTDefs.
Require Import Integers.
Require Import Values.
Require Import LemmaVTMap.
Require Import LemmaMaths.
Require Import Coq.Sorting.Mergesort.
Require Import Coq.Structures.Orders.
Require Import Coq.Sorting.Permutation.
Require Import EDFProofEnlarge.
Require Import EDFProofShrink.
Require Import EDFProofSamePeriod.
Require Import Psatz.


Local Open Scope Z_scope.

Section SCHEDULABILITYTEST.

  (* schedulability test *)
  
  Fixpoint enlarge_to_hyper (H: Z) (N: nat) (conf: PrioConfigPool) :=
    match N with
    | O => conf
    | S n => match (ZMap.get (Z.of_nat n) conf) with
            | mkPrioConfigValid T C =>
              ZMap.set (Z.of_nat n) (mkPrioConfigValid H (H / T * C))
                       (enlarge_to_hyper H n conf)
            end
    end.

  Lemma enlarge_to_hyper_period:
    forall i H N conf T C T0 C0,
      0 <= i < Z.of_nat N ->
      ZMap.get i conf = mkPrioConfigValid T0 C0 ->
      ZMap.get i (enlarge_to_hyper H N conf) = mkPrioConfigValid T C ->
      T = H /\ C = H / T0 * C0.
  Proof.
    intros i H N.
    revert i.
    induction N.
    {
      simpl.
      intros.
      omega.
    }
    {
      intros * Hi Hconfi.
      simpl.
      destruct (zeq i (Z.of_nat N)); subst.
      {
        rewrite Hconfi.
        rewrite ZMap.gss.
        intros H1.
        inv H1.
        split; auto. 
      }
      {        
        destruct (ZMap.get (Z.of_nat N) conf).
        rewrite ZMap.gso; auto.
        apply IHN; auto.
        rewrite Nat2Z.inj_succ in Hi.
        omega.        
      }
    }
  Qed.
        
  Definition schedulability_criterion (H: Z) (N: Z) (conf: PrioConfigPool) : Prop :=
    is_a_hyperperiod H N conf /\ H * H < Int.max_unsigned /\
         sum_budget (Z.to_nat N) (enlarge_to_hyper H (Z.to_nat N) conf) <= H.
      
End SCHEDULABILITYTEST.

Section EnlargeParam.

  Inductive enlargeParam: Type :=
  | EParam: Z (* task ID *) -> Z (* factor *) -> enlargeParam.

  Function enlargeParam_lt e1 e2 :=
    match e1, e2 with
    | EParam id1 T1, EParam id2 T2 =>
      if (zlt T1 T2) then true else false
    end.

  Function enlargeParam_le e1 e2 :=
    match e1, e2 with
    | EParam id1 T1, EParam id2 T2 =>
      if (zle T1 T2) then true else false
    end.

  Theorem enlargeParam_dec:
    forall x y : enlargeParam, {x = y} + {x <> y}.
  Proof.
    intros.
    destruct x as [id1 k1], y as [id2 k2].
    destruct (zeq id1 id2); subst.
    - destruct (zeq k1 k2); subst; auto.
      right.
      congruence.
    - right.
      congruence.
  Qed.

End EnlargeParam.

Module EParamOrder <: TotalLeBool.

  Definition t := enlargeParam.

  Function leb (x y: enlargeParam) := enlargeParam_le x y.

  Infix "<=p" := leb (at level 35).

  Theorem leb_total : forall a1 a2, a1 <=p a2 = true  \/ a2 <=p a1 = true.
  Proof.
    intros.
    destruct a1 as [id1 T1], a2 as [id2 T2].
    unfold leb, enlargeParam_le.
    destruct (zle T1 T2) as [Hdlt | Hdge]; auto.
    right.
    rewrite zle_true; auto.
    omega.
  Qed.      

End EParamOrder.

Module Import eparamsort := Sort EParamOrder.      
  
Section TRANSITIONING.
    
  Function do_enlarge (id k: Z) conf: PrioConfigPool :=
    match (ZMap.get id conf) with
    | mkPrioConfigValid T C =>
      ZMap.set id (mkPrioConfigValid (k * T) (k * C)) conf
    end.

  Lemma enlarge_preserves_schedulability1:
    forall N conf id k T0 C0,
      1 < k ->
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      config_pool_valid N conf ->
      config_pool_valid N (do_enlarge id k conf) ->
      (forall i T C,
          0 <= i < N ->
          ZMap.get i conf = mkPrioConfigValid T C ->
          T <= k * T0) ->
      system_schedulable N conf ->
      system_schedulable N (do_enlarge id k conf).
  Proof.
    intros * Hk Hid Hconfid Hconfvalid Hconfvalid' Hlarge Hschedulable.
    apply enlarge_preserves_schedulability with (conf:=conf) (id:=id) (k:=k) (T:=T0) (C:=C0); auto.
    - unfold LemmaOtherWorkload.id_is_enlarged_by_k.
      intros * Hconfid'.
      rewrite Hconfid in Hconfid'.
      inv Hconfid'.
      unfold do_enlarge.
      rewrite Hconfid.
      auto.
    - unfold LemmaWorkloadEnlargeSelf.enlarged_task_has_largest_period.
      intros * Hi Hconfid' Hconfi.
      rewrite Hconfid in Hconfid'.
      inv Hconfid'.
      specialize (Hlarge _ _ _ Hi Hconfi).
      auto.
  Qed.  

  Lemma shrink_preserves_schedulability1:
    forall N k conf1 conf2,
      1 < k ->
      conf_k_times N k conf1 conf2 ->
      config_pool_valid N conf1 ->
      config_pool_valid N conf2 ->
      system_schedulable N conf2 ->
      system_schedulable N conf1.
  Proof.
    intros.
    eapply shrink_preserves_schedulability; eauto.
  Qed.

  Fixpoint get_enlarge_params_aux (Ni: nat) (confi: PrioConfigPool) : list enlargeParam :=
    match Ni with
    | O => nil
    | S n => match (ZMap.get (Z.of_nat n) confi) with
            | mkPrioConfigValid T C =>
              EParam (Z.of_nat n) T :: get_enlarge_params_aux n confi
            end
    end.

  Fixpoint extract_id (l: list enlargeParam) :=
    match l with
    | nil => nil
    | EParam id _ :: tl => id :: extract_id tl
    end.

  Lemma extract_id_head_in:
    forall id k l,
      hd_error l = Some (EParam id k) ->
      In id (extract_id l).
  Proof.
    intros * Hhd.
    destruct l.
    - simpl in Hhd.
      discriminate.
    - simpl in Hhd.
      inv Hhd.
      simpl.
      left.
      auto.
  Qed.

  Lemma extract_id_any_in:
    forall id k l,
      In (EParam id k) l ->
      In id (extract_id l).
  Proof.
    intros *.
    induction l.
    {
      intros.
      simpl in H.
      omega.
    }
    {
      intros Hin.
      simpl in Hin.
      destruct Hin as [H1 | H2].
      - inv H1.
        simpl.
        auto.
      - destruct a.
        simpl.
        right.
        auto.
    }
  Qed.

  Lemma get_enlarge_id_bound:
    forall id k N conf,
      In (EParam id k) (get_enlarge_params_aux (Z.to_nat N) conf) ->
      0 <= id < N.
  Proof.
    intros * Hin.
    assert (Npos: 0 < N).
    {
      destruct N; try (simpl in Hin; omega).
      apply Pos2Z.is_pos.
    }
    remember (Z.to_nat N) as Nnat.
    assert (HNN: N = Z.of_nat Nnat).
    {
      rewrite HeqNnat.
      rewrite Z2Nat.id; omega.
    }
    rewrite HNN in *. clear HNN HeqNnat.
    induction Nnat.
    {
      simpl in Npos.
      omega.
    }
    {
      simpl in Hin.
      destruct (ZMap.get (Z.of_nat Nnat) conf).
      simpl in Hin.
      rewrite Nat2Z.inj_succ.
      destruct Hin as [H1 | H2].
      - inv H1.
        omega.
      - destruct (zeq (Z.of_nat Nnat) 0) as [Heq | Hneq].
        + assert (Z.to_nat (Z.of_nat Nnat) = Z.to_nat 0).
          {
            rewrite Heq.
            auto.
          }
          rewrite Nat2Z.id in H.
          rewrite H in H2.
          simpl in H2.
          omega.
        + assert (Hpos: 0 < Z.of_nat Nnat).
          {
            pose proof (Zle_0_nat Nnat).
            omega.
          }
          specialize (IHNnat H2 Hpos).
          omega.
    }
  Qed.  

  Lemma get_enlarge_extract_id_bound:
    forall id N conf,
      In id (extract_id (get_enlarge_params_aux N conf)) ->
      0 <= id < Z.of_nat N.
  Proof.
    intros *.
    induction N.
    {
      simpl.      
      omega.
    }
    {
      intros Hin.
      simpl in Hin.
      destruct (ZMap.get (Z.of_nat N) conf).
      simpl in Hin.
      rewrite Nat2Z.inj_succ.
      destruct Hin as [H1 | H2]; try omega.
      specialize (IHN H2).
      omega.
    }
  Qed.  
  
  Lemma get_enlarge_no_dup:
    forall N conf,
      NoDup (extract_id (get_enlarge_params_aux N conf)).
  Proof.
    intros.
    induction N.
    {
      simpl.
      constructor.
    }
    {
      simpl.
      destruct (ZMap.get (Z.of_nat N) conf).
      simpl.
      constructor; auto.
      destruct (in_dec Z.eq_dec (Z.of_nat N) (extract_id (get_enlarge_params_aux N conf))) as [H1 | H2]; auto.
      pose proof (get_enlarge_extract_id_bound _ _ _ H1).
      omega.
    }
  Qed.

  Lemma permutation_extract_id:
    forall l l',
      Permutation l l' ->
      Permutation (extract_id l) (extract_id l').
  Proof.
    intros * H.
    induction H.
    - simpl.
      constructor.
    - destruct x.
      simpl.
      constructor; auto.
    - destruct x, y.
      simpl.
      constructor; auto.
    - eapply Permutation_trans; eauto.
  Qed.      
      
  Lemma permutation_preserves_extract_id_nodup (A: Type):
    forall (l l': list A),
      Permutation l l' ->
      NoDup l ->
      NoDup l'.
  Proof.
    intros * Hperm Hdup.
    induction Hperm.
    - constructor.
    - inv Hdup.
      specialize (IHHperm H2).
      constructor; auto.
      unfold not.
      intros.
      assert (In x l).
      {
        apply Permutation_in with (l := l'); auto.
        apply Permutation_sym; auto.
      }
      congruence.
    - inv Hdup.
      inv H2.
      assert (Hneq: x <> y).
      {
        unfold not.
        intros.
        apply H1.
        constructor.
        auto.
      }
      constructor.
      + unfold not.
        intros.        
        inv H; try congruence.
      + constructor; auto.
        unfold not.
        intros.
        apply H1.
        simpl.
        auto.
    - auto.
  Qed.
      
  Lemma get_enlarge_k_value:
    forall id k N conf T C,
      In (EParam id k) (get_enlarge_params_aux (Z.to_nat N) conf) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      k = T.
  Proof.
    intros * Hin.
    assert (Npos: 0 < N).
    {
      destruct N; try (simpl in Hin; omega).
      apply Pos2Z.is_pos.
    }    
    remember (Z.to_nat N) as Nnat.
    assert (HNN: N = Z.of_nat Nnat).
    {
      rewrite HeqNnat.
      rewrite Z2Nat.id; omega.
    }
    rewrite HNN in *. clear HNN HeqNnat.
    induction Nnat.
    {
      simpl in Npos.
      omega.
    }
    {
      intros Hconf.
      assert (Hid: 0 <= id < Z.of_nat (S Nnat)).
      {
        eapply get_enlarge_id_bound; eauto.
        rewrite Nat2Z.id.
        eapply Hin.
      }
      pose proof (get_enlarge_no_dup (S Nnat) conf) as Hdup.
      simpl in Hin, Hdup.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T0 C0] eqn:Hconfid.
      simpl in Hdup.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq].
      - rewrite Heq in *.
        destruct Hin as [H1 | H2].
        + inv H1.
          rewrite Hconfid in Hconf.
          inv Hconf.
          auto.
        + inv Hdup.
          pose proof (extract_id_any_in _ _ _ H2).
          contradiction.
      - destruct Hin; try (inv H; omega).
        apply IHNnat; auto.
        destruct Nnat.
        + simpl in Hid. omega.
        + rewrite Nat2Z.inj_succ.
          omega.
    }
  Qed.

  Lemma get_enlarge_exhaustive:
    forall id T C conf N,
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      In (EParam id T) (get_enlarge_params_aux (Z.to_nat N) conf).
  Proof.
    intros * Hid Hconf.
    assert (Npos: 0 < N). omega.
    remember (Z.to_nat N) as Nnat.
    assert (HNN: N = Z.of_nat Nnat).
    {
      rewrite HeqNnat.
      rewrite Z2Nat.id; omega.
    }
    rewrite HNN in *. clear HNN HeqNnat.
    induction Nnat.
    {
      simpl in Npos.
      omega.
    }
    {
      simpl.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq].
      - rewrite Heq in *. clear Heq.
        rewrite Hconf.
        simpl.
        auto.
      - destruct (ZMap.get (Z.of_nat Nnat) conf).
        simpl.
        right.
        rewrite Nat2Z.inj_succ in *.
        apply IHNnat; try omega.
    }
  Qed.

  Variable H N: Z.
  Variable conf: PrioConfigPool.
  Hypothesis Npos: 0 < N.
  Hypothesis Hpos: 1 < H.
  Hypothesis Hconfvalid:
    forall i T C, 0 <= i < N ->
             ZMap.get i conf = mkPrioConfigValid T C ->
             1 < T /\ 0 < C < T.
  Hypothesis sched: schedulability_criterion H N conf.  
        
  (* The hyperized configuration is schedulable *)

  Definition conf0 := enlarge_to_hyper H (Z.to_nat N) conf.

  Lemma conf0_schedulable:
    system_schedulable N conf0.
  Proof.
    intros.
    unfold schedulability_criterion in sched.
    destruct sched as [? [? ? ]].

    assert (IH: forall i T C : Z, 0 <= i < N -> ZMap.get i conf0 = mkPrioConfigValid T C -> T = H).
    {
      unfold conf0.
      intros.
      destruct (ZMap.get i conf) eqn:Hconff.
      assert (H10: 0 <= i < Z.of_nat (Z.to_nat N)).
      {
        rewrite Z2Nat.id; omega.
      }
      
      pose proof (enlarge_to_hyper_period _ _ _ _ _ _ _ _ H10 Hconff H4).
      omega.
    }
    
    assert (Hconf0: config_pool_valid N conf0).
    {
      unfold config_pool_valid.
      unfold conf0.
      intros * Hi Hconfi.
      unfold conf0 in Hconfi.

      destruct (ZMap.get i conf) as [T0 C0] eqn:Hconff.
      assert (H10: 0 <= i < Z.of_nat (Z.to_nat N)).
      {
        rewrite Z2Nat.id; omega.
      }
      
      pose proof (enlarge_to_hyper_period _ H _ _ _ _ _ _ H10 Hconff Hconfi).
      inv H3.
      split; try omega.
      split; try nia.
      
      specialize (Hconfvalid _ _ _ Hi Hconff).
      apply budget_valid_after_hyper; try omega.
      destruct sched.
      destruct H3.
      specialize (H5 _ _ _ Hi Hconff).
      auto.
    }
    
    eapply same_period_schedulable; eauto.
  Qed.

  (* enlarging preserves schedulability *)  
  Definition enlarge_sequence: list enlargeParam :=
    eparamsort.sort (get_enlarge_params_aux (Z.to_nat N) conf).
  
  Fixpoint do_enlarge_sequence (confi: PrioConfigPool) (l: list enlargeParam) :=
    match l with
    | nil => confi
    | EParam id k :: tl =>
      do_enlarge_sequence (do_enlarge id k confi) tl
    end.

  Definition list_OK (Ni: Z) (Hi: Z) (confi: PrioConfigPool) (l: list enlargeParam): Prop :=
    Sorted.LocallySorted
         (fun x x0 : enlargeParam =>
            is_true ((fun x1 y : enlargeParam => enlargeParam_le x1 y) x x0)) l /\
    (forall id k, In (EParam id k) l -> 0 <= id < Ni /\ k =
              (match (ZMap.get id conf) with mkPrioConfigValid T C => T end)) /\
    (NoDup (extract_id l)) /\
    (forall id0 k0, hd_error l = Some (EParam id0 k0) ->
      forall id T C,
        0 <= id < Ni ->
        ZMap.get id confi = mkPrioConfigValid T C ->
        if (in_dec Z.eq_dec id (extract_id l)) then T = Hi else
          T <= k0 * Hi) /\
    config_pool_valid Ni confi.
  
  Lemma list_OK_head:
    forall Ni Hi confi l id k T C,
      list_OK Ni Hi confi l ->
      hd_error l = Some (EParam id k) ->
      ZMap.get id confi = mkPrioConfigValid T C ->
      T = Hi.
  Proof.
    intros * Hlist Hhd Hconf.
    unfold list_OK in Hlist.
    destruct Hlist as [_ [ Hbound [Hdup [Hlist Hconfivalid]]]].
    assert (Hin: In (EParam id k) l).
    {
      destruct l; simpl in Hhd; try discriminate.
      inv Hhd.
      simpl.
      auto.
    }
    specialize (Hbound _ _ Hin).
    destruct Hbound as [Hb1 _].
    assert (Hinid: In id (extract_id l)).      
    eapply extract_id_head_in; eauto.
    
    specialize (Hlist _ _ Hhd _ _ _ Hb1 Hconf).
    destruct (in_dec Z.eq_dec id (extract_id l)); congruence.
  Qed.  
  
  Lemma do_enlarge_step_list_OK:
    forall confi l id0 k0,
      list_OK N H confi l ->
      hd_error l = Some (EParam id0 k0) ->
      list_OK N H (do_enlarge id0 k0 confi) (tl l).
  Proof.
    intros * Hl Hhd.
    destruct (ZMap.get id0 confi) as [T0 C0] eqn:Hconf0.
    assert (HT0: T0 = H).
    {
      eapply list_OK_head; eauto.
    }
    subst.    
    
    destruct Hl as [Hsort [Hbound [Hdup [Hlist Hconfivalid]]]].
    unfold list_OK.
    split.
    {
      inv Hsort; simpl; auto.
      apply Sorted.LSorted_nil.
      apply Sorted.LSorted_nil.
    }
    {
      split.
      {
        intros.
        destruct l; simpl in *; try discriminate.
        apply Hbound.
        auto.
      }
      {
        split.
        {
          (* NoDup *)
          destruct l; try discriminate.
          simpl in *.
          inv Hhd.
          inv Hdup.
          auto.
        }
        {
          split.
          {
            specialize (Hlist _ _ Hhd).
            intros.
            destruct (zeq id id0); subst.
            {
              unfold do_enlarge in H2.
              specialize (Hlist _ _ _ H1 Hconf0).
              destruct (ZMap.get id0 confi).
              inv Hconf0.
              rewrite ZMap.gss in H2.
              inv H2.
              destruct l; simpl in Hhd; try discriminate.
              inv Hhd.
              simpl.
              destruct (in_dec Z.eq_dec id0 (extract_id l)).
              - (* duplication *)
                simpl in Hdup.
                inv Hdup.
                congruence.
              - simpl in H0.
                destruct l; simpl in H0; try discriminate.
                inv Hsort.
                inv H0.              
                simpl in H6.
                destruct (zle k0 k1) as [Hle| Hle]; try discriminate.
                apply Zmult_lt_0_le_compat_r; omega.
            }
            {
              unfold do_enlarge in H2.
              destruct (ZMap.get id0 confi).
              rewrite ZMap.gso in H2; auto.
              specialize (Hlist _ _ _ H1 H2).
              destruct l; simpl in Hhd; try discriminate.
              inv Hhd.
              simpl.
              destruct (in_dec Z.eq_dec id (extract_id (EParam id0 k0 :: l))) as [Hidin | Hidnot].
              - simpl in Hidin.
                destruct Hidin; try omega.
                destruct (in_dec Z.eq_dec id (extract_id l)); try congruence.              
              - simpl in Hidnot.
                destruct (in_dec Z.eq_dec id (extract_id l)) as [H10 | H11].
                + assert (id0 = id \/ In id (extract_id l)). auto.
                  congruence.
                + transitivity (k0 * H); auto.
                  destruct l; simpl in H0; try discriminate.
                  inv Hsort.
                  inv H0.              
                  simpl in H7.
                  destruct (zle k0 k1) as [Hle| Hle]; try discriminate.
                  apply Zmult_lt_0_le_compat_r; omega.                
            }
          }
          {
            unfold config_pool_valid.
            intros * Hi0 Hconfi0.
            unfold do_enlarge in Hconfi0.
            rewrite Hconf0 in Hconfi0.
            destruct (zeq i id0) as [Heq|Hneq].
            {
              rewrite Heq in *.
              rewrite ZMap.gss in Hconfi0.
              inv Hconfi0.
              destruct l; simpl in Hhd; try discriminate.
              inv Hhd.
              assert (Hin: In (EParam id0 k0) (EParam id0 k0 :: l)).
              {
                simpl.
                auto.
              }
              specialize (Hbound _ _ Hin).
              unfold schedulability_criterion in sched.
              destruct sched as [Hmod ?].              
              assert (Hk0: 1 < k0 <= H).
              {
                destruct (ZMap.get id0 conf) eqn:Hconff.
                inv Hbound.
                specialize (Hconfvalid _ _ _ Hi0 Hconff).
                split; try omega.
                unfold is_a_hyperperiod in Hmod.
                destruct Hmod.
                specialize (H3 _ _ _ Hi0 Hconff).
                apply period_le_hyper_period; omega.
              }
              specialize (Hconfivalid _ _ _ Hi0 Hconf0).
              nia.
            }
            {
              rewrite ZMap.gso in Hconfi0; auto.
              specialize (Hconfivalid _ _ _ Hi0 Hconfi0).
              auto.
            }
          }
        }
      }
    }
  Qed.
    
  Lemma do_enlarge_step_schedulable:
    forall confi l id0 k0,
      list_OK N H confi l ->
      system_schedulable N confi ->
      hd_error l = Some (EParam id0 k0) ->
      system_schedulable N (do_enlarge id0 k0 confi).
  Proof.
    intros * Hlist Hsched Hhd.
    destruct (ZMap.get id0 confi) as [T0 C0] eqn:Hconf0.
    
    assert (HT0: T0 = H).
    {
      eapply list_OK_head; eauto.
    }
    subst.
    assert (Hin: In (EParam id0 k0) l).
    {
      destruct l; simpl in Hhd; try discriminate.
      inv Hhd.      
      simpl.
      auto.
    }    

    assert (HOK: list_OK N H (do_enlarge id0 k0 confi) (tl l)).
    {
      apply do_enlarge_step_list_OK; auto.
    }

    destruct HOK as [_ [_ [_ [_ Hvalid']]]].
    unfold list_OK in Hlist.
    destruct Hlist as [_ [Hbound [_ [Hlist Hvalid]]]].

    assert (Hid: 0 <= id0 < N).
    {
      specialize (Hbound _ _ Hin).
      omega.
    }

    assert (1 < k0).
    {
      destruct (Hbound _ _ Hin) as [_ Hbound'].
      destruct (ZMap.get id0 conf) eqn:Hconff.
      specialize (Hconfvalid _ _ _ Hid Hconff).
      omega.
    }
    
    eapply enlarge_preserves_schedulability1; eauto.
    intros * Hii Hconfi.
    specialize (Hlist _ _ Hhd _ _ _ Hii Hconfi).
    destruct (in_dec Z.eq_dec i (extract_id l)); subst.
    - specialize (Hbound _ _ Hin).
      destruct Hbound.
      replace H with (1 * H) at 1; try omega.
      apply Zmult_lt_0_le_compat_r; omega.
    - auto.
  Qed.
  
  Lemma do_enlarge_sequence_schedulable:
    forall confi l,
      list_OK N H confi l ->
      system_schedulable N confi ->
      system_schedulable N (do_enlarge_sequence confi l).
  Proof.
    intros *.
    revert confi.
    induction l.
    {
      intros.
      simpl.
      auto.
    }
    {
      intros * HOK Hsched.
      destruct a as [id k].
      simpl.
      assert (H1: system_schedulable N (do_enlarge id k confi)).
      {
        eapply do_enlarge_step_schedulable; eauto.
      }
      assert (H2: list_OK N H (do_enlarge id k confi) (tl (EParam id k :: l))).
      {
        eapply do_enlarge_step_list_OK; eauto.
      }
      eapply IHl; auto.
    }
  Qed.

  Definition list_val (Ni: Z) (k: Z) (confi: PrioConfigPool) (l: list enlargeParam): Prop :=
    forall i T C,
      0 <= i < Ni ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      ((not (In i (extract_id l)) /\ ZMap.get i confi = mkPrioConfigValid (H * T) (H * C)) \/
      ((In i (extract_id l)) /\ ZMap.get i confi = mkPrioConfigValid H (H / T * C))).  

  Lemma do_enlarge_step_val:
    forall confi l id0 k0,
      list_OK N H confi l ->
      list_val N H confi l ->
      hd_error l = Some (EParam id0 k0) ->
      list_val N H (do_enlarge id0 k0 confi) (tl l).
  Proof.
    intros * HOK Hval Hhd.
    unfold list_val in *.
    intros * Hi Hconfi.

    destruct l; simpl in Hhd; try discriminate.
    inv Hhd.
    simpl.
    
    specialize (Hval _ _ _ Hi Hconfi).

    destruct (zeq i id0) as [Heq|Hneq].
    {
      rewrite Heq in *. clear Heq.
      left.
      assert (Hin: In id0 (extract_id (EParam id0 k0 :: l))).
      {
        simpl.
        auto.
      }
      destruct Hval as [[? ? ] | [Hin1 Hval1]]; try congruence.

      assert (Hin': ~ In id0 (extract_id l)).
      {
        unfold list_OK in HOK.
        destruct HOK as [_ [HOKk [Hdup _]]].
        simpl in Hdup.
        inv Hdup.
        auto.
      }

      assert (Hkk: k0 = T).
      {
        unfold list_OK in HOK.
        destruct HOK as [_ [HOKk [Hdup _]]].
        assert (Hin2: In (EParam id0 k0) (EParam id0 k0 :: l)).
        {
          simpl.
          auto.
        }
        specialize (HOKk _ _ Hin2).
        rewrite Hconfi in HOKk.
        omega.
      }
      rewrite Hkk in *.

      split; auto.

      unfold do_enlarge.
      rewrite Hval1.
      rewrite ZMap.gss.

      replace (T * H) with (H * T) by lia.
      specialize (Hconfvalid _ _ _ Hi Hconfi).
      destruct sched.
      unfold is_a_hyperperiod in H0.
      destruct H0.
      specialize (H2 _ _ _ Hi Hconfi).

      assert (H5: T * (H / T * C) = H / T * T * C). lia.
      rewrite mod_0_divide_mul in H5; try omega.
      rewrite H5.
      auto.
    }
    {
      unfold do_enlarge.
      destruct (ZMap.get id0 confi).
      rewrite ZMap.gso; auto.
      
      destruct Hval as [[Hin1 Hval1] | [Hin2 Hval2]].
      {
        left.
        split; auto.
        simpl in Hin1.
        unfold not.
        intros H5.        
        assert (id0 = i \/ In i (extract_id l)). auto.
        congruence.
      }
      {
        right.
        split; auto.
        simpl in Hin2.
        destruct Hin2; auto.
        congruence.
      }
    }
  Qed.
      
  Lemma do_enlarge_sequence_val:
    forall confi l,
      list_OK N H confi l ->
      list_val N H confi l ->
      list_val N H (do_enlarge_sequence confi l) nil.
  Proof.
    intros *.
    revert confi.
    induction l.
    {
      intros.
      simpl.
      auto.
    }
    {
      intros * HOK Hval.
      destruct a as [id k].
      simpl.

      assert (H1: list_val N H (do_enlarge id k confi) (tl (EParam id k :: l))).
      {
        apply do_enlarge_step_val; auto.
      }
      assert (H2: list_OK N H (do_enlarge id k confi) (tl (EParam id k :: l))).
      {
        eapply do_enlarge_step_list_OK; eauto.
      }
      eapply IHl; auto.
    }
  Qed.
  
  Definition confseq: PrioConfigPool := do_enlarge_sequence conf0 enlarge_sequence.

  Lemma list_OK0:
    list_OK N H conf0 enlarge_sequence.
  Proof.
    unfold list_OK.
    split.
    {
      (* sorted *)
      unfold enlarge_sequence.
      apply Sorted_sort.
    }
    {
      split.
      {
        (* bounded *)
        unfold enlarge_sequence.
        pose proof (Permuted_sort (get_enlarge_params_aux (Z.to_nat N) conf)) as Hperm.
        intros * Hinsort.
        assert (Hin: In (EParam id k) (get_enlarge_params_aux (Z.to_nat N) conf)).
        {
          apply Permutation_sym in Hperm.
          eapply Permutation_in; eauto.
        }
        pose proof (get_enlarge_id_bound _ _ _ _ Hin) as Hid.
        split; auto.
        destruct (ZMap.get id conf) as [T C] eqn:Hconfi.
        specialize (Hconfvalid _ _ _ Hid Hconfi).
        pose proof (get_enlarge_k_value _ _ _ _ _ _ Hin Hconfi) as Hk.
        omega.
      }
      {
        split.
        {
          (* dup *)
          unfold enlarge_sequence.
          pose proof (Permuted_sort (get_enlarge_params_aux (Z.to_nat N) conf)) as Hperm.
          pose proof (get_enlarge_no_dup (Z.to_nat N) conf) as Hdup.
          pose proof (permutation_extract_id _ _ Hperm) as Hperme.
          eapply permutation_preserves_extract_id_nodup; eauto.
        }
        {
          split.
          {
            (* list *)
            intros * Hhd * Hid Hconfid.
            unfold conf0 in Hconfid.
            assert (HT: T = H).
            {
              assert (H10: 0 <= id < Z.of_nat (Z.to_nat N)).
              {
                rewrite Z2Nat.id; omega.
              }
              destruct (ZMap.get id conf) eqn:Hconff.
              pose proof (enlarge_to_hyper_period _ _ _ _ _ _ _ _ H10 Hconff Hconfid).
              omega.
            }
            subst.
            destruct (in_dec Z.eq_dec id (extract_id enlarge_sequence)); auto.
            contradiction n.
            unfold enlarge_sequence.
            assert (Hperm: Permutation (extract_id (get_enlarge_params_aux (Z.to_nat N) conf))
                                       (extract_id (sort (get_enlarge_params_aux (Z.to_nat N) conf)))).
            {
              apply permutation_extract_id.
              apply Permuted_sort.
            }
            eapply Permutation_in; eauto.
            destruct (ZMap.get id conf) as [T0 C0] eqn:Hconf.
            apply extract_id_any_in with (k := T0).
            eapply get_enlarge_exhaustive; eauto.
          }
          {
            (* config_pool_valid *)
            unfold conf0.
            unfold config_pool_valid.
            intros * Hi Hconfi.
            assert (H10: 0 <= i < Z.of_nat (Z.to_nat N)).
            {
              rewrite Z2Nat.id; omega.
            }
            destruct (ZMap.get i conf) eqn:Hconff.
            pose proof (enlarge_to_hyper_period _ _ _ _ _ _ _ _ H10 Hconff Hconfi) as H11.
            inv H11.
            specialize (Hconfvalid _ _ _ Hi Hconff).
            destruct sched.
            destruct H0.
            specialize (H2 _ _ _ Hi Hconff).
            assert (0 < H / z * z0 < H).
            {
              apply budget_valid_after_hyper; omega.
            }
            nia.
          }
        }
      }
    }
  Qed.

  Lemma list_val0:
    list_val N H conf0 enlarge_sequence.
  Proof.
    unfold list_val.
    intros * Hi Hconfi.
    right.
    split.
    {
      unfold enlarge_sequence.
      assert (Hin: In (EParam i T) (get_enlarge_params_aux (Z.to_nat N) conf)).
      {
        apply get_enlarge_exhaustive with (C:=C); auto.
      }
      apply extract_id_any_in with (k := T); auto.
      pose proof (Permuted_sort (get_enlarge_params_aux (Z.to_nat N) conf)).
      apply Permutation_in with (l := (get_enlarge_params_aux (Z.to_nat N) conf)); auto.
    }
    {
      unfold conf0.
      assert (H1: 0 <= i < Z.of_nat (Z.to_nat N)).
      {
        rewrite Z2Nat.id; omega.
      }
      destruct (ZMap.get i (enlarge_to_hyper H (Z.to_nat N) conf)) eqn:Hconff.
      pose proof (enlarge_to_hyper_period _ _ _ _ _ _ _ _ H1 Hconfi Hconff).
      inv H0.
      auto.
    }
  Qed.
    
  Lemma confseq_schedulable:
    system_schedulable N confseq.
  Proof.
    unfold confseq.
    eapply do_enlarge_sequence_schedulable; eauto.
    eapply list_OK0.
    eapply conf0_schedulable.
  Qed.

  Lemma list_val_confseq:
    list_val N H confseq nil.
  Proof.
    unfold confseq.
    apply do_enlarge_sequence_val; auto.
    apply list_OK0.    
    apply list_val0.
  Qed.
  
  Lemma confseq_k_times_conf:
    conf_k_times N H conf confseq.
  Proof.
    unfold conf_k_times.
    pose proof list_val_confseq as Hval.
    intros * Hi Hconfi.
    unfold list_val in Hval.
    specialize (Hval _ _ _ Hi Hconfi).
    simpl in Hval.
    destruct Hval as [[? Hv ] | [? Hv]]; try omega.
    auto.
  Qed.

  Lemma conf_schedulable:
    system_schedulable N conf.
  Proof.
    pose proof confseq_k_times_conf as Hk.
    pose proof confseq_schedulable.

    assert (Hconf: config_pool_valid N conf).
    {
      unfold config_pool_valid.
      intros * Hi Hconfi.
      specialize (Hconfvalid _ _ _ Hi Hconfi).

      assert (T < Int.max_unsigned).
      {
        destruct sched.
        destruct H1.
        specialize (H3 _ _ _ Hi Hconfi).
        assert (H5: T <= H).
        {
          apply period_le_hyper_period; auto; try omega.
        }

        nia.
      }
      omega.
    }

    assert (Hconff: config_pool_valid N confseq).
    {
      unfold config_pool_valid.
      intros i T' C' Hi Hconfi'.
      destruct (ZMap.get i conf) as [T C] eqn:Hconfi.
      specialize (Hk _ _ _ Hi Hconfi).
      rewrite Hk in Hconfi'.
      inv Hconfi'.

      destruct sched.
      specialize (Hconf _ _ _ Hi Hconfi).
      
      assert (H10: T <= H).
      {
        unfold is_a_hyperperiod in H1.
        destruct H1.
        specialize (H3 _ _ _ Hi Hconfi).
        apply period_le_hyper_period; auto; try omega.
      }
      nia.
    }
    
    eapply shrink_preserves_schedulability1 with (k := H) (conf2 := confseq); eauto.
  Qed.
    
End TRANSITIONING.

Section MAIN.
  
  (* main theorem: schedulability test is valid *)
  
  Theorem schedulability_main (H: Z) (N: Z) (conf: PrioConfigPool):
    0 < N ->
    1 < H ->
    (forall i T C : Z, 0 <= i < N -> ZMap.get i conf = mkPrioConfigValid T C -> 1 < T /\ 0 < C < T) ->
    schedulability_criterion H N conf ->
    system_schedulable N conf.
  Proof.
    intros.
    eapply conf_schedulable; eauto.
  Qed.
        
End MAIN.



