Require Import Coqlib.
Require Import Maps.
Require Import VTDefs.
Require Import Integers.
Require Import Values.
Require Import LemmaVTMap.
Require Import LemmaMaths.
Require Import Psatz.

Local Open Scope Z_scope.

Section KTIMES.

  Definition conf_k_times (N k: Z) (conf1 conf2: PrioConfigPool) : Prop :=
    forall i T1 C1,
      0 <= i < N ->
      ZMap.get i conf1 = mkPrioConfigValid T1 C1 ->
      ZMap.get i conf2 = mkPrioConfigValid (k * T1) (k * C1).

  Definition conf_k_divisor (N k: Z) (conf: PrioConfigPool) : Prop :=
    forall i T C,
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      T mod k = 0 /\ C mod k = 0.      
  
  Definition TAC_k_times (N k: Z) (TAC1 TAC2: TimeContainerPool) : Prop :=
    forall i t,
      0 <= i < N ->
      0 <= t ->
      ZMap.get i TAC1 (t / k) = (ZMap.get i TAC2 t) / k.

  Definition TAC_mod_k_boundary (N k bound: Z) (TAC: TimeContainerPool) : Prop :=
    forall i t,
      0 <= i < N ->
      0 <= t <= bound ->
      t mod k = 0 ->
      (ZMap.get i TAC t) mod k = 0.

  Lemma TAC_mod_k_boundary_shrink:
    forall N k b1 b2 TAC,
      b1 <= b2 ->
      TAC_mod_k_boundary N k b2 TAC ->
      TAC_mod_k_boundary N k b1 TAC.
  Proof.
    intros * Hb Hmod.
    unfold TAC_mod_k_boundary in *.
    intros * Hi Ht Hmodd.
    apply Hmod; auto.
    omega.
  Qed.
  
  Lemma conf_k_times_implies_divisor:
    forall k conf1 conf2 N,
      conf_k_times N k conf1 conf2 ->
      conf_k_divisor N k conf2.
  Proof.
    intros * Hconfk.
    unfold conf_k_times in Hconfk.
    unfold conf_k_divisor.
    intros * Hi Hconfi.
    destruct (ZMap.get i conf1) as [T1 C1] eqn:Hconf11.
    specialize (Hconfk _ _ _ Hi Hconf11).
    rewrite Hconfk in Hconfi.
    inv Hconfi.
    rewrite Z.mul_comm.
    rewrite Z_mod_mult.
    rewrite Z.mul_comm.
    rewrite Z_mod_mult.
    auto.
  Qed.

  Lemma priEntry_lt_k_times:
    forall k d1 i1 d2 i2,
      0 < k ->
      priEntry_lt (PriEntry (k * d1) i1) (PriEntry (k * d2) i2) =
      priEntry_lt (PriEntry d1 i1) (PriEntry d2 i2).      
  Proof.
    intros * Hk.
    unfold priEntry_lt.
    destruct (zlt d1 d2).
    {
      rewrite zlt_true; try nia.
      auto.
    }
    {
      rewrite zlt_false; try nia.
      destruct (zeq d1 d2).
      {
        rewrite e.
        rewrite zeq_true.
        auto.
      }
      {
        rewrite zeq_false; try nia.
        auto.
      }
    }
  Qed.
      
End KTIMES.
  

Section SCHEDULE_BIGSTEP.

  Inductive VT_valid_bigstep: Z (* pid *) -> Z (* tick *) -> Z (* N *) -> PrioConfigPool -> Z -> Prop :=
    | VT_valid_bigstep_stagnates: forall i t N conf k,
      1 < k ->
      0 <= t ->
      id_is_blocked i (t / k * k) N conf (TAC_pool_at_t t (Z.to_nat N) conf) ->
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t =
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) (t / k * k) ->
      VT_valid_bigstep i t N conf k
    | VT_valid_bigstep_increase: forall i t N conf k,
      1 < k ->
      0 <= t ->
      id_not_blocked i (t / k * k) N conf (TAC_pool_at_t t (Z.to_nat N) conf) ->
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t =
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) (t / k * k) + t mod k ->
      VT_valid_bigstep i t N conf k.      

  Lemma vt_valid_bigstep_mod_0:
    forall i t N conf k,
      0 <= t ->
      1 < k ->
      t mod k = 0 ->
      0 <= i < N ->
      VT_valid_bigstep i t N conf k.
  Proof.
    intros * Ht Hk Hmod Hi.
    assert (HN: 0 < N). omega.
    assert (Htt: t / k * k = t).
    {
      rewrite mod_0_divide_mul; omega.
    }
    
    destruct (id_blocked_dec i t N conf HN) as [H1 | H1].
    {
      apply VT_valid_bigstep_stagnates; auto.
      - rewrite Htt. auto.
      - rewrite Htt. auto.
    }
    {
      apply VT_valid_bigstep_increase; auto; rewrite Htt; auto; try omega.
    }
  Qed.

  Lemma TAC_mod_k_boundary_0:
    forall N k bound conf,
      TAC_mod_k_boundary N k bound (TAC_pool_at_t 0 (Z.to_nat N) conf).
  Proof.
    intros.
    unfold TAC_pool_at_t, TAC_mod_k_boundary.
    simpl.
    intros * Hi Ht Hmod.
    unfold real_time_container_pool.
    rewrite ZMap.gi.
    auto.
  Qed.

  Lemma vt_map_not_finished_k_divisor:
    forall t k T C vt,
      1 < k ->
      0 <= t ->
      1 < T ->
      0 < C < T ->
      t mod k <> 0 ->
      (vt (t / k * k)) mod k = 0 ->
      (vt (t / T * T)) mod k = 0 ->
      T mod k = 0 ->
      C mod k = 0 ->
      monotonic_continuous vt ->
      vt_map_not_finished t T C vt = vt_map_not_finished (t / k * k) T C vt.
  Proof.
    intros * Hk Ht HT HC Hmodt Hmodvt Hmodvt' HmodT HmodC Hmono.
    unfold vt_map_not_finished.
    assert (HTk: k <= T).
    {
      apply period_le_hyper_period; omega.
    }
    assert (H1: 0 < T / k). apply Z.div_str_pos; omega.
    assert (H2: t / T * T <= t / k * k).
    {
      assert (HTT: T = T / k * k).
      {
        rewrite mod_0_divide_mul; omega.
      }
      rewrite HTT.
      apply period_enlarge_le; try omega.
    }
    assert (H3: t / k * k <= t).
    {
      apply divide_mul_le; omega.
    }
    assert (H4: 0 <= t / k * k).
    {
      apply div_mul_nonneg; omega.
    }
    assert (H5: vt (t / k * k) <= vt t).
    {
      apply monotonic_continuous_monotonic; auto.
    }

    rewrite divide_same_mod_k; try omega.

    assert (H7: t < t / k * k + k).
    {
      apply x_lt_floor_plus; omega.
    }
    
    assert (H6: vt t - vt (t / k * k) < k).
    {
      assert (vt t - vt (t / k * k) <= t - t / k * k).
      {
        apply monotonic_continuous_range; auto; omega.
      }
      omega.
    }
    
    destruct (zlt (vt (t / k * k) - vt (t / T * T)) C).
    {
      assert (Hlt: vt t < vt (t / T * T) + C).
      {
        replace (vt t) with (vt (t / k * k) + (vt t - vt (t / k * k))) by omega.
        apply zlt_incr_less_k with (k:=k); try omega.
        rewrite Zplus_mod.
        rewrite Hmodvt'.
        rewrite HmodC.
        auto.
      }
      destruct (zlt (vt t - vt (t / T * T)) C); try discriminate; try omega; auto.      
    }
    {
      destruct (zlt (vt t - vt (t / T * T)) C); try discriminate; try omega; auto.
    }
  Qed.
    
  Lemma id_is_blocked_whole_k:
    forall N k conf t i TAC,
      0 <= t ->
      1 < k ->
      0 <= i < N ->
      conf_k_divisor N k conf ->
      config_pool_valid N conf ->
      TAC_mod_k_boundary N k t TAC ->
      id_is_blocked i (t / k * k) N conf TAC ->
      (forall y, 0 <= y < N -> monotonic_continuous (ZMap.get y TAC)) ->
      id_is_blocked i t N conf TAC.
  Proof.
    intros * Ht Hk Hi Hconfk Hconfvalid HTACmod Hblock Hmono.
    destruct (zeq (t mod k) 0) as [Hmod | Hmod].
    {
      rewrite mod_0_divide_mul in Hblock; try omega.
      auto.
    }
    {
      unfold id_is_blocked in *.
      destruct Hblock as [x [Hx [Hx' Hval]]].
      exists x.
      split; auto.
      split; auto.
      destruct (ZMap.get x conf) as [T C] eqn:Hconfx.
      destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconfi.
      destruct Hval as [Hfinish Hpri].      

      assert (H1: 1 < T /\ 0 < C < T).
      {
        specialize (Hconfvalid _ _ _ Hx Hconfx).
        omega.
      }
      assert (H2: 0 <= t / k * k).
      {
        apply div_mul_nonneg; omega.
      }
      assert (H3: t / k * k <= t).
      {
        apply divide_mul_le; omega.
      }
      assert (H4: (t / k * k) mod k = 0).
      {
        apply Z_mod_mult.
      }
      assert (H5: ZMap.get x TAC (t / k * k) mod k = 0).
      {
        apply HTACmod; auto; try omega.
      }
      assert (H11: 0 <= t / T * T <= t).
      {
        split.
        - apply div_mul_nonneg; omega.
        - apply divide_mul_le; omega.
      }
      assert (H14: T mod k = 0).
      {
        specialize (Hconfk _ _ _ Hx Hconfx).
        omega.
      }
      assert (H12: (t / T * T) mod k = 0).
      {
        rewrite Zmult_mod.
        rewrite H14.
        rewrite Z.mul_0_r.
        auto.
      }
      assert (H9: ZMap.get x TAC (t / T * T) mod k = 0).
      {
        apply HTACmod; auto; try omega.
      }
      
      assert (H6: T mod k = 0 /\ C mod k = 0).
      {
        specialize (Hconfk _ _ _ Hx Hconfx).
        omega.
      }        
        
      assert (Hfinish': vt_map_not_finished t T C (ZMap.get x TAC) = true).
      {
        rewrite vt_map_not_finished_k_divisor with (k := k); auto; try omega.
      }

      assert (H7: 1 < Ti).
      {
        specialize (Hconfvalid _ _ _ Hi Hconfi).
        omega.
      }
      assert (H8: Ti mod k = 0).
      {
        specialize (Hconfk _ _ _ Hi Hconfi).
        omega.
      }        
      
      assert (Hpri': priEntry_lt (PriEntry (t / T * T + T) x) (PriEntry (t / Ti * Ti + Ti) i) = true).
      {
        rewrite divide_same_mod_k in Hpri; try omega.
        rewrite divide_same_mod_k in Hpri; try omega.
        auto.
      }

      split; auto.
    }
  Qed.

  Lemma id_not_blocked_whole_k:
    forall N k conf t i TAC,
      0 <= t ->
      1 < k ->
      0 <= i < N ->
      conf_k_divisor N k conf ->
      config_pool_valid N conf ->
      TAC_mod_k_boundary N k t TAC ->
      id_not_blocked i (t / k * k) N conf TAC ->
      (forall y, 0 <= y < N -> monotonic_continuous (ZMap.get y TAC)) ->      
      id_not_blocked i t N conf TAC.
  Proof.
    intros * Ht Hk Hi Hconfk Hconfvalid HTACmod Hblock Hmono.
    destruct (zeq (t mod k) 0) as [Hmod | Hmod].
    {
      rewrite mod_0_divide_mul in Hblock; try omega.
      auto.
    }
    {
      unfold id_not_blocked in *.
      intros * Hi0 Hi0' Hconfi0 Hconfi.
      specialize (Hblock _ _ _ _ _ Hi0 Hi0' Hconfi0 Hconfi).      

      assert (H1: 1 < T /\ 0 < C < T).
      {
        specialize (Hconfvalid _ _ _ Hi0 Hconfi0).
        omega.
      }
      assert (H2: 0 <= t / k * k).
      {
        apply div_mul_nonneg; omega.
      }
      assert (H3: t / k * k <= t).
      {
        apply divide_mul_le; omega.
      }
      assert (H4: (t / k * k) mod k = 0).
      {
        apply Z_mod_mult.
      }
      assert (H5: ZMap.get i0 TAC (t / k * k) mod k = 0).
      {
        apply HTACmod; auto; try omega.
      }
      assert (H11: 0 <= t / T * T <= t).
      {
        split.
        - apply div_mul_nonneg; omega.
        - apply divide_mul_le; omega.
      }
      assert (H14: T mod k = 0).
      {
        specialize (Hconfk _ _ _ Hi0 Hconfi0).
        omega.
      }
      assert (H12: (t / T * T) mod k = 0).
      {
        rewrite Zmult_mod.
        rewrite H14.
        rewrite Z.mul_0_r.
        auto.
      }
      assert (H9: ZMap.get i0 TAC (t / T * T) mod k = 0).
      {
        apply HTACmod; auto; try omega.
      }
      
      assert (H6: T mod k = 0 /\ C mod k = 0).
      {
        specialize (Hconfk _ _ _ Hi0 Hconfi0).
        omega.
      }        

      
      assert (Hfinish': vt_map_not_finished t T C (ZMap.get i0 TAC) = vt_map_not_finished (t / k * k) T C (ZMap.get i0 TAC)).
      {
        rewrite vt_map_not_finished_k_divisor with (k := k); auto; try omega.
      }
      rewrite Hfinish'.      

      assert (H7: 1 < Tid).
      {
        specialize (Hconfvalid _ _ _ Hi Hconfi).
        omega.
      }
      assert (H8: Tid mod k = 0).
      {
        specialize (Hconfk _ _ _ Hi Hconfi).
        omega.
      }        

      rewrite divide_same_mod_k in Hblock; try omega.
      rewrite divide_same_mod_k in Hblock; try omega.
      auto.
    }
  Qed.

  Lemma conf_k_divisor_then_vt_valid_bigstep:
    forall i t N conf k,
      0 <= t ->      
      1 < k ->
      0 <= i < N ->
      conf_k_divisor N k conf ->
      config_pool_valid N conf ->
      VT_valid_bigstep i t N conf k /\ TAC_mod_k_boundary N k t (TAC_pool_at_t t (Z.to_nat N) conf).
  Proof.
    intros * Ht Hk.
    revert i.
    remember (Z.to_nat t) as nt.
    assert (Htt: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    rewrite Htt in *. clear Htt Heqnt Ht.
    induction nt.
    {
      intros * Hi Hconfk Hconf.
      simpl.
      split.
      - apply vt_valid_bigstep_mod_0; auto.
        omega.
      - apply TAC_mod_k_boundary_0.
    }
    {
      intros * Hi Hconfk Hconf.
      pose proof IHnt as IH.
      specialize (IHnt _ Hi Hconfk Hconf).
      destruct IHnt as [IHvt IHmod].
      pose proof IHvt as Hvtvalid.
      replace (Z.of_nat (S nt)) with (Z.of_nat nt + 1) by xomega.
      destruct (zeq ((Z.of_nat nt + 1) mod k) 0) as [Hmod | Hmod].
      {
        split.
        {
          apply vt_valid_bigstep_mod_0; auto.
          omega.
        }
        {
          unfold TAC_mod_k_boundary.
          intros x t0 Hx Ht0 Ht0mod.
          specialize (IH x Hx Hconfk Hconf).
          destruct IH as [IHx1 IHx2].

          assert (Hmodstart: ZMap.get x (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf) (Z.of_nat nt / k * k) mod k = 0).
          {
            assert (Ht0': 0 <= Z.of_nat nt / k * k <= Z.of_nat nt).
            {
              split.
              - apply div_mul_nonneg; omega.
              - apply divide_mul_le; omega.
            }
            assert (Ht1: (Z.of_nat nt / k * k) mod k = 0).
            {
              apply Z_mod_mult.
            }
            
            specialize (IHmod _ _ Hx Ht0' Ht1).
            auto.
          }
            
          destruct (zeq t0 (Z.of_nat nt + 1)) as [Heq | Hneq].
          {
            inv IHx1.
            {
              (* blocked *)
              assert (Hblock': id_is_blocked x (Z.of_nat nt) N conf (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf)).
              {
                apply id_is_blocked_whole_k with (k := k); auto.
                - intros y Hy.
                  apply TAC_pool_all_monotonic_continuous; auto.
                  rewrite Z2Nat.id; omega.
              }

              assert (Hvt1: ZMap.get x (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf) (Z.of_nat nt + 1) = ZMap.get x (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf) (Z.of_nat nt)).
              {
                apply id_is_blocked_TAC_stagnates; auto.
              }

              rewrite Hvt1.
              rewrite H2.
              auto.
            }
            {
              assert (Hblock': id_not_blocked x (Z.of_nat nt) N conf (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf)).
              {
                apply id_not_blocked_whole_k with (k := k); auto.
                - intros y Hy.
                  apply TAC_pool_all_monotonic_continuous; auto.
                  rewrite Z2Nat.id; omega.
              }
              
              assert (Hvt1: ZMap.get x (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf) (Z.of_nat nt + 1) = ZMap.get x (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf) (Z.of_nat nt) + 1).
              {
                apply id_not_blocked_TAC_increases; auto.
              }

              rewrite Hvt1.
              rewrite H2.

              replace (ZMap.get x (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf) (Z.of_nat nt / k * k) + Z.of_nat nt mod k + 1) with (ZMap.get x (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf) (Z.of_nat nt / k * k) + (Z.of_nat nt mod k + 1)) by omega.              
              rewrite Zplus_mod.
              rewrite Hmodstart.
              rewrite Z.add_0_l.
              assert (H10: Z.of_nat nt mod k = k - 1).
              {
                apply mod_0_minus_1_2; try omega.
              }
              rewrite H10.
              replace (k -1 + 1) with k by omega.
              rewrite Z_mod_same_full.
              auto.
            }
          }
          {
            assert (Ht0': 0 <= t0 <= Z.of_nat nt). omega.
            specialize (IHmod _ _ Hx Ht0' Ht0mod).
            rewrite vertical_map_canonical; try omega.
            rewrite vertical_map_canonical in IHmod; try omega.            
          }
        }
      }
      {
        assert (Hdiv: (Z.of_nat nt + 1) / k = Z.of_nat nt / k).
        {
          rewrite <- z_divide_same; try omega.
        }
        assert (H1: (Z.of_nat nt) / k * k <= Z.of_nat nt).
        {
          apply divide_mul_le; omega.
        }
        assert (H3: 0 <= Z.of_nat nt / k * k).
        {
          apply div_mul_nonneg; omega.
        }

        assert (Hmod': TAC_mod_k_boundary N k (Z.of_nat nt + 1)
     (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf)).
        {
          unfold TAC_mod_k_boundary.
          intros * Hi0 Ht0 Hmodt0.

          assert (H10: 0 <= t0 <= Z.of_nat nt).
          {
            destruct (zeq t0 (Z.of_nat nt + 1)); try omega.
            rewrite e in Hmodt0.
            omega.
          }          
          specialize (IHmod _ _ Hi0 H10 Hmodt0).
          rewrite vertical_map_canonical; try omega.
          rewrite vertical_map_canonical in IHmod; try omega.          
        }

        split; auto.
        
        inv IHvt.
        {
          assert (Hvt': VT_valid_bigstep i (Z.of_nat nt + 1) N conf k).
          {
            (* stagnates *)
            apply VT_valid_bigstep_stagnates; auto; try omega.
            - rewrite Hdiv.
              rewrite id_is_blocked_canonical; auto; try omega.
              rewrite id_is_blocked_canonical in H2; auto; try omega.            
            - rewrite Hdiv.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry in H4. 
              rewrite vertical_map_canonical in H4; try omega.           
              rewrite H4.
              symmetry.
              
              apply id_is_blocked_TAC_stagnates; auto.
              apply id_is_blocked_whole_k with (k:=k); auto.
              + intros y Hy.
                apply TAC_pool_all_monotonic_continuous; auto.
                rewrite Z2Nat.id; omega.
          }
          auto.
        }
        {
          apply VT_valid_bigstep_increase; auto; try omega.
          - rewrite Hdiv.
            rewrite id_not_blocked_canonical; auto; try omega.
            rewrite id_not_blocked_canonical in H2; auto; try omega.            
          - rewrite Hdiv.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry in H4. 
            rewrite vertical_map_canonical in H4; try omega.
            assert (H10: (Z.of_nat nt + 1) mod k = Z.of_nat nt mod k + 1).
            {
              apply mod_increase; try omega.
            }
            rewrite H10.
            replace (ZMap.get i (TAC_pool_at_t (Z.of_nat nt / k * k) (Z.to_nat N) conf) (Z.of_nat nt / k * k) + (Z.of_nat nt mod k + 1)) with ((ZMap.get i (TAC_pool_at_t (Z.of_nat nt / k * k) (Z.to_nat N) conf) (Z.of_nat nt / k * k) + Z.of_nat nt mod k) + 1) by omega.
            rewrite H4.
            symmetry.
            
            apply id_not_blocked_TAC_increases; auto.
            apply id_not_blocked_whole_k with (k:=k); auto.
            + intros y Hy.
              apply TAC_pool_all_monotonic_continuous; auto.
              rewrite Z2Nat.id; omega.
        }
      }
    }
  Qed.

End SCHEDULE_BIGSTEP.




Section PRIQUEUE.

  Lemma id_is_blocked_sync:
    forall N k i conf1 conf2 t TAC1 TAC2,
      1 < k ->
      conf_k_times N k conf1 conf2 ->
      config_pool_valid N conf1 ->
      config_pool_valid N conf2 ->
      TAC_k_times N k TAC1 TAC2 ->
      0 <= i < N ->
      0 <= t ->
      id_is_blocked i t N conf1 TAC1 ->
      TAC_mod_k_boundary N k (t * k) TAC2 ->      
      id_is_blocked i (t * k) N conf2 TAC2.
  Proof.
    intros * Hk Hconfk Hconf1 Hconf2 HTACk Hi Ht Hblock1 HTACmod.
    unfold id_is_blocked in *.
    destruct Hblock1 as [x [Hx [Hx' Hval]]].
    exists x.
    split; auto.
    split; auto.
    destruct (ZMap.get x conf1) as [T C] eqn:Hconf1x.
    destruct (ZMap.get i conf1) as [Ti Ci] eqn:Hconf1i.

    pose proof (Hconfk _ _ _ Hx Hconf1x) as Hconf2x.
    pose proof (Hconfk _ _ _ Hi Hconf1i) as Hconf2i.
    rewrite Hconf2x.
    rewrite Hconf2i.

    assert (Hd1: t * k / (k * T) * (k * T) + k * T = k * (t / T * T + T)).
    {
      replace (t * k) with (k * t) by lia.
      rewrite Zdiv_mult_cancel_l; try omega.
      nia.
    }
    rewrite Hd1.

    assert (Hd2: t * k / (k * Ti) * (k * Ti) + k * Ti = k * (t / Ti * Ti + Ti)).
    {
      replace (t * k) with (k * t) by lia.
      rewrite Zdiv_mult_cancel_l; try omega.
      nia.
    }
    rewrite Hd2.   

    destruct Hval as [Hfinish Hpri].
    assert (Hpri': priEntry_lt (PriEntry (k * (t / T * T + T)) x) (PriEntry (k * (t / Ti * Ti + Ti)) i) = true).
    {
      rewrite priEntry_lt_k_times; auto; try omega.
    }

    assert (HT: 1 < T).
    {
      specialize (Hconf1 _ _ _ Hx Hconf1x).
      omega.
    }
    assert (Htpos: 0 <= t / T).
    {
      apply Z_div_pos; omega.
    }
    assert (Httpos: 0 <= t / T * T * k).
    {
      apply Zmult_gt_0_le_0_compat; try omega.
      nia.
    }
    
    assert (Hfinish': vt_map_not_finished (t * k) (k * T) (k * C) (ZMap.get x TAC2) = true).
    {
      unfold vt_map_not_finished in *.
      unfold TAC_k_times in HTACk.
      assert (Htt: t * k / k = t).
      {
        rewrite Z_div_mult_full; omega.
      }
      rewrite <- Htt in Hfinish.
      rewrite HTACk in Hfinish; try nia.
      rewrite Z_div_mult_full in Hfinish; try omega.

      assert (H1: t * k / (k * T) * (k * T) = t / T * T * k).
      {
        replace (t * k) with (k * t) by lia.
        rewrite Zdiv_mult_cancel_l; try omega.
        nia.
      }
      rewrite H1.
      assert (H2: t / T * T = t / T * T * k / k).
      {
        rewrite Z_div_mult_full; omega.
      }
      rewrite H2 in Hfinish.
      rewrite HTACk in Hfinish; try nia.

      destruct (zlt (ZMap.get x TAC2 (t * k) / k - ZMap.get x TAC2 (t / T * T * k) / k) C); try discriminate.
      assert (Hlt: (ZMap.get x TAC2 (t * k) - ZMap.get x TAC2 (t / T * T * k)) < (k * C)).
      {
        apply zlt_k_times; try omega.        
        - apply HTACmod; auto; try nia.
          apply Z_mod_mult.
      }
      destruct (zlt (ZMap.get x TAC2 (t * k) - ZMap.get x TAC2 (t / T * T * k)) (k * C)); try omega; auto.
    }

    split; auto.
  Qed.

  Lemma id_not_blocked_sync:
    forall N k i conf1 conf2 t TAC1 TAC2,
      1 < k ->
      conf_k_times N k conf1 conf2 ->
      config_pool_valid N conf1 ->
      config_pool_valid N conf2 ->
      TAC_k_times N k TAC1 TAC2 ->
      0 <= i < N ->
      0 <= t ->
      id_not_blocked i t N conf1 TAC1 ->
      TAC_mod_k_boundary N k (t * k) TAC2 ->       
      id_not_blocked i (t * k) N conf2 TAC2.
  Proof.
    intros * Hk Hconfk Hconf1 Hconf2 HTACk Hi Ht Hblock1 HTACmod.
    unfold id_not_blocked in *.
    intros x T' C' Ti' Ci' Hx Hx' Hconf2x Hconf2i.

    destruct (ZMap.get x conf1) as [T C] eqn:Hconf1x.
    pose proof (Hconfk _ _ _ Hx Hconf1x) as Hconf2x'.
    rewrite Hconf2x in Hconf2x'.
    inv Hconf2x'.

    destruct (ZMap.get i conf1) as [Ti Ci] eqn:Hconf1i.
    pose proof (Hconfk _ _ _ Hi Hconf1i) as Hconf2i'.
    rewrite Hconf2i in Hconf2i'.
    inv Hconf2i'.
    assert (Hbb: mkPrioConfigValid Ti Ci = mkPrioConfigValid Ti Ci). auto.    
    specialize (Hblock1 _ _ _ _ _  Hx Hx' Hconf1x Hbb).

    assert (Hd1: t * k / (k * T) * (k * T) + k * T = k * (t / T * T + T)).
    {
      replace (t * k) with (k * t) by lia.
      rewrite Zdiv_mult_cancel_l; try omega.
      nia.
    }
    rewrite Hd1.

    assert (Hd2: t * k / (k * Ti) * (k * Ti) + k * Ti = k * (t / Ti * Ti + Ti)).
    {
      replace (t * k) with (k * t) by lia.
      rewrite Zdiv_mult_cancel_l; try omega.
      nia.
    }
    rewrite Hd2.   

    destruct Hblock1 as [Hfinish | Hpri].
    {
      assert (HT: 1 < T).
      {
        specialize (Hconf1 _ _ _ Hx Hconf1x).
        omega.
      }
      assert (Htpos: 0 <= t / T).
      {
        apply Z_div_pos; omega.
      }
      assert (Httpos: 0 <= t / T * T * k).
      {
        apply Zmult_gt_0_le_0_compat; try omega.
        nia.
      }

      left.
      unfold vt_map_not_finished in *.
      unfold TAC_k_times in HTACk.
      assert (Htt: t * k / k = t).
      {
        rewrite Z_div_mult_full; omega.
      }
      rewrite <- Htt in Hfinish.
      rewrite HTACk in Hfinish; try nia.
      rewrite Z_div_mult_full in Hfinish; try omega.

      assert (H1: t * k / (k * T) * (k * T) = t / T * T * k).
      {
        replace (t * k) with (k * t) by lia.
        rewrite Zdiv_mult_cancel_l; try omega.
        nia.
      }
      rewrite H1.
      assert (H2: t / T * T = t / T * T * k / k).
      {
        rewrite Z_div_mult_full; omega.
      }
      rewrite H2 in Hfinish.
      rewrite HTACk in Hfinish; try nia.

      assert (H3: 0 <= t / T * T * k <= t * k).
      {
        split.
        - nia.
        - assert (t / T * T <= t).
          {
            apply divide_mul_le; omega.
          }
          apply Z.mul_le_mono_nonneg_r; try omega.
      }
      
      destruct (zlt (ZMap.get x TAC2 (t * k) / k - ZMap.get x TAC2 (t / T * T * k) / k) C); try discriminate.
      assert (Hlt: (ZMap.get x TAC2 (t * k) - ZMap.get x TAC2 (t / T * T * k)) >= (k * C)).
      {
        apply zge_k_times; try omega.        
        - apply HTACmod; auto; try omega.
          apply Z_mod_mult.
      }
      destruct (zlt (ZMap.get x TAC2 (t * k) - ZMap.get x TAC2 (t / T * T * k)) (k * C)); try omega; auto.
    }
    {
      right.
      rewrite priEntry_lt_k_times; auto; try omega.
    }
  Qed.
  
End PRIQUEUE.






Section SHRINK.
  
  Lemma TAC_k_times_hold:
    forall N k conf1 conf2 t,
      0 <= t ->
      1 < k ->      
      conf_k_times N k conf1 conf2 ->
      config_pool_valid N conf1 ->
      config_pool_valid N conf2 ->
      TAC_k_times N k (TAC_pool_at_t (t / k) (Z.to_nat N) conf1)
                  (TAC_pool_at_t t (Z.to_nat N) conf2).
  Proof.
    intros * Ht Hk Hconfk Hconf1 Hconf2.
    remember (Z.to_nat t) as nt.
    assert (Htt: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    rewrite Htt in *. clear Heqnt Htt.
    induction nt.
    {
      unfold TAC_pool_at_t.
      simpl.
      unfold TAC_k_times.
      intros * Hi Ht0.
      unfold real_time_container_pool.
      rewrite ZMap.gi.
      auto.
    }
    {
      assert (H1: Z.of_nat (S nt) / k = (Z.of_nat nt + 1) / k). 
      {
        replace (Z.of_nat (S nt)) with (Z.of_nat nt + 1) by xomega.
        auto.
      }
      rewrite H1. 

      replace (Z.of_nat (S nt)) with (Z.of_nat nt + 1) by xomega.
      
      assert (Ht': 0 <= Z.of_nat nt). omega.
      specialize (IHnt Ht'). clear Ht'.

      unfold TAC_k_times.
      intros * Hi Ht0.
      assert (H12: 0 <= t0 / k).
      {
        apply Z_div_pos; omega.
      }
      assert (H14: 0 <= Z.of_nat nt / k).
      {
        apply Z_div_pos; omega.
      }

      assert (Hvt0: ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt) / k =
                    ZMap.get i (TAC_pool_at_t (Z.of_nat nt / k) (Z.to_nat N) conf1) (Z.of_nat nt / k)).
      {
        symmetry.
        apply IHnt; auto.
        omega.
      }      
      
      destruct (zeq ((Z.of_nat nt + 1) mod k) 0) as [Hmod|Hmod].
      {
        assert (H10: (Z.of_nat nt + 1) / k = (Z.of_nat nt / k) + 1).
        {
          apply Zdiv_incr_exact; try omega.
        }
        rewrite H10. 
        
        destruct (zlt t0 (Z.of_nat nt + 1)) as [Hbefore|Hnow].
        {
          assert (H11: t0 / k <= Z.of_nat nt / k).
          {
            apply Z_div_le; try omega.
          }
          rewrite vertical_map_canonical; try omega.
          symmetry.
          rewrite vertical_map_canonical; try omega.

          specialize (IHnt _ _ Hi Ht0).
          rewrite vertical_map_canonical in IHnt; try omega.
          symmetry in IHnt.
          rewrite vertical_map_canonical in IHnt; try omega.          
        }
        {
          assert (H11: Z.of_nat nt / k + 1 <= t0 / k).
          {            
            rewrite <- H10.
            apply Z_div_le; omega.
          }
          rewrite vertical_map_canonical2; try xomega.
          symmetry.
          rewrite vertical_map_canonical2; try xomega.

          assert (Hmono: monotonic_continuous (ZMap.get i (TAC_pool_at_t (Z.of_nat nt / k + 1) (Z.to_nat N) conf1))).
          {
            apply TAC_pool_all_monotonic_continuous; auto; try omega.
            - rewrite Z2Nat.id; omega.
          }
          assert (H20: Z.of_nat nt / k >= 0). omega.
          specialize (Hmono _ H20).
          destruct Hmono as [Hstagnates | Hincrease].
          {
            rewrite Hstagnates.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            rewrite <- Hvt0.

            symmetry in Hstagnates.
            rewrite vertical_map_canonical in Hstagnates; try omega.
            symmetry in Hstagnates.
            
            pose proof (TAC_stagnates_id_is_blocked _ _ _ _ H14 Hi Hstagnates) as Hblock1.

            assert (H30: 0 <= Z.of_nat nt). omega.
            assert (H31: conf_k_divisor N k conf2).
            {
              eapply conf_k_times_implies_divisor; eauto.
            }
            pose proof (conf_k_divisor_then_vt_valid_bigstep _ _ _ _ _ H30 Hk Hi H31 Hconf2).
            destruct H as [Hvtvalid HTACmod].
            
            assert (Hblock2: id_is_blocked i (Z.of_nat nt / k * k) N conf2
                                           (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2)).
            {
              eapply id_is_blocked_sync; eauto.
              - eapply TAC_mod_k_boundary_shrink; eauto.
                apply divide_mul_le; omega.
            }
            
            assert (Hvt: ZMap.get i (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf2) (Z.of_nat nt + 1) = ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt)).
            {
              apply id_is_blocked_TAC_stagnates; auto; try omega.
              apply id_is_blocked_whole_k with (k := k); auto; try omega.
              - intros y Hy.
                apply TAC_pool_all_monotonic_continuous; auto.
                rewrite Z2Nat.id; omega.
            }
            rewrite Hvt.
            auto.
          }
          {
            rewrite Hincrease.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            rewrite <- Hvt0.

            symmetry in Hincrease.
            rewrite vertical_map_canonical in Hincrease; try omega.
            symmetry in Hincrease.

            assert (H30: 0 <= Z.of_nat nt). omega.
            assert (H31: conf_k_divisor N k conf2).
            {
              eapply conf_k_times_implies_divisor; eauto.
            }
            pose proof (conf_k_divisor_then_vt_valid_bigstep _ _ _ _ _ H30 Hk Hi H31 Hconf2).
            destruct H as [Hvtvalid HTACmod].
            
            pose proof (TAC_increases_id_not_blocked _ _ _ _ H14 Hi Hincrease) as Hblock1.
            assert (Hblock2: id_not_blocked i (Z.of_nat nt / k * k) N conf2
                                           (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2)).
            {
              eapply id_not_blocked_sync; eauto.
              - eapply TAC_mod_k_boundary_shrink; eauto.
                apply divide_mul_le; omega.
            }
            
            assert (Hvt: ZMap.get i (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf2) (Z.of_nat nt + 1) = ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt) + 1).
            {
              apply id_not_blocked_TAC_increases; auto; try omega.
              apply id_not_blocked_whole_k with (k := k); auto; try omega.
              - intros y Hy.
                apply TAC_pool_all_monotonic_continuous; auto.
                rewrite Z2Nat.id; omega.              
            }
            rewrite Hvt.

            assert (H18: 0 <= Z.of_nat nt / k * k <= Z.of_nat nt).
            {
              split.
              - apply div_mul_nonneg; omega.
              - apply divide_mul_le; omega.
            }
            
            assert (Hvt1: ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt) = ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) + (Z.of_nat nt) mod k).
            {
              inv Hvtvalid.
              {
                exfalso.
                assert (H19: 0 < N). omega.
                rewrite id_is_blocked_canonical in H2; auto; try omega.
                rewrite id_not_blocked_canonical in Hblock2; auto; try omega. 
                apply (id_blocked_absurd _ _ _ _ H19 H2 Hblock2); auto; try omega.
              }
              {
                auto.
              }
            }
            rewrite Hvt1.

            assert (Hmod': (ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k)) mod k = 0).
            {
              apply HTACmod; auto; try omega.
              - apply Z_mod_mult.
            }

            assert (Hmod'': Z.of_nat nt mod k = k - 1).
            {
              apply mod_0_minus_1_2; omega.
            }
            rewrite Hmod''.

            assert (Hnonneg: 0 <= ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k)).
            {
              apply TAC_pool_at_t_nonneg; try omega.
            }
            
            replace (ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) + (k - 1) + 1) with (ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) + k) by omega.
            rewrite Nadd_div_exact; auto; try omega.
            rewrite Nadd_div_exact; auto; try omega.
            assert (H19: (k - 1) / k = 0).
            {
              rewrite Zdiv.Zdiv_small; try omega.
            }
            rewrite H19.
            rewrite Z_div_same_full; try omega.
          }
        }
      }
      {
        assert (H10: (Z.of_nat nt + 1) / k = (Z.of_nat nt / k)).
        {
          symmetry.
          rewrite z_divide_same; try omega.
        }
        rewrite H10.

        destruct (zlt t0 (Z.of_nat nt + 1)) as [Hbefore|Hnow].
        {
          assert (H11: t0 / k <= Z.of_nat nt / k).
          {
            apply Z_div_le; try omega.
          }
          rewrite vertical_map_canonical; try omega.
          symmetry.
          rewrite vertical_map_canonical; try omega.

          specialize (IHnt _ _ Hi Ht0).
          rewrite vertical_map_canonical in IHnt; try omega.
          symmetry in IHnt.
          rewrite vertical_map_canonical in IHnt; try omega.          
        }
        {
          assert (H11: Z.of_nat nt / k <= t0 / k).
          {            
            apply Z_div_le; omega.
          }
          rewrite vertical_map_canonical2; try xomega.
          symmetry.
          rewrite vertical_map_canonical2; try xomega.

          rewrite <- Hvt0.

          assert (Hmono: monotonic_continuous (ZMap.get i (TAC_pool_at_t (Z.of_nat nt / k + 1) (Z.to_nat N) conf1))).
          {
            apply TAC_pool_all_monotonic_continuous; auto; try omega.
            - rewrite Z2Nat.id; omega.
          }
          assert (H20: Z.of_nat nt / k >= 0). omega.
          specialize (Hmono _ H20).
          
          assert (H30: 0 <= Z.of_nat nt). omega.
          assert (H31: conf_k_divisor N k conf2).
          {
            eapply conf_k_times_implies_divisor; eauto.
          }
          pose proof (conf_k_divisor_then_vt_valid_bigstep _ _ _ _ _ H30 Hk Hi H31 Hconf2).
          destruct H as [Hvtvalid HTACmod].

          inv Hvtvalid.
          {
            (* blocked *)
            assert (Hblocknow: id_is_blocked i (Z.of_nat nt) N conf2 (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2)).
            {
              apply id_is_blocked_whole_k with (k := k); auto.
              - intros y Hy.
                apply TAC_pool_all_monotonic_continuous; auto.
                rewrite Z2Nat.id; omega.
            }
            
            assert (Hvt: ZMap.get i (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf2) (Z.of_nat nt + 1) = ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt)).
            {
              apply id_is_blocked_TAC_stagnates; auto; try omega.
            }
            rewrite Hvt.
            auto.
          }
          {
            (* not blocked *)
            assert (Hblocknow: id_not_blocked i (Z.of_nat nt) N conf2 (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2)).
            {
              apply id_not_blocked_whole_k with (k := k); auto.
              - intros y Hy.
                apply TAC_pool_all_monotonic_continuous; auto.
                rewrite Z2Nat.id; omega.
            }
            
            assert (Hvt: ZMap.get i (TAC_pool_at_t (Z.of_nat nt + 1) (Z.to_nat N) conf2) (Z.of_nat nt + 1) = ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt) + 1).
            {
              apply id_not_blocked_TAC_increases; auto; try omega.
            }
            rewrite Hvt.

            rewrite H3.

            assert (H24: 0 <= Z.of_nat nt / k * k).
            {
              apply div_mul_nonneg; omega.
            }
            assert (H28: Z.of_nat nt / k * k <= Z.of_nat nt).
            {
              apply divide_mul_le; omega.
            }
            assert (H25: 0 <= ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k)).
            {
              apply TAC_pool_at_t_nonneg; try omega.              
            }
            assert (H26: 0 <= Z.of_nat nt mod k < k - 1).
            {
              apply mod_increase_bound; try omega.
            }
            assert (H27: ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) mod k = 0).
            {
              apply HTACmod; auto; try omega.
              - apply Z_mod_mult.
            }

            replace (ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) + Z.of_nat nt mod k + 1) with (ZMap.get i (TAC_pool_at_t (Z.of_nat nt) (Z.to_nat N) conf2) (Z.of_nat nt / k * k) + (Z.of_nat nt mod k + 1)) by omega. 
            rewrite divide_plus_equal; try omega.
            rewrite divide_plus_equal; try omega.
          }
        }
      }
    }
  Qed.
        
        
    
End SHRINK.


Section MAIN.

  Lemma shrink_preserves_schedulability:
    forall N k conf1 conf2,
      1 < k ->      
      conf_k_times N k conf1 conf2 ->
      config_pool_valid N conf1 ->
      config_pool_valid N conf2 ->
      system_schedulable N conf2 ->
      system_schedulable N conf1.
  Proof.
    intros * Hk Hconfk Hconf1 Hconf2 Hsched2.
    unfold system_schedulable.
    intros * Ht.
    unfold schedulable_upto.
    intros * Hp.
    destruct (ZMap.get p conf1) as [T C] eqn:Hconf1p.
    unfold schedulable_per_level_upto.
    intros * Hk1 Hk2.
    
    assert (HTC: 1 < T /\ 0 < C < T).
    {
      specialize (Hconf1 _ _ _ Hp Hconf1p).
      omega.
    }

    assert (Hstart: 0 <= k0 * T). nia.   

    assert (Hvtk1: TAC_k_times N k (TAC_pool_at_t ((k0*T + T) *k / k) (Z.to_nat N) conf1)
                               (TAC_pool_at_t ((k0 * T + T) * k) (Z.to_nat N) conf2)).
    {
      apply TAC_k_times_hold; auto.
      nia.
    }

    unfold TAC_k_times in Hvtk1.
    rewrite Z_div_mult_full in Hvtk1; try omega.

    assert (Hvt1: ZMap.get p (TAC_pool_at_t t (Z.to_nat N) conf1) (k0 * T + T) =
                  ZMap.get p (TAC_pool_at_t (k0 * T + T) (Z.to_nat N) conf1) ((k0 * T + T) * k / k)).
    {
      rewrite Z_div_mult_full; try omega.
      rewrite vertical_map_canonical; try omega.
    }
    rewrite Hvt1.
    rewrite Hvtk1; try nia.

    assert (Hvt2: ZMap.get p (TAC_pool_at_t t (Z.to_nat N) conf1) (k0 * T) =
                  ZMap.get p (TAC_pool_at_t (k0 * T + T) (Z.to_nat N) conf1) ((k0 * T) * k / k)).
    {
      rewrite Z_div_mult_full; try omega.
      rewrite vertical_map_canonical; try omega.
      symmetry.
      rewrite vertical_map_canonical; try omega.      
    }
    rewrite Hvt2.
    rewrite Hvtk1; try nia.

    assert (Hvtbound: k * C <= ZMap.get p (TAC_pool_at_t ((k0 * T + T) * k) (Z.to_nat N) conf2) ((k0 * T + T) * k) -
                      ZMap.get p (TAC_pool_at_t ((k0 * T + T) * k) (Z.to_nat N) conf2) (k0 * T * k)).
    {
      unfold conf_k_times in Hconfk.
      specialize (Hconfk _ _ _ Hp Hconf1p).
      unfold system_schedulable in Hsched2.
      assert (Ht': 0 <= ((k0 * T + T) * k)). nia.
      specialize (Hsched2 _ Ht').
      unfold schedulable_upto in Hsched2.
      specialize (Hsched2 _ Hp).
      rewrite Hconfk in Hsched2.
      unfold schedulable_per_level_upto in Hsched2.
      replace (k0 * T * k) with (k0 * (k * T)) by lia.
      replace ((k0 * T + T) * k) with (k0 * (k * T) + k * T) at 2 by lia.
      apply Hsched2; try nia.
    }

    assert (Hnonneg: 0 <= ZMap.get p (TAC_pool_at_t ((k0 * T + T) * k) (Z.to_nat N) conf2) (k0 * T * k)).
    {
      apply TAC_pool_at_t_nonneg; try nia.
    }
    
    assert (Hmod: (ZMap.get p (TAC_pool_at_t ((k0 * T + T) * k) (Z.to_nat N) conf2) (k0 * T * k)) mod k = 0).
    {
      assert (H10: 0 <= (k0 * T + T) * k). nia.
      assert (H11: conf_k_divisor N k conf2).
      {
        eapply conf_k_times_implies_divisor; eauto.
      }
      pose proof (conf_k_divisor_then_vt_valid_bigstep _ _ _ _ _ H10 Hk Hp H11 Hconf2).
      destruct H as [? HTACmod].
      apply HTACmod; auto; try omega.
      - nia.
      - apply Z_mod_mult.
    }
    
    rewrite div_sub_mod_0; try nia.
    transitivity (C * k / k).
    - rewrite Z_div_mult_full; omega.
    - apply Z_div_le; try nia.
  Qed.
    
End MAIN.


