Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import LemmaVTMap.
Require Import LemmaInterference.


Open Scope Z_scope.

Section IDLETIME.

  Fixpoint last_idle_slot_aux N idle ts nrange conf :=
    match nrange with
    | O => None
    | S n =>
      let i := highest_not_finished_map idle (ts + Z.of_nat n) (TAC_pool_at_t (ts + Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (ts + Z.of_nat n) conf) in
      if (zeq idle i) then Some (ts + Z.of_nat n) else
        last_idle_slot_aux N idle ts n conf
    end.

  Function last_idle_slot N idle ts te conf :=
    last_idle_slot_aux N idle ts (Z.to_nat (te - ts)) conf.
    
  Lemma last_idle_slot_aux_range:
    forall N idle ts nrange conf v,
      last_idle_slot_aux N idle ts nrange conf = Some v ->
      ts <= v < ts + Z.of_nat nrange.
  Proof.
    intros * H.
    induction nrange.
    {
      simpl in H.
      congruence.
    }
    {
      simpl in H.      
      destruct (zeq idle
         (highest_not_finished_map idle (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf))).
      {
        inversion H.
        rewrite Nat2Z.inj_succ.
        omega.
      }
      {
        specialize (IHnrange H).
        rewrite Nat2Z.inj_succ.
        omega.
      }
    }
  Qed.

  Lemma last_idle_slot_range:
    forall N idle ts te conf v,
      ts <= te ->
      last_idle_slot N idle ts te conf = Some v ->
      ts <= v < te.
  Proof.
    intros * Ht Hv.
    unfold last_idle_slot in Hv.
    pose proof (last_idle_slot_aux_range N idle ts (Z.to_nat (te - ts)) conf v) as H.
    rewrite Z2Nat.id in H; try omega.
    specialize (H Hv).
    omega.
  Qed.

  Lemma busy_schedule_prev_also_busy:
    forall N ts t conf,
      0 <= t ->
      last_idle_slot N N ts (ts + t + 1) conf = None ->
      last_idle_slot N N ts (ts + t) conf = None.
  Proof.
    intros * Ht Hnone.
    unfold last_idle_slot in *.
    replace (ts + t -ts) with t by omega.
    replace (ts + t + 1 - ts) with (t +1) in Hnone by omega.
    remember (Z.to_nat t) as nt.
    assert (H0: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    assert (H1: t + 1 = Z.of_nat (S nt)).
    {
      rewrite Nat2Z.inj_succ; omega.
    }
    rewrite H1 in Hnone.
    rewrite Nat2Z.id in Hnone.
    simpl in Hnone.
    destruct (zeq N
                (highest_not_finished_map N (ts + Z.of_nat nt)
                   (TAC_pool_at_t (ts + Z.of_nat nt) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (ts + Z.of_nat nt) conf))); auto.
    congruence.
  Qed.

  Lemma busy_schedule_every_tick_is_real:
    forall N ts te conf,
      0 < N ->
      ts <= te ->
      last_idle_slot N N ts te conf = None ->
      (forall t, ts <= t < te ->
            0 <= highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf (real_priqueue (Z.to_nat N) t conf) < N).
  Proof.
    intros * HN Hts Hidle t Ht.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte in *.
    clear Hte Heqnrange Hts.
    unfold last_idle_slot in Hidle.
    assert (H1: (Z.to_nat (ts + Z.of_nat nrange - ts)) = nrange).
    {
      replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
      rewrite Nat2Z.id; auto.
    }
    rewrite H1 in Hidle. clear H1.
    induction nrange.
    {
      simpl in Ht.
      omega.
    }
    {
      simpl in Hidle.
      assert (Hrealnow: highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                                 (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <> N).
      {
        destruct (zeq N
                (highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf))); auto.
        congruence.
      }      
      rewrite zeq_false in Hidle; auto.
      destruct (zeq t (ts + Z.of_nat nrange)) as [Heq | Hneq].
      {
        assert (0 <= highest_not_finished_map N (ts + Z.of_nat nrange) (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                             (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <= N).
        {
          apply highest_not_finished_map_res_bound; try omega.
          - replace N with (Z.of_nat (Z.to_nat N)) at 1.
            apply real_priqueue_sound.
            rewrite Z2Nat.id; omega.
        }
        rewrite Heq in *.
        omega.
      }
      {
        apply IHnrange; auto.
        rewrite Nat2Z.inj_succ in Ht.
        omega.
      }
    }
  Qed.
      
End IDLETIME.


Section OVERLAPPING.

  Function is_overlapping fromid Tf toid d1 d2 d3 : bool :=
    if (zle d1 (d2 / Tf * Tf)) then
      if (zlt (d2 / Tf * Tf) d2) then
        if (priEntry_lt (PriEntry (d2 / Tf * Tf + Tf) fromid)
                        (PriEntry d3 toid)) then
        true else false else false else false.

  Fixpoint workload_overlapping_total (Nnat: nat) (toid d1 d2 d3: Z) conf :Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_overlapping_total n toid d1 d2 d3 conf else
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid Tn Cn =>
          if (is_overlapping (Z.of_nat n) Tn toid d1 d2 d3) then
            Cn + workload_overlapping_total n toid d1 d2 d3 conf else
            workload_overlapping_total n toid d1 d2 d3 conf
        end
    end.

  Fixpoint workload_overlapping_remain_total (Nnat: nat) (toid d1 d2 d3 t: Z) conf TAC :Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_overlapping_remain_total n toid d1 d2 d3 t conf TAC else
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid Tf Cf =>          
          if (is_overlapping (Z.of_nat n) Tf toid d1 d2 d3) then           
            workload_overlapping_remain_total n toid d1 d2 d3 t conf TAC +
            Z.min Cf (ZMap.get (Z.of_nat n) TAC t - ZMap.get (Z.of_nat n) TAC (t / Tf * Tf))
          else workload_overlapping_remain_total n toid d1 d2 d3 t conf TAC
        end
    end.

  Function is_leftover fromid Tf toid d1 d2 : bool :=
    if (zle d1 (d2 - Tf)) then
      if (zeq (d2 mod Tf) 0) then
        if (zlt toid fromid) then true else false
                                              else false else false.

  Fixpoint workload_leftover_total (Nnat: nat) (toid d1 d2: Z) conf :Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_leftover_total n toid d1 d2 conf else
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid Tn Cn =>
          if (is_leftover (Z.of_nat n) Tn toid d1 d2) then
            Cn + workload_leftover_total n toid d1 d2 conf else
            workload_leftover_total n toid d1 d2 conf
        end
    end.

  Lemma workload_overlapping_remain_total_canonical_aux:
    forall Nnat Nnat0 toid d1 d2 d3 t t2 conf,
      Z.of_nat Nnat <= Z.of_nat Nnat0 ->
      0 <= t <= t2 ->
      config_pool_valid (Z.of_nat Nnat0) conf ->
      workload_overlapping_remain_total Nnat toid d1 d2 d3 t conf (TAC_pool_at_t t2 Nnat0 conf) =
      workload_overlapping_remain_total Nnat toid d1 d2 d3 t conf (TAC_pool_at_t t Nnat0 conf).
  Proof.
    intros * HN Ht Hconfvalid.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nnat <= Z.of_nat Nnat0). xomega.      
      rewrite (IHNnat H1).
      destruct (zeq toid (Z.of_nat Nnat)); auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:Hconf.
      assert (HT: 1 < T).
      {
        assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat Nnat0). xomega.
        specialize (Hconfvalid _ _ _ H2 Hconf).
        omega.
      }
      destruct (is_overlapping (Z.of_nat Nnat) T toid d1 d2 d3); auto.

      assert (H: forall e a b c d, a = c -> b = d -> e + Z.min C (a - b) = e + Z.min C (c - d)).
      {
        intros.
        subst.
        auto.
      }
      apply H.
      - rewrite vertical_map_canonical; try omega.
        auto.
      - assert (H3: 0 <= t / T * T).
        {
          apply div_mul_nonneg; omega.
        }
        assert (H4: t / T * T <= t).
        {
          apply Nfloor_le; omega.
        }
        rewrite vertical_map_canonical; try omega.
        symmetry.
        rewrite vertical_map_canonical; try omega.
    }
  Qed.
    
  Lemma workload_overlapping_remain_total_canonical:
    forall Nnat toid d1 d2 d3 t t2 conf,
      0 <= t <= t2 ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_overlapping_remain_total Nnat toid d1 d2 d3 t conf (TAC_pool_at_t t2 Nnat conf) =
      workload_overlapping_remain_total Nnat toid d1 d2 d3 t conf (TAC_pool_at_t t Nnat conf).
  Proof.
    intros.
    apply workload_overlapping_remain_total_canonical_aux; auto.
    omega.
  Qed.
  
  Lemma is_overlapping_true_end_not_aligned:
    forall N Tf id d1 d2 d3,
      is_overlapping N Tf id d1 d2 d3 = true ->
      d2 / Tf * Tf <> d2.
  Proof.
    intros.
    unfold is_overlapping in H.
    unfold not.
    intros Heq.
    rewrite Heq in H.
    destruct (zle d1 d2); try omega; try congruence.
    destruct (zlt d2 d2); try omega; try congruence.
  Qed.

  Lemma is_overlapping_false_empty_slot2:
    forall N Tf id d1 d2,
      0 < Tf ->
      0 <= d2 ->
      is_overlapping N Tf id d1 d2 d2 = false.
  Proof.
    intros.
    unfold is_overlapping.
    destruct (zle d1 (d2 / Tf * Tf)); auto.
    destruct (zlt (d2 / Tf * Tf) d2); auto.
    destruct (priEntry_lt (PriEntry (d2 / Tf * Tf + Tf) N) (PriEntry d2 id)) eqn:Hpri; auto.
    pose proof (priEntry_lt_deadline_le _ _ _ _ Hpri).
    assert (d2 < d2 / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; try omega.
    }
    omega.
  Qed.

End OVERLAPPING.


Section OVERLAPPING_ONLY_DEPEND_ON_CONF.

  Definition id_is_enlarged_by_k id k conf1 conf2 :=
    forall i T C,
      ZMap.get id conf1 = mkPrioConfigValid T C ->      
      ZMap.get i conf2 = ZMap.get i (ZMap.set id (mkPrioConfigValid (k*T) (k*C)) conf1).
  
  Lemma workload_overlapping_total_same:    
    forall Nnat conf conf2 id k d1 d2 d3,      
      id_is_enlarged_by_k id k conf conf2 ->
      workload_overlapping_total Nnat id d1 d2 d3 conf2 = workload_overlapping_total Nnat id d1 d2 d3 conf.
  Proof.
    intros * Henlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.
      destruct (ZMap.get id conf) as [T0 C0] eqn:Hconfid.
      specialize (Henlarge (Z.of_nat Nnat) _ _ Hconfid).
      rewrite ZMap.gso in Henlarge; auto.
      rewrite Henlarge.
      rewrite Hconfn.
      rewrite <- IHNnat.
      auto.
    }
  Qed.

  Lemma workload_leftover_total_same:    
    forall Nnat conf conf2 id k d1 d2,      
      id_is_enlarged_by_k id k conf conf2 ->
      workload_leftover_total Nnat id d1 d2 conf2 = workload_leftover_total Nnat id d1 d2 conf.
  Proof.
    intros * Henlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.
      destruct (ZMap.get id conf) as [T0 C0] eqn:Hconfid.
      specialize (Henlarge (Z.of_nat Nnat) _ _ Hconfid).
      rewrite ZMap.gso in Henlarge; auto.
      rewrite Henlarge.
      rewrite Hconfn.
      rewrite <- IHNnat.
      auto.
    }
  Qed.  
      
End OVERLAPPING_ONLY_DEPEND_ON_CONF.


Section SCHED_TYPE.

  Inductive sched_type: Type :=
  | Sched_self                     (* the task "id" *)
  | Sched_interference             (* tasks that must be finished before "id" *)
  | Sched_leftover                 (* tasks lower than "id" in the current period but higher in the next *)
  | Sched_overlap                  (* tasks that contributes to workload_remain in the next period *).

  Lemma sched_type_eq_dec:
    forall x y : sched_type, {x = y} + {x <> y}.
  Proof.
    intros.
    destruct x; destruct y; auto; (right; congruence).
  Qed.

  Lemma sched_type_eq_true:
    forall (A: Type) (ty: sched_type) (a b: A),
      (if sched_type_eq_dec ty ty then a else b) = a.
  Proof.
    intros.
    case (sched_type_eq_dec ty ty); intros.
    auto.
    elim n; auto.
  Qed.

  Lemma sched_type_eq_false:
    forall (A: Type) (x y: sched_type) (a b: A),
      x <> y ->
      (if sched_type_eq_dec x y then a else b) = b.
  Proof.
    intros.
    case (sched_type_eq_dec x y); intros.
    elim H; auto.
    auto.
  Qed.
                           
  (* Given that ts <= t < te, and 0 <= fromid, toid < N *)
  Function get_sched_type (fromid T toid t todeadline: Z) :sched_type :=
    if (zeq fromid toid) then Sched_self else
      let deadline := t / T * T + T in
      if (priEntry_lt (PriEntry deadline fromid) (PriEntry todeadline toid)) then Sched_interference else
        if (zeq deadline todeadline) then Sched_leftover else
          Sched_overlap.

  Lemma get_sched_type_is_self:
    forall fromid T toid t todeadline,
      get_sched_type fromid T toid t todeadline = Sched_self ->
      fromid = toid.
  Proof.
    intros.
    unfold get_sched_type in H.
    destruct (zeq fromid toid); auto.
    destruct (priEntry_lt (PriEntry (t / T * T + T) fromid) (PriEntry todeadline toid)); try congruence.
    destruct (zeq (t / T * T + T) todeadline); try congruence.
  Qed.

  Fixpoint count_type_occurrence_aux N idle ts nrange conf toid todeadline ty :=
    match nrange with
    | O => 0
    | S n =>
      let i := highest_not_finished_map idle (ts + Z.of_nat n)
                                        (TAC_pool_at_t (ts + Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (ts + Z.of_nat n) conf) in
      if (zeq i idle) then count_type_occurrence_aux N idle ts n conf toid todeadline ty else
        match (ZMap.get i conf) with
        | mkPrioConfigValid T C =>
          let cur_ty := get_sched_type i T toid (ts + Z.of_nat n) todeadline in
          if (sched_type_eq_dec cur_ty ty) then
            count_type_occurrence_aux N idle ts n conf toid todeadline ty + 1 else
            count_type_occurrence_aux N idle ts n conf toid todeadline ty
        end
    end.

  Function count_type_occurrence N idle ts te conf toid todeadline ty :=
    count_type_occurrence_aux N idle ts (Z.to_nat (te - ts)) conf toid todeadline ty.

  Fixpoint count_single_task_type_aux N idle fromid ts nrange conf toid todeadline ty: Z :=
    match nrange with
    | O => 0
    | S n =>
      let i := highest_not_finished_map idle (ts + Z.of_nat n)
                                        (TAC_pool_at_t (ts + Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (ts + Z.of_nat n) conf) in
      if (zeq i fromid) then 
        match (ZMap.get i conf) with
        | mkPrioConfigValid T C =>
          let cur_ty := get_sched_type i T toid (ts + Z.of_nat n) todeadline in
          if (sched_type_eq_dec cur_ty ty) then
            count_single_task_type_aux N idle fromid ts n conf toid todeadline ty + 1 else
            count_single_task_type_aux N idle fromid ts n conf toid todeadline ty
        end
          else count_single_task_type_aux N idle fromid ts n conf toid todeadline ty
    end.

  Function count_single_task_type_occurrence N idle fromid ts te conf toid todeadline ty: Z :=
    count_single_task_type_aux N idle fromid ts (Z.to_nat (te - ts)) conf toid todeadline ty.    

  Fixpoint count_type_occurrence2_aux (Nid: nat) (N idle ts te: Z) conf (toid todeadline: Z) ty: Z :=
    match Nid with
    | O => 0
    | S n => count_type_occurrence2_aux n N idle ts te conf toid todeadline ty +
            count_single_task_type_occurrence N N (Z.of_nat n) ts te conf toid todeadline ty
    end.  
  
  Lemma count_single_task_type_occurance_0_when_empty_duration:
    forall N1 N2 N3 t conf toid todeadline ty,
      count_single_task_type_occurrence N1 N2 N3 t t conf toid todeadline ty = 0.
  Proof.
    unfold count_single_task_type_occurrence.
    intros.
    replace (t - t) with 0 by omega.
    simpl.
    auto.
  Qed.
                                                                                   
  Lemma count_type_occurance2_aux_0_when_empty_duration:
    forall Nid N1 N2 t conf toid todeadline ty,
      count_type_occurrence2_aux Nid N1 N2 t t conf toid todeadline ty = 0.
  Proof.
    intros.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHNid.
      rewrite count_single_task_type_occurance_0_when_empty_duration; auto.
    }
  Qed.

  Lemma count_single_task_type_occurrence_stagnates_when_not_scheduled:
    forall N fromid ts te conf toid todeadline ty,
      ts <= te ->
      highest_not_finished_map N te (TAC_pool_at_t te (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) te conf) <> fromid ->
      count_single_task_type_occurrence N N fromid ts (te + 1) conf toid todeadline ty =      
      count_single_task_type_occurrence N N fromid ts te conf toid todeadline ty.
  Proof.
    intros * Hte Hid Hidle.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te - ts)) as range.
    assert (H1: te = ts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (te + 1 - ts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (te + 1 - ts) with (te - ts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite zeq_false; auto.
  Qed.

  Lemma count_single_task_type_occurrence_incr_when_scheduled_and_match:
    forall N ts te conf toid todeadline ty i T C,
      ts <= te ->
      i = highest_not_finished_map N te (TAC_pool_at_t te (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) te conf) ->
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_sched_type i T toid te todeadline = ty ->      
      count_single_task_type_occurrence N N i ts (te + 1) conf toid todeadline ty =      
      count_single_task_type_occurrence N N i ts te conf toid todeadline ty + 1.
  Proof.
    intros * Hts Hhigh Hi Hconf Htype.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te - ts)) as range.
    assert (H1: te = ts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (te + 1 - ts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (te + 1 - ts) with (te - ts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite <- Hhigh.
    rewrite zeq_true.
    rewrite Hconf.
    rewrite Htype.
    rewrite sched_type_eq_true; auto.
  Qed.

  Lemma count_single_task_type_occurrence_stagnates_when_scheduled_but_mismatch:
    forall N ts te conf toid todeadline ty i T C,
      ts <= te ->
      i = highest_not_finished_map N te (TAC_pool_at_t te (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) te conf) ->
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_sched_type i T toid te todeadline <> ty ->      
      count_single_task_type_occurrence N N i ts (te + 1) conf toid todeadline ty =      
      count_single_task_type_occurrence N N i ts te conf toid todeadline ty.
  Proof.
    intros * Hts Hhigh Hi Hconf Htype.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te - ts)) as range.
    assert (H1: te = ts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (te + 1 - ts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (te + 1 - ts) with (te - ts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite <- Hhigh.
    rewrite zeq_true.
    rewrite Hconf.
    rewrite sched_type_eq_false; auto.
  Qed.

  Definition count_single_slot N idle fromid ts conf toid todeadline ty :Z :=
      (if zeq
         (highest_not_finished_map idle ts (TAC_pool_at_t ts (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) ts conf)) fromid
    then
     match
       ZMap.get
         (highest_not_finished_map idle ts (TAC_pool_at_t ts (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) ts conf)) conf
     with
     | mkPrioConfigValid T _ =>
         if sched_type_eq_dec
              (get_sched_type
                 (highest_not_finished_map idle ts 
                    (TAC_pool_at_t ts (Z.to_nat N) conf) conf 
                    (real_priqueue (Z.to_nat N) ts conf)) T toid ts todeadline) ty
         then 1
         else 0
     end
       else 0).
    
  Lemma count_single_task_type_occurrence_aux_left_step:
    forall N idle fromid ts range conf toid todeadline ty,
      0 < Z.of_nat range ->
      count_single_task_type_aux N idle fromid ts range conf toid todeadline ty =
      count_single_task_type_aux N idle fromid (ts + 1) (pred range) conf toid todeadline ty +
      count_single_slot N idle fromid ts conf toid todeadline ty.
  Proof.
    intros * Hrange.
    revert ts.
    induction range.
    {
      simpl in Hrange.      
      intros.
      omega.
    }
    {
      intros.
      destruct (zeq (Z.of_nat range) 0) as [Heq | Hneq].
      {
        assert (H2: Z.to_nat (Z.of_nat range) = Z.to_nat 0).
        { rewrite Heq. auto. }
        simpl in H2.
        rewrite Nat2Z.id in H2.
        
        rewrite H2 in *.
        simpl.
        replace (ts + 0) with ts by omega.
        auto.
      }
      {
        simpl.
        destruct (zeq
         (highest_not_finished_map idle (ts + Z.of_nat range)
            (TAC_pool_at_t (ts + Z.of_nat range) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat range) conf)) fromid) as [Hsched|Hother].
        {
          destruct (ZMap.get
       (highest_not_finished_map idle (ts + Z.of_nat range)
          (TAC_pool_at_t (ts + Z.of_nat range) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (ts + Z.of_nat range) conf)) conf) as [T C] eqn:Hconf.
          destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map idle (ts + Z.of_nat range)
               (TAC_pool_at_t (ts + Z.of_nat range) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (ts + Z.of_nat range) conf)) T toid 
            (ts + Z.of_nat range) todeadline) ty) as [Hmatch | Hmismatch].
          {
            rewrite IHrange; try xomega.
            assert (forall a b c, a + 1 = b -> a + c + 1 = b + c). intros. omega.
            apply H.
            remember (pred range) as n.
            assert (Hr: range = S n).
            {
              rewrite Heqn.
              rewrite NPeano.Nat.succ_pred; auto.
              omega.
            }
            rewrite Hr.
            simpl.

            replace (ts + 1 + Z.of_nat n) with (ts + Z.of_nat range).
            rewrite Hsched.
            rewrite zeq_true.
            rewrite Hsched in *.
            rewrite Hconf.            
            rewrite Hmatch.
            rewrite sched_type_eq_true.
            auto.
            - xomega.
          }
          {
            rewrite IHrange; try xomega.
            assert (forall a b c, a = b -> a + c = b + c). intros. omega.
            apply H.
            remember (pred range) as n.
            assert (Hr: range = S n).
            {
              rewrite Heqn.
              rewrite NPeano.Nat.succ_pred; auto.
              omega.
            }
            rewrite Hr.
            simpl.

            replace (ts + 1 + Z.of_nat n) with (ts + Z.of_nat range).
            rewrite Hsched.
            rewrite zeq_true.
            rewrite Hsched in *.
            rewrite Hconf.            
            rewrite sched_type_eq_false; auto.
            - xomega.
          }
        }
        {
          rewrite IHrange; try xomega.
          assert (forall a b c, a = b -> a + c = b + c). intros. omega.
          apply H.
          remember (pred range) as n.
          assert (Hr: range = S n).
          {
            rewrite Heqn.
            rewrite NPeano.Nat.succ_pred; auto.
            omega.
          }
          rewrite Hr.
          simpl.

          replace (ts + 1 + Z.of_nat n) with (ts + Z.of_nat range).          
          rewrite zeq_false; auto.
          - xomega.
        }
      }
    }
  Qed.

  Lemma count_single_task_type_occurrence_nonneg:
    forall N idle fromid t1 t2 conf toid todeadline ty,
      0 <= count_single_task_type_occurrence N idle fromid t1 t2 conf toid todeadline ty.
  Proof.
    intros *.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (t2 - t1)) as range.
    clear Heqrange.
    induction range.
    {
      simpl.
      omega.
    }
    {
      simpl.
      destruct (zeq
         (highest_not_finished_map idle (t1 + Z.of_nat range)
            (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) fromid); try omega.
      destruct (ZMap.get
       (highest_not_finished_map idle (t1 + Z.of_nat range)
          (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) conf).
      destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map idle (t1 + Z.of_nat range)
               (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) z toid
            (t1 + Z.of_nat range) todeadline) ty); try omega.
    }
  Qed.
  
  Lemma count_single_task_type_occurrence_composition:
    forall N idle fromid t1 t2 t3 conf toid todeadline ty,
      t1 <= t2 <= t3 ->
      count_single_task_type_occurrence N idle fromid t1 t2 conf toid todeadline ty +
      count_single_task_type_occurrence N idle fromid t2 t3 conf toid todeadline ty =
      count_single_task_type_occurrence N idle fromid t1 t3 conf toid todeadline ty.      
  Proof.
    intros * Ht.
    remember (Z.to_nat (t2 -t1)) as range.
    assert (H1: t2 = t1 + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite H1 in *. clear H1 Heqrange.

    induction range.
    {
      simpl.
      replace (t1 + 0) with t1 by omega.
      rewrite count_single_task_type_occurance_0_when_empty_duration.
      omega.
    }
    {
      rewrite <- IHrange; try xomega. clear IHrange.
      unfold count_single_task_type_occurrence.      
      
      replace (t1 + Z.of_nat (S range) - t1) with (Z.of_nat (S range)) by omega.
      replace (t1 + Z.of_nat range - t1) with (Z.of_nat range) by omega.
      repeat rewrite Nat2Z.id.        

      assert (H1: t1 + Z.of_nat (S range) = t1 + Z.of_nat range + 1). xomega.
      rewrite H1. clear H1.

      assert (H2: Z.to_nat (t3 - (t1 + Z.of_nat range + 1)) =
                  pred (Z.to_nat (t3 - (t1 + Z.of_nat range)))).
      {
        rewrite <- Z2Nat.inj_pred.
        rewrite <- Z.sub_1_r.
        replace (t3 - (t1 + Z.of_nat range + 1)) with (t3 - (t1 + Z.of_nat range) - 1) by omega.
        auto.
      }
      rewrite H2.
      
      assert (H3: count_single_task_type_aux N idle fromid (t1 + Z.of_nat range) 
                                             (Z.to_nat (t3 - (t1 + Z.of_nat range))) conf toid todeadline ty =
             count_single_task_type_aux N idle fromid (t1 + Z.of_nat range + 1) 
                                        (pred (Z.to_nat (t3 - (t1 + Z.of_nat range)))) conf toid todeadline ty +
             count_single_slot N idle fromid (t1 + Z.of_nat range) conf toid todeadline ty).
      {
        apply count_single_task_type_occurrence_aux_left_step; auto.
        - rewrite Z2Nat.id; xomega.
      }
      rewrite H3. clear H3.

      assert (H: forall a b c d, a = b + c -> a + d = b + (d + c)). intros. omega.
      apply H.
      simpl.
      unfold count_single_slot.

      destruct (zeq
         (highest_not_finished_map idle (t1 + Z.of_nat range)
            (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) fromid) as [Heq|Hneq]; try omega.
      
      destruct (ZMap.get
       (highest_not_finished_map idle (t1 + Z.of_nat range)
          (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) conf).
      destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map idle (t1 + Z.of_nat range)
               (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) z toid 
            (t1 + Z.of_nat range) todeadline) ty); omega.
    }
  Qed.  

  Lemma count_type_occurance2_aux_composition:
    forall Nid N t1 t2 t3 conf toid todeadline ty,
      t1 <= t2 <= t3 ->
      Z.of_nat Nid <= N ->
      count_type_occurrence2_aux Nid N N t1 t2 conf toid todeadline ty +
      count_type_occurrence2_aux Nid N N t2 t3 conf toid todeadline ty =
      count_type_occurrence2_aux Nid N N t1 t3 conf toid todeadline ty.
  Proof.
    intros * Ht HNid.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite <- IHNid; try xomega.
      assert (count_single_task_type_occurrence N N (Z.of_nat Nid) t1 t2 conf toid todeadline ty + count_single_task_type_occurrence N N (Z.of_nat Nid) t2 t3 conf toid todeadline ty =
              count_single_task_type_occurrence N N (Z.of_nat Nid) t1 t3 conf toid todeadline ty).
      {
        apply count_single_task_type_occurrence_composition; auto.
      }
      omega.
    }
  Qed.
  
  Lemma count_type_occurance2_aux_stagnates_when_slot_idle:
    forall Nid N ts te conf toid todeadline ty,
      ts <= te ->
      Z.of_nat Nid <= N ->
      highest_not_finished_map N te 
            (TAC_pool_at_t te (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) te conf) = N ->
      count_type_occurrence2_aux Nid N N ts (te + 1) conf toid todeadline ty =      
      count_type_occurrence2_aux Nid N N ts te conf toid todeadline ty.
  Proof.
    intros * Hte HN Hidle.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nid < Z.of_nat (S Nid)). xomega.
      assert (HN': Z.of_nat Nid <= N). xomega.
      rewrite <- (IHNid HN').

      rewrite count_single_task_type_occurrence_stagnates_when_not_scheduled; auto; try omega.
    }
  Qed.
  
  Lemma count_type_occurance2_aux_stagnates_when_not_scheduled:
    forall Nid N ts te conf toid todeadline ty,
      ts <= te ->
      Z.of_nat Nid <= highest_not_finished_map N te 
            (TAC_pool_at_t te (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) te conf) ->
      count_type_occurrence2_aux Nid N N ts (te + 1) conf toid todeadline ty =      
      count_type_occurrence2_aux Nid N N ts te conf toid todeadline ty.
  Proof.
    intros * Hte HN Hidle.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nid < Z.of_nat (S Nid)). xomega.
      rewrite <- IHNid; try omega.

      rewrite count_single_task_type_occurrence_stagnates_when_not_scheduled; auto; try omega.
    }
  Qed.
  
  Lemma count_type_occurance2_aux_increase_when_scheduled_task_match:
    forall Nid N ts te conf toid todeadline ty i T C,
      ts <= te ->
      i < Z.of_nat Nid <= N ->     
      i = highest_not_finished_map N te 
            (TAC_pool_at_t te (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) te conf) ->
      0 <= i <= N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_sched_type i T toid te todeadline = ty ->
      count_type_occurrence2_aux Nid N N ts (te + 1) conf toid todeadline ty =      
      count_type_occurrence2_aux Nid N N ts te conf toid todeadline ty + 1.
  Proof.
    intros * Hte HN Hsched Hi Hconfi Hmatch.
    induction Nid.
    {
      simpl in HN.
      omega.
    }
    {
      simpl.
      destruct (zeq (Z.of_nat Nid) i) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite count_type_occurance2_aux_stagnates_when_not_scheduled; auto; try omega.
        rewrite count_single_task_type_occurrence_incr_when_scheduled_and_match with (T:=T) (C:=C); auto; try omega.
      }
      {
        rewrite count_single_task_type_occurrence_stagnates_when_not_scheduled; auto; try omega.
        rewrite IHNid; auto; try xomega.
      }
    }
  Qed.

  Lemma count_type_occurance2_aux_increase_when_scheduled_task_mismatch:
    forall Nid N ts te conf toid todeadline ty i T C,
      ts <= te ->
      i < Z.of_nat Nid <= N ->     
      i = highest_not_finished_map N te 
            (TAC_pool_at_t te (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) te conf) ->
      0 <= i <= N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_sched_type i T toid te todeadline <> ty ->
      count_type_occurrence2_aux Nid N N ts (te + 1) conf toid todeadline ty =      
      count_type_occurrence2_aux Nid N N ts te conf toid todeadline ty.
  Proof.
    intros * Hte HN Hsched Hi Hconfi Hmatch.
    induction Nid.
    {
      simpl in HN.
      omega.
    }
    {
      simpl.
      destruct (zeq (Z.of_nat Nid) i) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite count_type_occurance2_aux_stagnates_when_not_scheduled; auto; try omega.
        rewrite count_single_task_type_occurrence_stagnates_when_scheduled_but_mismatch with (T:=T) (C:=C); auto; try omega.
      }
      {
        rewrite count_single_task_type_occurrence_stagnates_when_not_scheduled; auto; try omega.
        rewrite IHNid; auto; try xomega.
      }
    }
  Qed.
  
  Lemma count_type_occurrence2_aux_eq:
    forall N ts nrange conf toid todeadline ty,
      0 < N ->
      count_type_occurrence_aux N N ts nrange conf toid todeadline ty =
      count_type_occurrence2_aux (Z.to_nat N) N N ts (ts + Z.of_nat nrange) conf toid todeadline ty.
  Proof.
    intros * HN.
    induction nrange.
    {
      simpl.
      replace (ts + 0) with ts by omega.
      rewrite count_type_occurance2_aux_0_when_empty_duration.
      auto.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = ts + Z.of_nat nrange + 1). xomega.
      assert (H2: Z.of_nat (Z.to_nat N) = N). { rewrite Z2Nat.id; omega. }
      rewrite H1. clear H1.
      simpl.
      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N) as [Hidle | Hsched].
      {
        rewrite IHnrange.
        rewrite count_type_occurance2_aux_stagnates_when_slot_idle; auto; try omega.
      }
      {
        remember (highest_not_finished_map N (ts + Z.of_nat nrange)
             (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
             (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) as i.        
        assert (H3: 0 <= i <= N).
        {
          rewrite Heqi.
          apply highest_not_finished_map_res_bound; auto.
          - remember (Z.to_nat N) as nN.
            replace N with (Z.of_nat nN).
            apply real_priqueue_sound.
        }
        destruct (ZMap.get i conf) as [T C] eqn:Hconf.        
        destruct (sched_type_eq_dec (get_sched_type i T toid (ts + Z.of_nat nrange) todeadline) ty) as [Hmatch|Hmismatch].
        {
          rewrite IHnrange.
          rewrite count_type_occurance2_aux_increase_when_scheduled_task_match with (i:=i) (T:=T) (C:=C); auto; try xomega.
        }
        {
          rewrite IHnrange.
          rewrite count_type_occurance2_aux_increase_when_scheduled_task_mismatch with (i:=i) (T:=T) (C:=C); auto; try xomega.
        }
      }
    }
  Qed.
  
  Lemma count_type_occurrence2_eq:
    forall N ts te conf toid todeadline ty,
      0 < N ->
      ts <= te ->
      count_type_occurrence N N ts te conf toid todeadline ty =
      count_type_occurrence2_aux (Z.to_nat N) N N ts te conf toid todeadline ty.      
  Proof.
    intros * HN Ht.
    unfold count_type_occurrence.
    remember (Z.to_nat (te - ts)) as range.
    replace te with (ts + Z.of_nat range).
    - rewrite count_type_occurrence2_aux_eq; auto.
    - rewrite Heqrange.
      rewrite Z2Nat.id; omega.
  Qed.

  Lemma count_type_occurrence_composition:
    forall N t1 t2 t3 conf toid todeadline ty,
      0 < N ->
      t1 <= t2 <= t3 ->
      count_type_occurrence N N t1 t2 conf toid todeadline ty +
      count_type_occurrence N N t2 t3 conf toid todeadline ty =
      count_type_occurrence N N t1 t3 conf toid todeadline ty.      
  Proof.
    intros * HN Ht.
    repeat rewrite count_type_occurrence2_eq; try omega.
    apply count_type_occurance2_aux_composition; auto.
    rewrite Z2Nat.id; omega.
  Qed.

  Lemma count_type_occurrence_nonneg:
    forall N te ts conf toid todeadline ty,
      0 <= count_type_occurrence N N ts te conf toid todeadline ty.
  Proof.
    intros.
    unfold count_type_occurrence.
    destruct (zle ts te).
    {
      remember (Z.to_nat (te - ts)) as nrange.
      clear Heqnrange.
      induction nrange.
      {
        simpl.
        omega.
      }
      {
        simpl.
        destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N); auto.
        destruct (
     ZMap.get
       (highest_not_finished_map N (ts + Z.of_nat nrange)
          (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) conf).
        destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map N (ts + Z.of_nat nrange)
               (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) z toid
            (ts + Z.of_nat nrange) todeadline) ty); omega.
      }
    }
    {
      assert (te - ts < 0). omega.
      destruct (te - ts); try discriminate.
    }
  Qed.
      
  Lemma sched_type_sum_aux_inc_when_schedule_not_idle:
    forall N conf id todeadline ts nrange,
      0 <= highest_not_finished_map N (ts + Z.of_nat nrange)
                                  (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) < N ->
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_overlap =      
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_overlap + 1.
  Proof.
    intros * Hreal.
    simpl.
    repeat rewrite zeq_false; try omega.
    destruct (ZMap.get
       (highest_not_finished_map N (ts + Z.of_nat nrange)
          (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) conf) as [T0 C0] eqn:Hconf0.
    destruct ((get_sched_type
            (highest_not_finished_map N (ts + Z.of_nat nrange)
               (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) T0 id 
            (ts + Z.of_nat nrange) todeadline)) eqn:Hty.
    {
      (* Sched_self *)
      rewrite sched_type_eq_true.
      repeat rewrite sched_type_eq_false; try congruence.
      omega.
    }
    {
      (* Sched_interference *)
      rewrite sched_type_eq_true.
      repeat rewrite sched_type_eq_false; try congruence.
      omega.
    }
    {
      rewrite sched_type_eq_true.
      repeat rewrite sched_type_eq_false; try congruence.
      omega.
    }
    {
      rewrite sched_type_eq_true.
      repeat rewrite sched_type_eq_false; try congruence.
      omega.
    }
  Qed.

  Lemma sched_type_sum_aux_stagnates_when_schedule_idle:
    forall N conf id todeadline ts nrange,
      highest_not_finished_map N (ts + Z.of_nat nrange)
                                  (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) = N ->
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts (S nrange) conf id todeadline Sched_overlap =      
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_overlap.
  Proof.
    intros * Hreal.
    simpl.
    rewrite Hreal.
    repeat rewrite zeq_true.
    auto.
  Qed.
  
  Lemma sched_type_is_complete_when_schedule_is_busy_aux:
    forall N conf id T C ts nrange todeadline,
      0 <= ts ->
      1 < T ->
      0 < C ->
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->      
      last_idle_slot N N ts (ts + Z.of_nat nrange) conf = None ->
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_overlap = Z.of_nat nrange.
  Proof.
    intros * Hts HT HC Hid Hconfid Hidle.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      assert (H2: last_idle_slot N N ts (ts + Z.of_nat nrange) conf = None).
      {
        assert (H3: ts + Z.of_nat (S nrange) = ts + Z.of_nat nrange + 1).
        { rewrite Nat2Z.inj_succ; omega. }
        rewrite H3 in Hidle.
        apply busy_schedule_prev_also_busy; auto.
        xomega.
      }
      specialize (IHnrange H2). 
      rewrite sched_type_sum_aux_inc_when_schedule_not_idle.
      - rewrite Nat2Z.inj_succ; omega.
      - apply busy_schedule_every_tick_is_real with (ts := ts) (te := ts + Z.of_nat (S nrange)); auto; try omega.
      rewrite Nat2Z.inj_succ; omega.
    }
  Qed.
  
  Lemma sched_type_is_complete_when_schedule_is_busy:
    forall N conf id T C ts todeadline,
      0 <= ts ->
      ts mod T = 0 ->
      1 < T ->
      0 < C ->
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      last_idle_slot N N ts (ts + T) conf = None ->
      count_type_occurrence N N ts (ts + T) conf id todeadline Sched_self +
      count_type_occurrence N N ts (ts + T) conf id todeadline Sched_interference +
      count_type_occurrence N N ts (ts + T) conf id todeadline Sched_leftover +
      count_type_occurrence N N ts (ts + T) conf id todeadline Sched_overlap = T.
  Proof.
    intros * Hts Htsmod HT HC Hid Hconfid Hidle.
    unfold count_type_occurrence.
    remember (Z.to_nat T) as nrange.
    replace (ts + T - ts) with T by omega.
    assert (HTv: T = Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite HTv in *.
    clear Heqnrange HTv.
    rewrite Nat2Z.id.
    apply sched_type_is_complete_when_schedule_is_busy_aux with (T:=Z.of_nat nrange) (C := C); auto.
  Qed.

  Lemma sched_type_is_bounded_by_window_aux:
    forall N conf id ts nrange todeadline,
      0 <= ts ->
      0 <= id < N ->
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_self +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_interference +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_leftover +
      count_type_occurrence_aux N N ts nrange conf id todeadline Sched_overlap <= Z.of_nat nrange.
  Proof.
    intros * Hts Hid.
    induction nrange.
    {
      simpl.
      omega.
    }
    {
      destruct (zeq (highest_not_finished_map N (ts + Z.of_nat nrange)
                                  (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                  (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N) as [Heq | Hneq].
      {        
        rewrite sched_type_sum_aux_stagnates_when_schedule_idle; auto.
        xomega.
      }
      {
        assert (Hb: 0 <= highest_not_finished_map N (ts + Z.of_nat nrange)
           (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
           (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <= N).
        {
          apply highest_not_finished_map_res_bound; auto; try omega.
          - remember (Z.to_nat N) as Nnat.
            replace N with (Z.of_nat Nnat).
            apply real_priqueue_sound.
            rewrite HeqNnat.
            rewrite Z2Nat.id; omega.
        }
        rewrite sched_type_sum_aux_inc_when_schedule_not_idle.
        - rewrite Nat2Z.inj_succ; omega.
        - omega.
      }
    }
  Qed.

  Lemma sched_type_is_bounded_by_window:
    forall N conf id ts te todeadline,
      0 <= ts <= te ->
      0 <= id < N ->
      count_type_occurrence N N ts te conf id todeadline Sched_self +
      count_type_occurrence N N ts te conf id todeadline Sched_interference +
      count_type_occurrence N N ts te conf id todeadline Sched_leftover +
      count_type_occurrence N N ts te conf id todeadline Sched_overlap <= te - ts.
  Proof.
    intros * Ht Hid.
    unfold count_type_occurrence.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te - ts = Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hte.
    apply sched_type_is_bounded_by_window_aux; auto.
    omega.
  Qed.
  
End SCHED_TYPE.


Section SCHED_SELF.

  Lemma count_occurrence_self_bounded_by_budget_single_period_aux:
    forall N ts nrange conf id T C,
      0 <= id < N ->
      0 <= ts ->
      1 < T ->
      0 < C ->
      ts mod T = 0 ->
      Z.of_nat nrange <= T ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      count_type_occurrence_aux N N ts nrange conf id (ts+T) Sched_self <= Z.min C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
  Proof.
    intros * Hid Hts HT HC Hmod Hnrange Hconf.
    induction nrange.
    {
      simpl.
      replace (ts + 0) with ts by omega.
      rewrite Z.min_r; omega.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H1.
      simpl.
      assert (HN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.
      assert (Hmono: monotonic_continuous (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf))).
      {
        apply TAC_pool_all_monotonic_continuous; auto; try xomega.
      }        
      assert (H2: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) \/
                  ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
      {
        replace (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange)) with (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange)).
        - apply Hmono; omega.
        - rewrite vertical_map_canonical; omega.
      }

      assert (Hrange': Z.of_nat nrange <= T). xomega.
      specialize (IHnrange Hrange'). clear Hrange'.

      assert (H3:    Z.min C
     (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
      ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts) <=
   Z.min C
     (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) -
      ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)).
      {
        destruct H2 as [Hvt0 | Hvt1]; [rewrite Hvt0; omega | ].
        rewrite Hvt1.
        destruct (zle C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)) as [Hl | Hr].
        - rewrite Z.min_l; try omega.
          rewrite Z.min_l; omega.
        - rewrite Z.min_r; try omega.
          rewrite Z.min_r; omega.
      }
        
      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N) as [Hidle|Hreal]; try omega.
      
      (* not idle *)
      destruct (ZMap.get
       (highest_not_finished_map N (ts + Z.of_nat nrange)
          (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) conf) as [T0 C0] eqn:Hconf0.
      destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map N (ts + Z.of_nat nrange)
               (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) T0 id 
            (ts + Z.of_nat nrange) (ts+T)) Sched_self) as [Hself | Hothers]; try omega.

      (* self is scheduled *)
      pose proof (get_sched_type_is_self _ _ _ _ _ Hself) as Hideq.
      assert (H5: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
      {
        apply vt_increase_for_the_scheduled_task; auto; try omega.
        rewrite <- HN.
        auto.
      }
      assert (H6: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts < C).
      {
        assert (H7: vt_map_not_finished (ts + Z.of_nat nrange) T C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)) = true).
        {
          apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:= conf); auto; try omega.
          rewrite <- HN.
          auto.
        }
        unfold vt_map_not_finished in H7.
        rewrite divide_plus_equal in H7; try omega.
        rewrite mod_0_divide_mul in H7; try omega.
        assert (H8: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts = ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
        { rewrite vertical_map_canonical; omega. }
        rewrite H8 in H7.
        destruct (zlt
         (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
          ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts) C); try omega; try discriminate.
      }
      rewrite H5.
      rewrite Z.min_r in IHnrange; try omega.
      rewrite Z.min_r; try omega.      
    }
  Qed.
  
  Lemma count_occurrence_self_bounded_by_budget_single_period:
    forall N ts conf id T C,
      0 <= id < N ->
      0 <= ts ->      
      1 < T ->
      0 < C ->
      ts mod T = 0 ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      count_type_occurrence N N ts (ts + T) conf id (ts+T) Sched_self <= C.
  Proof.
    intros.
    unfold count_type_occurrence.
    assert (Z.of_nat (Z.to_nat (ts + T - ts)) <= T).
    { rewrite Z2Nat.id; omega. }
    replace (ts + T - ts) with T by omega.
    assert (count_type_occurrence_aux N N ts (Z.to_nat T) conf id (ts+T) Sched_self <= Z.min C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat (Z.to_nat T)) (Z.to_nat N) conf) (ts + Z.of_nat (Z.to_nat T)) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)).
    {
      apply count_occurrence_self_bounded_by_budget_single_period_aux; auto.
      rewrite Z2Nat.id; omega.
    }
    transitivity (Z.min C
                        (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat (Z.to_nat T)) (Z.to_nat N) conf) (ts + Z.of_nat (Z.to_nat T))- ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)); auto.
    apply Z.le_min_l.
  Qed.

End SCHED_SELF.



Section SCHED_OVERLAP.

  (* This requires that the current task arrives no earlier than ts *)
  Lemma non_overlapping_no_remain:
    forall fromid toid Tf Cf T ts vt,
      0 <= ts ->
      ts mod T = 0 ->
      1 < Tf <= T ->
      is_overlapping fromid Tf toid ts (ts+T) (ts + 2*T) = false ->
      workload_remain toid fromid Tf Cf (ts + T) (ts + 2 * T) vt = 0.
  Proof.
    intros * Hts Htsmod HT Hnonoverlap.
    unfold is_overlapping in Hnonoverlap.
    unfold workload_remain.
    destruct (zeq ((ts + T) mod Tf) 0); auto.
    assert (ts <= ((ts + T) / Tf * Tf)).
    {
      apply mod_not_0_adjust_to_smaller_period; omega.
    }
    destruct (zle ts ((ts + T) / Tf * Tf)); try omega.
    assert (((ts + T) / Tf * Tf) < (ts + T)).
    {
      assert (ts + T - (ts + T) / Tf * Tf >= 1).
      { apply z_gt_mod; omega. }
      omega.
    }
    destruct (zlt ((ts + T) / Tf * Tf) (ts + T)); try omega.
    destruct (priEntry_lt (PriEntry ((ts + T) / Tf * Tf + Tf) fromid) (PriEntry (ts + 2 * T) toid)); auto.
    congruence.
  Qed.

  Lemma is_overlapping_implies_remain:
    forall fromid toid Tf Cf d1 d2 d3 vt,
      0 <= d2 ->
      0 < Tf ->
      0 < Cf ->
      is_overlapping fromid Tf toid d1 d2 d3 = true ->
      workload_remain toid fromid Tf Cf d2 d3 vt =
      Cf - Z.min Cf (vt d2 - vt (d2 / Tf * Tf)).
  Proof.
    intros * Hd2 HT HC Hoverlap.
    unfold is_overlapping in Hoverlap.
    unfold workload_remain.
    destruct (zeq (d2 mod Tf) 0) as [Heq|Hneq].
    {      
      rewrite mod_0_divide_mul in Hoverlap; try omega.
      destruct (zle d1 d2); try omega; try congruence.
      destruct (zlt d2 d2); try omega.
      congruence.
    }
    {
      destruct (zle d1 (d2 / Tf * Tf)); try congruence.
      destruct (zlt (d2 / Tf * Tf) d2); try congruence.
      destruct (priEntry_lt (PriEntry (d2 / Tf * Tf + Tf) fromid) (PriEntry d3 toid)); try congruence.
      destruct (zle (vt d2 - vt (d2 / Tf * Tf)) Cf).
      - rewrite Z.max_r; try omega.
        rewrite Z.min_r; omega.
      - rewrite Z.max_l; try omega.
        rewrite Z.min_l; omega.
    }
  Qed.
    
  Lemma workload_overlapping_remain_total_eq:
    forall Nnat id p T conf TAC,
      0 <= p ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
      ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      workload_overlapping_total Nnat id (p * T) ((p+1)*T) ((p+1)*T + T) conf -
      workload_remain_total Nnat id ((p+1)*T) ((p+1)*T + T) conf TAC =
      workload_overlapping_remain_total Nnat id (p*T) (p*T + T) ((p+1)*T+ T) (p*T + T) conf TAC.
  Proof.
    intros * Hp HT Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: 0 <= p * T). nia.
      assert (H2: (p*T) mod T = 0). apply Z_mod_mult.
      assert (H3: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (H4: config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (H5: forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T).
      { intros. eapply Hlarge; eauto; omega. }
      specialize (IHNnat H4 H5).
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.

      assert (H6: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
      assert (H7: 1 < Tn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      assert (H8: Tn <= T).
      { specialize (Hlarge _ _ _ H6 Hconfn). omega. }
      assert (H9: 0 < Cn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }

      replace ((p+1) * T) with (p*T + T) in * by nia.      
      destruct (is_overlapping (Z.of_nat Nnat) Tn id (p * T) (p * T + T) (p * T + T + T)) eqn:Hoverlap.
      {
        replace (p * T + T + T) with (p*T + 2*T) in * by omega.
        rewrite is_overlapping_implies_remain with (d1:=p*T); auto; try omega.

        replace (Cn + workload_overlapping_total Nnat id (p * T) (p * T + T) (p * T + 2 * T) conf -
   (Cn -
    Z.min Cn
      (ZMap.get (Z.of_nat Nnat) TAC (p * T + T) -
       ZMap.get (Z.of_nat Nnat) TAC ((p * T + T) / Tn * Tn)) +
    workload_remain_total Nnat id (p * T + T) (p * T + 2 * T) conf TAC)) with
            (workload_overlapping_total Nnat id (p * T) (p * T + T) (p * T + 2 * T) conf -
             workload_remain_total Nnat id (p * T + T) (p * T + 2 * T) conf TAC +
            Z.min Cn
      (ZMap.get (Z.of_nat Nnat) TAC (p * T + T) -
       ZMap.get (Z.of_nat Nnat) TAC ((p * T + T) / Tn * Tn))) by omega.

        rewrite <- IHNnat.
        auto.
      }
      {
        assert (Hremain: workload_remain id (Z.of_nat Nnat) Tn Cn (p * T + T) (p * T + T + T) (ZMap.get (Z.of_nat Nnat) TAC) = 0).
        {
          replace (p* T + T + T) with (p*T + 2*T) in *by omega.
          apply non_overlapping_no_remain; auto; try omega.
        }
        omega.
      }
    }
  Qed.
  
  Lemma none_overlapping_then_type_not_overlap:
    forall fromid Tf toid T ts t,
      0 <= ts ->
      ts <= t < ts + T ->
      1 < Tf <= T ->
      is_overlapping fromid Tf toid ts (ts+T) (ts + 2*T) = false ->
      get_sched_type fromid Tf toid t (ts+T) <> Sched_overlap.
  Proof.
    intros * Hts Ht HTf Hnone.
    unfold is_overlapping in Hnone.
    unfold get_sched_type.
    destruct (zeq fromid toid); try congruence.
    assert (H1: ts <= (ts + T) / Tf * Tf <= ts + T).
    {
      split.
      - apply mod_not_0_adjust_to_smaller_period; auto.
      - apply divide_mul_le; omega.
    }
    destruct (zeq ((ts +T) mod Tf) 0) as [Hmod | Hmod].
    {
      unfold not.
      intros H.
      destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) fromid) (PriEntry (ts + T) toid)) eqn:Hpri; try congruence.
      destruct (zeq (t / Tf * Tf + Tf) (ts + T)) as [Heq | Hneq]; try congruence.
      unfold priEntry_lt in Hpri.
      destruct (zlt (t / Tf * Tf + Tf) (ts + T)); try congruence.      
      assert (t / Tf * Tf + Tf <= (ts + T) - 0 * Tf).
      { apply mod_0_bound_increase_k; omega. }
      lia.
    }
    {
      assert (H2: (ts + T) - (ts + T) / Tf * Tf >= 1).
      { apply z_gt_mod; omega. }
      destruct (zle ts ((ts + T) / Tf * Tf)); try omega.
      destruct (zlt ((ts + T) / Tf * Tf) (ts + T)); try omega.
      destruct (priEntry_lt (PriEntry ((ts + T) / Tf * Tf + Tf) fromid)
                            (PriEntry (ts + 2 * T) toid)) eqn:Hpri0; try congruence.
      unfold priEntry_lt in Hpri0.
      rewrite zlt_true in Hpri0; try omega.
      congruence.
    }
  Qed.

  Lemma is_overlapping_yet_not_overlap_when_t_is_small:
    forall fromid Tf toid T ts t,
      0 <= ts ->      
      ts <= t < (ts + T) / Tf * Tf ->
      1 < Tf <= T ->
      is_overlapping fromid Tf toid ts (ts +T) (ts+ 2*T) = true ->
      get_sched_type fromid Tf toid t (ts+T) <> Sched_overlap.
  Proof.
    intros * Hts Ht HTf Hoverlap.
    unfold get_sched_type, not.
    intros H.
    destruct (zeq fromid toid); try congruence.
    destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) fromid) (PriEntry (ts + T) toid)) eqn:Hpri; try congruence.
    destruct (zeq (t / Tf * Tf + Tf) (ts + T)) as [Heq | Hneq]; try congruence.
    unfold priEntry_lt in Hpri.
    destruct (zlt (t / Tf * Tf + Tf) (ts + T)); try congruence.
    assert (H1 : t / Tf * Tf + Tf <= (ts + T) / Tf * Tf - 0 * Tf).
    {
      apply mod_0_bound_increase_k; try omega.
      apply Z_mod_mult.
    }

    assert (H2: (ts + T) / Tf * Tf <= ts + T).
    { apply divide_mul_le; omega. }
    omega.
  Qed.
  
  Lemma is_overlapping_then_single_task_type_count_is_0_with_small_t:
    forall N0 fromid toid ts T Tf Cf conf nrange,
      0 <= ts ->
      ts + Z.of_nat nrange <= (ts + T) / Tf * Tf ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= T ->
      is_overlapping fromid Tf toid ts (ts+T) (ts+2*T) = true ->
      count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_overlap = 0.
  Proof.
    intros * Hts Hrange Hconf HTf Hoverlap.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      rewrite IHnrange; try omega.
      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.
      rewrite Heq.
      rewrite Hconf.
      rewrite sched_type_eq_false; auto.
      apply is_overlapping_yet_not_overlap_when_t_is_small; auto.
      omega.
    }
  Qed.

  Lemma is_overlapping_type_is_overlap_generalized:
    forall fromid Tf toid ts T t,
      0 <= ts ->
      1 < Tf <= T ->
      (ts + T) / Tf * Tf <= t < ts + T ->
      is_overlapping fromid Tf toid ts (ts+T) (ts+2*T) = true ->
      fromid <> toid ->
      get_sched_type fromid Tf toid t (ts+T) = Sched_overlap.
  Proof.
    intros * Hts HTf Ht Hoverlap Hid.
    unfold is_overlapping in Hoverlap.
    unfold get_sched_type.
    rewrite zeq_false; auto.
    destruct (zle ts ((ts + T) / Tf * Tf)); try congruence.
    destruct (zlt ((ts + T) / Tf * Tf) (ts + T)); try congruence.
    destruct (priEntry_lt (PriEntry ((ts + T) / Tf * Tf + Tf) fromid)
                          (PriEntry (ts + 2 * T) toid)) eqn:Hpri; try congruence.

    unfold priEntry_lt.
    assert (H1: ((ts + T) / Tf * Tf) mod Tf = 0). apply Z_mod_mult.
    assert (H2: ts + T < (ts + T) / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; omega.
    }
    assert (H3: (ts + T) / Tf * Tf = ((ts + T) / Tf * Tf ) / Tf * Tf).
    {
      symmetry.
      rewrite mod_0_divide_mul; try omega.    
    }
    
    assert (H4: (ts + T) / Tf * Tf / Tf * Tf <= t / Tf * Tf).
    {
      apply div_mul_le; omega.
    }
    rewrite zlt_false; try omega.
    repeat rewrite zeq_false; try omega.
    auto.
  Qed.
  
  Lemma is_overlapping_type_is_overlap:
    forall fromid Tf toid d1 d2 d3,
      0 <= d2 ->
      0 < Tf ->      
      is_overlapping fromid Tf toid d1 d2 d3 = true ->
      fromid <> toid ->
      get_sched_type fromid Tf toid (d2 / Tf * Tf) d2 = Sched_overlap.
  Proof.
    intros * Hd2 HT Hoverlap Hid.
    unfold is_overlapping in Hoverlap.
    unfold get_sched_type.
    rewrite zeq_false; auto.
    destruct (zle d1 (d2 / Tf * Tf)); try congruence.
    destruct (zlt (d2 / Tf * Tf) d2); try congruence.
    destruct (priEntry_lt (PriEntry (d2 / Tf * Tf + Tf) fromid)
                          (PriEntry d3 toid)) eqn:Hpri; try congruence.

    unfold priEntry_lt.
    assert (H1: (d2 / Tf * Tf) mod Tf = 0). apply Z_mod_mult.
    assert (H3: 0 <= d2 / Tf * Tf).
    {
      apply div_mul_nonneg; omega.
    }
    rewrite mod_0_divide_mul; try omega.

    assert (H2: d2 < d2 / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; omega.
    }
    rewrite zlt_false; try omega.
    repeat rewrite zeq_false; try omega.
    auto.
  Qed.
  
  Lemma is_overlapping_then_single_task_type_count_is_valid:
    forall N0 fromid toid ts T Tf Cf conf nrange,
      0 < N0 ->
      0 <= ts ->
      (ts + T) / Tf * Tf < ts + Z.of_nat nrange <= ts + T ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= T ->
      0 < Cf ->
      is_overlapping fromid Tf toid ts (ts+T) (ts+2*T) = true ->
      fromid <> toid ->
      0 <= fromid < N0 ->
      count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_overlap = Z.min Cf (ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf) (ts + Z.of_nat nrange) - ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf) ((ts + Z.of_nat nrange) / Tf * Tf)).
  Proof.
    intros * HN0 Hts Hrange Hconf HTf HCf Hoverlap Hid Hfrom.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl in Hrange.
      assert (ts <= (ts + T) / Tf * Tf).
      { apply mod_not_0_adjust_to_smaller_period; omega. }
      omega.
    }
    {
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      assert (H2: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H2.
      assert (H3: ((ts + T) / Tf * Tf) mod Tf = 0).
      { apply Z_mod_mult. }
      assert (H10: 0 <= (ts + T) / Tf * Tf).
      {
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H11: 0 <= fromid < Z.of_nat (Z.to_nat N0)).
      { rewrite Z2Nat.id; omega. }
      assert (H12: (ts + Z.of_nat nrange + 1) mod Tf <> 0).
      {            
        apply a_between_floor_and_self_mod_not_0 with (c := ts + T); omega.
      }
      assert (H15: 0 <= (ts + Z.of_nat nrange) / Tf * Tf).
      { 
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H20: (ts + Z.of_nat nrange) / Tf * Tf <= ts + Z.of_nat nrange).
      {
        apply divide_mul_le; omega.
      }
      
      simpl.
      
      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Hsched|Hnot].
      {
        rewrite Hsched.
        rewrite Hconf.

        assert (Htype: get_sched_type fromid Tf toid (ts + Z.of_nat nrange) (ts+T) = Sched_overlap).
        {
          apply is_overlapping_type_is_overlap_generalized; auto.
          omega.
        }
        rewrite Htype.
        rewrite sched_type_eq_true.

        destruct (zeq (ts + Z.of_nat nrange) ((ts + T) / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
            
          assert (H0: count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_overlap = 0).
          {
            apply is_overlapping_then_single_task_type_count_is_0_with_small_t with (Tf:=Tf) (Cf:=Cf); auto.
            omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
        
          assert (H4: ((ts + T) / Tf * Tf + 1) / Tf * Tf = (ts + T) / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H5: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf) ((ts + T) / Tf * Tf + 1) = ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf) ((ts + T) / Tf * Tf) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto.
            rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.

          rewrite mod_equal; try omega.
          assert (H14: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange + 1) =
                        ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto.
            omega.
            rewrite Z2Nat.id; omega.            
          }

          rewrite H14.

          assert (H16: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf) (ts + Z.of_nat nrange) -
      ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
               ((ts + Z.of_nat nrange) / Tf * Tf) < Cf).
          {
            symmetry in Hsched.
            replace N0 with (Z.of_nat (Z.to_nat N0)) in Hsched.
            
            rewrite Nat2Z.id in Hsched.
            pose proof (highest_not_finished_map_condition_met _ _ _ _ _ _ _ Hsched H11 Hconf).
            - unfold vt_map_not_finished in H.
              rewrite vertical_map_canonical; try omega.
              assert ((ts + Z.of_nat nrange) / Tf * Tf <= ts + Z.of_nat nrange).
              { apply divide_mul_le; omega. }
              assert (ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
     ((ts + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                                                   ((ts + Z.of_nat nrange) / Tf * Tf)).
              {
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }
              rewrite H4.
              destruct (zlt
        (ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
           (ts + Z.of_nat nrange) -
         ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                  ((ts + Z.of_nat nrange) / Tf * Tf)) Cf); try omega.
              discriminate H.
            - rewrite Z2Nat.id; omega.
          }

          repeat rewrite Z.min_r; omega.
        }
      }
      {
        (* fromid not scheduled *)
        destruct (zeq (ts + Z.of_nat nrange) ((ts + T) / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
          
          assert (H0: count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_overlap = 0).
          {
            apply is_overlapping_then_single_task_type_count_is_0_with_small_t with (Tf:=Tf) (Cf:=Cf); auto.
            omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
          
          assert (H4: ((ts + T) / Tf * Tf + 1) / Tf * Tf = (ts + T) / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H6: vt_map_not_finished ((ts + T) / Tf * Tf) Tf Cf
                                          (ZMap.get fromid (TAC_pool_at_t ((ts + T) / Tf * Tf) (Z.to_nat N0) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          assert (H5: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                               ((ts + T) / Tf * Tf + 1) =
                      ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                               ((ts + T) / Tf * Tf)).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          rewrite mod_equal; try omega.

          destruct (vt_map_not_finished (ts + Z.of_nat nrange) Tf Cf
     (ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange + 1) =
                   ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                            (ts + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            auto.
          }
          {
            unfold vt_map_not_finished in Hfinish.            
            
            assert (H16: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                                              (ts + Z.of_nat nrange) \/
                    ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
                                              (ts + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }

            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.
              assert (H18: ZMap.get fromid (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf)
        ((ts + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                                                      ((ts + Z.of_nat nrange) / Tf * Tf)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }

              rewrite H18.

              destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                 (ts + Z.of_nat nrange) -
               ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                        ((ts + Z.of_nat nrange) / Tf * Tf)) Cf); try discriminate.

              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }
  Qed.      
     
  Lemma non_overlapping_not_counted_by_single_task_type_count:
    forall N0 fromid toid ts T Tf Cf conf nrange,
      0 <= ts ->
      Z.of_nat nrange <= T ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= T ->
      is_overlapping fromid Tf toid ts (ts+T) (ts+2*T) = false ->
      count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_overlap = 0.
  Proof.
    intros * Hts Hrange Hconf HTf Hoverlap.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      simpl.
      rewrite IHnrange; try omega.
      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.

      rewrite Heq.
      rewrite Hconf.
      assert (Htype: get_sched_type fromid Tf toid (ts + Z.of_nat nrange) (ts+T) <> Sched_overlap).
      {
        apply none_overlapping_then_type_not_overlap; auto; try omega.
      }
      rewrite sched_type_eq_false; auto.
    }
  Qed.

  Lemma count_single_task_self_sched_others_is_0:
    forall N1 N2 ts te conf i todeadline ty,      
      ts <= te ->
      ty <> Sched_self ->
      count_single_task_type_occurrence N1 N2 i ts te conf i todeadline ty = 0.
  Proof.
    intros * Ht Hty.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te - ts)) as range.
    clear Heqrange.
    induction range.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHrange.
      destruct (zeq
         (highest_not_finished_map N2 (ts + Z.of_nat range)
            (TAC_pool_at_t (ts + Z.of_nat range) (Z.to_nat N1) conf) conf
            (real_priqueue (Z.to_nat N1) (ts + Z.of_nat range) conf)) i) as [Heq|Hneq]; auto.
      {
        rewrite Heq.        
        destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconf.
        rewrite sched_type_eq_false; auto.
        unfold not.
        intros H.
        unfold get_sched_type in H.
        rewrite zeq_true in H.
        congruence.
      }
    }
  Qed.
  
  Lemma count_single_task_self_sched_overlap_is_0:
    forall N1 N2 ts te conf i todeadline,
      ts <= te ->
      count_single_task_type_occurrence N1 N2 i ts te conf i todeadline Sched_overlap = 0.
  Proof.
    intros * Ht.
    apply count_single_task_self_sched_others_is_0; auto.
    congruence.
  Qed.
      
  Lemma count_occurrence_overlapping_bounded_aux:
    forall Nnat N0 id ts T conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 ->
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence2_aux Nnat N0 N0
                           ts (ts + T) conf id (ts+T) Sched_overlap =
      workload_overlapping_remain_total Nnat id ts (ts+T) (ts+T+T) (ts + T) conf (TAC_pool_at_t (ts + T + T) (Z.to_nat N0) conf).
  Proof.
    intros * HNnat HN0 Hts HT Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.
      rewrite <- IHNnat.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_overlap_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.
        destruct (is_overlapping (Z.of_nat Nnat) Tf id ts (ts+T) (ts+T+T)) eqn:Hoverlap.
        {
          assert (H: forall a b c, b = c -> a + b = a + c). intros. omega.
          apply H.
          remember (Z.to_nat T) as nT.
          assert (HTT: T = Z.of_nat nT).
          { rewrite HeqnT. rewrite Z2Nat.id; omega. }
          rewrite HTT in *. clear HTT HeqnT.

          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.          
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: 0 < Cf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H5: Tf <= Z.of_nat nT).
          { eapply Hlarge; eauto. }          
          
          apply is_overlapping_then_single_task_type_count_is_valid; auto; try omega.
          - pose proof (is_overlapping_true_end_not_aligned _ _ _ _ _ _ Hoverlap).
            assert ((ts + Z.of_nat nT) / Tf * Tf <= ts + Z.of_nat nT).
            { apply divide_mul_le; omega. }
            omega.
          - replace (ts+2 * Z.of_nat nT) with (ts + Z.of_nat nT + Z.of_nat nT) by lia.
            auto.
        }
        {
          assert (H: forall a b, b = 0 -> a + b = a). intros. omega.
          apply H.
          remember (Z.to_nat T) as nT.
          assert (HTT: T = Z.of_nat nT).
          { rewrite HeqnT. rewrite Z2Nat.id; omega. }
          rewrite HTT in *. clear HTT HeqnT.
          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: Tf <= Z.of_nat nT).
          { eapply Hlarge; eauto. }
          apply non_overlapping_not_counted_by_single_task_type_count with (Tf := Tf) (Cf := Cf); auto; try omega.
          - replace (ts+2 * Z.of_nat nT) with (ts + Z.of_nat nT + Z.of_nat nT) by lia.
            auto.
        }
      }
    }
  Qed.    

  Lemma count_occurrence_overlapping_bounded:
    forall Nnat id ts T conf,
      0 < Z.of_nat Nnat ->
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           ts (ts + T) conf id (ts+T) Sched_overlap =
      workload_overlapping_remain_total Nnat id ts (ts+T) (ts+T+T) (ts + T) conf (TAC_pool_at_t (ts + T + T) Nnat conf).
  Proof.
    intros * HN Hts HT Hconfvalid Hlarge.
    
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    rewrite count_occurrence_overlapping_bounded_aux; auto; try omega.
    - rewrite H1. auto.
  Qed.    
  
End SCHED_OVERLAP.



Section SCHED_LEFTOVER.

  Lemma count_single_task_self_sched_leftover_is_0:
    forall N1 N2 ts te conf i todeadline,
      ts <= te ->
      count_single_task_type_occurrence N1 N2 i ts te conf i todeadline Sched_leftover = 0.
  Proof.
    intros * Ht.
    apply count_single_task_self_sched_others_is_0; auto.
    congruence.
  Qed.   

  Lemma is_leftover_false_get_type_not_leftover:
    forall i Ti toid T ts te,
      0 < Ti <= T ->
      is_leftover i Ti toid ts (ts+T) = false ->
      ts <= te <= ts + T ->
      get_sched_type i Ti toid te (ts+T) <> Sched_leftover.
  Proof.
    intros * HT Hleftover Hrange.
    unfold not.
    intros H.
    unfold is_leftover in Hleftover.    
    unfold get_sched_type in H.

    destruct (zeq i toid); try congruence.
    destruct (priEntry_lt (PriEntry (te / Ti * Ti + Ti) i) (PriEntry (ts + T) toid)) eqn:Hpri; try congruence.
    unfold priEntry_lt in Hpri.    
    destruct (zeq (te / Ti * Ti + Ti) (ts + T)) as [Heq | Hneq]; try congruence.
    rewrite Heq in Hpri.
    rewrite zlt_false in Hpri; try omega.
    destruct (zlt i toid); try congruence.
    rewrite zle_true in Hleftover; try omega.
    rewrite zlt_true in Hleftover; try omega.

    assert (Hmod: (ts + T) mod Ti = 0).
    {
      rewrite <- Heq.
      rewrite mod_0_preserve; try omega.
      apply Z_mod_mult.
    }
    rewrite Hmod in Hleftover.
    rewrite zeq_true in Hleftover.
    congruence.
  Qed.
  
  Lemma count_single_task_self_sched_leftover_is_0_when_not_leftover:
    forall N1 N2 ts te conf i Ti Ci toid T,
      0 < Ti <= T ->
      ts <= te <= ts + T ->
      is_leftover i Ti toid ts (ts+T) = false ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      count_single_task_type_occurrence N1 N2 i ts te conf toid (ts+T) Sched_leftover = 0.
  Proof.
    intros * HT Ht Hleftover Hconfi.
    unfold count_single_task_type_occurrence.
    remember (Z.to_nat (te -ts)) as range.
    assert (Hrange: Z.of_nat range <= T).
    { rewrite Heqrange. rewrite Z2Nat.id; omega. }
    clear Heqrange Ht.
    
    induction range.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHrange; try xomega.
      remember ((highest_not_finished_map N2 (ts + Z.of_nat range)
            (TAC_pool_at_t (ts + Z.of_nat range) (Z.to_nat N1) conf) conf
            (real_priqueue (Z.to_nat N1) (ts + Z.of_nat range) conf))) as cur.
      destruct (zeq cur i) as [Heq | Hneq]; auto.
      {
        rewrite Heq.
        rewrite Hconfi.
        rewrite sched_type_eq_false; auto.
        apply is_leftover_false_get_type_not_leftover; auto; try xomega.
      }
    }
  Qed.        

  Lemma is_leftover_type_is_leftover_generalized:
    forall fromid Tf toid t ts T,
      0 <= ts ->
      1 < Tf <= T ->
      ts + T - Tf <= t < ts + T ->
      is_leftover fromid Tf toid ts (ts+T) = true ->
      get_sched_type fromid Tf toid t (ts+T) = Sched_leftover.
  Proof.
    intros * Hts HTf Ht Hleftover.
    unfold is_leftover in Hleftover.
    rewrite zle_true in Hleftover; try omega.
    destruct (zeq ((ts + T) mod Tf) 0) as [Hmod|Hmod]; try congruence.
    destruct (zlt toid fromid); try congruence.

    unfold get_sched_type.
    rewrite zeq_false; try omega.

    assert (H1: (ts + T - Tf) mod Tf = 0).
    {
      replace (ts + T) with ((ts + T - Tf) + Tf) in Hmod by omega.
      rewrite Zplus_mod in Hmod.
      rewrite Z_mod_same_full in Hmod.
      replace ((ts + T - Tf) mod Tf + 0) with ((ts + T - Tf) mod Tf) in Hmod by omega.
      rewrite Zmod_mod in Hmod.
      auto.
    }
    assert (H: t / Tf * Tf + Tf = ts + T).
    {
      remember (t - ts - T + Tf) as d.
      replace t with ((ts + T - Tf) + d) by omega.
      rewrite divide_plus_equal; try omega.
      - rewrite mod_0_divide_mul; omega.
    }
    rewrite H.

    unfold priEntry_lt.
    rewrite zlt_false; try omega.
    repeat rewrite zeq_true.
    rewrite zlt_false; try omega.
    auto.
  Qed.

  Lemma is_leftover_yet_not_leftover_when_t_is_small:
    forall fromid Tf toid T ts t,
      0 <= ts ->      
      ts <= t < ts + T - Tf ->
      1 < Tf <= T ->
      is_leftover fromid Tf toid ts (ts+T) = true ->
      get_sched_type fromid Tf toid t (ts+T) <> Sched_leftover.
  Proof.
    intros * Hts Ht HTf Hleftover.
    unfold get_sched_type, not.
    intros H.    
    destruct (zeq fromid toid); try congruence.
    destruct (priEntry_lt (PriEntry (t / Tf * Tf + Tf) fromid) (PriEntry (ts + T) toid)); try congruence.
    destruct (zeq (t / Tf * Tf + Tf) (ts + T)) as [Heq | Hneq]; try congruence.

    unfold is_leftover in Hleftover.
    destruct (zle ts (ts + T - Tf)); try congruence.
    destruct (zeq ((ts + T) mod Tf) 0) as [Hmod|Hmod]; try congruence.

    assert (t / Tf * Tf + Tf <= (ts + T) - Tf).
    {
      apply mod_0_bound_increase; omega.
    }
    omega.
  Qed.
  
  Lemma is_leftover_then_single_task_type_count_is_0_with_small_t:
    forall N0 fromid toid ts T Tf Cf conf nrange,
      0 <= ts ->
      Z.of_nat nrange <= T - Tf ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf <= T ->
      is_leftover fromid Tf toid ts (ts+T) = true ->
      count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_leftover = 0.
  Proof.
    intros * Hts Hrange Hconf HTf Hleftover.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.
      rewrite Heq.
      rewrite Hconf.
      rewrite sched_type_eq_false; auto.
      apply is_leftover_yet_not_leftover_when_t_is_small; auto.
      xomega.
    }
  Qed.
  
  Lemma count_single_task_self_sched_leftover_bounded_by_budget_aux:
    forall N ts nrange conf i Ti Ci toid T,
      0 <= N ->
      1 < Ti <= T ->
      0 < Ci ->
      0 <= ts ->      
      T - Ti < Z.of_nat nrange <= T ->
      0 <= i < N ->      
      is_leftover i Ti toid ts (ts+T) = true ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_leftover = Z.min Ci
         (ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - (ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti))).
  Proof.
    intros * HN HT HC Hts Hrange Hi Hleftover Hconf.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    
    induction nrange.
    {      
      simpl in *.
      replace (ts + 0) with ts by omega.
      replace (ts + T - Ti) with ts by omega.
      rewrite Z.min_r; try omega.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H1.
      simpl.
      assert (HNN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.

      assert (H2: ts + Z.of_nat (S nrange) = ts + Z.of_nat nrange + 1).
      { rewrite Nat2Z.inj_succ; omega. }
      assert (H11: 0 <= i < Z.of_nat (Z.to_nat N)).
      { rewrite Z2Nat.id; omega. }
      assert (H20: (ts + T) mod Ti = 0).
      {
        unfold is_leftover in Hleftover.
        destruct (zle ts (ts + T - Ti)); try congruence.
        destruct (zeq ((ts + T) mod Ti) 0); auto; try congruence.
      }
      assert (H21: (ts + T - Ti) mod Ti = 0).
      {
        replace (ts + T) with (ts + T - Ti + Ti) in H20 by omega.
        rewrite Zplus_mod in H20.
        rewrite Z_mod_same_full in H20.
        replace ((ts + T - Ti) mod Ti + 0) with ((ts + T - Ti) mod Ti) in H20 by omega.
        rewrite Zmod_mod in H20.
        auto.
      }
      
      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) i) as [Hsched|Hother].
      {
        rewrite Hsched.
        rewrite Hconf.

        assert (Htype: get_sched_type i Ti toid (ts + Z.of_nat nrange) (ts+T) = Sched_leftover).
        {
          apply is_leftover_type_is_leftover_generalized; auto; omega.
        }
        rewrite Htype.
        rewrite sched_type_eq_true.

        destruct (zeq (Z.of_nat nrange) (T - Ti)) as [Heq | Hneq].
        {          
          rewrite Heq in *. clear Heq.
          
          assert (H0: count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_leftover = 0).
          {
            apply is_leftover_then_single_task_type_count_is_0_with_small_t with (Tf:=Ti) (Cf:=Ci); auto.
            xomega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.

          replace (ts + (T - Ti)) with (ts + T - Ti) in * by omega.
          assert (H5: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti + 1) = ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti) + 1).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto; try omega.
            - rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          
          assert (H14: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) =
                       ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto; try omega.
            - rewrite Z2Nat.id; omega.            
          }

          rewrite H14.

          assert (H15: (ts + Z.of_nat nrange) / Ti * Ti = ts + T - Ti).
          {
            apply mod_lower_bound; try omega.
          }
          
          assert (H16: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
      ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti) < Ci).
          {
            symmetry in Hsched.
            replace N with (Z.of_nat (Z.to_nat N)) in Hsched.
            
            rewrite Nat2Z.id in Hsched.
            pose proof (highest_not_finished_map_condition_met _ _ _ _ _ _ _ Hsched H11 Hconf).
            - unfold vt_map_not_finished in H.
              rewrite vertical_map_canonical; try omega.
              rewrite H15 in H.
              assert ((ts + Z.of_nat nrange) / Ti * Ti <= ts + Z.of_nat nrange).
              { apply divide_mul_le; omega. }
              
              assert (H17: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + T - Ti) = ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti)).
              {
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }
              rewrite H17 in H.
              destruct (zlt
        (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
           (ts + Z.of_nat nrange) -
         ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti)) Ci); try omega.
              discriminate.
          }
          repeat rewrite Z.min_r; try omega.
        }
      }
      {
        (* i not scheduled *)
        destruct (zeq (Z.of_nat nrange) (T - Ti)) as [Heq | Hneq].
        {          
          rewrite Heq in *. clear Heq.
          
          assert (H0: count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid (ts+T) Sched_leftover = 0).
          {
            apply is_leftover_then_single_task_type_count_is_0_with_small_t with (Tf:=Ti) (Cf:=Ci); auto.
            xomega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.

          assert (H6:    vt_map_not_finished (ts + T - Ti) Ti Ci
                                             (ZMap.get i (TAC_pool_at_t (ts + T - Ti) (Z.to_nat N) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          replace (ts + (T - Ti)) with (ts + T - Ti) in * by omega.
          assert (H5: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti + 1) = ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) (ts + T - Ti)).
          {
            rewrite vertical_map_canonical; try omega.            
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Ti) (C:=Ci); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=Ti) (C:=Ci); auto.              
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.        
        }
        {
          rewrite IHnrange; try omega.
          destruct (vt_map_not_finished (ts + Z.of_nat nrange) Ti Ci
                                        (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                                 (ts + Z.of_nat nrange + 1) =
                   ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                            (ts + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Ti) (C:=Ci); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=Ti) (C:=Ci); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            auto.
          }
          {
            unfold vt_map_not_finished in Hfinish.            
            
            assert (H16: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                                              (ts + Z.of_nat nrange) \/
                    ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                                              (ts + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }

            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.

              assert (H18: ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
        (ts + T - Ti) = ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                                                      (ts + T - Ti)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }

              rewrite H18.
              
              assert (H15: (ts + Z.of_nat nrange) / Ti * Ti = ts + T - Ti).
              {
                apply mod_lower_bound; try omega.
              }
              rewrite H15 in Hfinish.
              
              destruct (zlt
              (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                 (ts + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                 (ts + T - Ti)) Ci); try discriminate.

              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }
  Qed.    
    
  Lemma count_occurrence_leftover_bounded_aux:
    forall Nnat N0 id ts T conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence2_aux Nnat N0 N0
                           ts (ts + T) conf id (ts+T) Sched_leftover <=
      workload_leftover_total Nnat id ts (ts+T) conf.
  Proof.
    intros * HNnat HN0 Hts HT Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_leftover_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H3: Tf <= T).
        {
          apply Hlarge with (i:=Z.of_nat Nnat) (Cn:=Cf); auto.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        
        destruct (is_leftover (Z.of_nat Nnat) Tf id ts (ts+T)) eqn:Hleftover.
        {
          assert (H: forall a b c d, a <= d -> b <= c -> a + b <= c + d). intros. omega.
          apply H; auto.
          remember (Z.to_nat T) as nT.
          assert (HH: ts + T = ts + Z.of_nat nT).
          { rewrite HeqnT. rewrite Z2Nat.id; omega. }
          rewrite HH.
          rewrite count_single_task_self_sched_leftover_bounded_by_budget_aux with (Ti:=Tf) (Ci:=Cf); auto; try omega.
          - apply Z.le_min_l.
          - rewrite HeqnT. rewrite Z2Nat.id; auto; omega.
        }        
        {
          assert (H: count_single_task_type_occurrence N0 N0 (Z.of_nat Nnat) ts 
                                                       (ts + T) conf id (ts+T) Sched_leftover = 0).
          {
            apply count_single_task_self_sched_leftover_is_0_when_not_leftover with (Ti:=Tf) (Ci:=Cf); auto; try omega.
          }
          rewrite H.
          omega.
        }
      }
    }      
  Qed.
  
  Lemma count_occurrence_leftover_bounded:
    forall Nnat id ts T conf,
      0 < Z.of_nat Nnat ->
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           ts (ts + T) conf id (ts+T) Sched_leftover <=
      workload_leftover_total Nnat id ts (ts+T) conf.
  Proof.
    intros * HN Hts HT Hconfvalid Hlarge.
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    rewrite count_occurrence_leftover_bounded_aux; auto; try omega. 
  Qed.

End SCHED_LEFTOVER.


Section Sched_interference.

  Lemma count_single_task_interference_value_in_one_period_aux:
    forall N ts conf toid i Ti Ci nrange todeadline,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      ts mod Ti = 0 ->
      Z.of_nat nrange <= Ti ->
      priEntry_lt (PriEntry (ts + Ti) i) (PriEntry todeadline toid) = true ->
      count_single_task_type_occurrence N N i ts (ts + Z.of_nat nrange) conf toid todeadline Sched_interference =
      Z.min Ci (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
  Proof.
    intros * Hts Hconf HC Hi Hibound Hmod Hrange Hpri.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.    
    induction nrange.
    {
      simpl.
      replace (ts + 0) with ts by omega.
      rewrite Z.min_r; omega.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H1.      
      simpl.
      assert (HNN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.
      assert (Hmono: monotonic_continuous (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf))).
      {
        apply TAC_pool_all_monotonic_continuous; auto; try xomega.
      }

      assert (H2: (ts + Z.of_nat nrange) / Ti * Ti = ts).
      {
        rewrite divide_plus_equal; try omega.
        rewrite mod_0_divide_mul; omega.
      }
      
      rewrite IHnrange; try xomega.

      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) i) as [Hsched|Hother].
      {
        rewrite Hsched.
        rewrite Hconf.
        
        assert (Htype: get_sched_type i Ti toid (ts + Z.of_nat nrange) todeadline = Sched_interference).
        {
          unfold get_sched_type.
          rewrite zeq_false; auto.
          rewrite H2.
          rewrite Hpri.
          auto.
        }
        rewrite Htype.
        rewrite sched_type_eq_true.

        assert (H10: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
        {
          apply vt_increase_for_the_scheduled_task; auto; try omega.
          rewrite <- HNN.
          auto.
        }
        rewrite H10.

        assert (H11: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts < Ci).
        {
          assert (H12: vt_map_not_finished (ts + Z.of_nat nrange) Ti Ci (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)) = true).
          {
            apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:= conf); auto; try omega.
            rewrite <- HNN.
            auto.
          }
          unfold vt_map_not_finished in H12.
          rewrite divide_plus_equal in H12; try omega.
          rewrite mod_0_divide_mul in H12; try omega.

          assert (H14: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts = ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
          { rewrite vertical_map_canonical; omega. }
          rewrite H14 in H12.
          
          destruct (zlt
                      (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
                       ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts) Ci); try omega; try discriminate. 
        }

        rewrite Z.min_r; try omega.
        rewrite Z.min_r; try omega.
      }
      {
        (* i not scheduled *)
        destruct (vt_map_not_finished (ts + Z.of_nat nrange) Ti Ci (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf))) eqn:Hfinish.
        {
          (* i is ready, vt stagnates *)
          assert (Hvt: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange)).
          {            
            apply vt_stagnates_if_task_not_scheduled with (T:=Ti) (C:=Ci); auto; try omega.
            - rewrite <- HNN.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=Ti) (C:=Ci); auto.
            - rewrite <- HNN.
              auto.
          }
          rewrite Hvt.
          auto.
        }
        {
          (* i finished *)
          assert (H10: 0 <= ts + Z.of_nat nrange). omega.
          pose proof (monotonic_vt_step_bound _ _ _ conf H10 Hibound) as H11.
          
          unfold vt_map_not_finished in Hfinish.
          rewrite divide_plus_equal in Hfinish; try omega.
          rewrite mod_0_divide_mul in Hfinish; try omega.
          assert (H12: ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts = ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
          {
            rewrite vertical_map_canonical; auto; omega.
          }
          rewrite H12 in Hfinish.

          destruct (zlt
              (ZMap.get i (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts) Ci); try discriminate.
          repeat rewrite Z.min_l; try omega.
        }
      }
    }
  Qed.
  
  Lemma count_single_task_interference_bounded_by_budget_in_one_period:
    forall N ts conf toid i Ti Ci todeadline,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->
      i <> toid ->
      ts mod Ti = 0 ->
      priEntry_lt (PriEntry (ts + Ti) i) (PriEntry todeadline toid) = true ->
      count_single_task_type_occurrence N N i ts (ts + Ti) conf toid todeadline Sched_interference <=
      Ci.
  Proof.
    intros * Hts Hconf HTf HCf Hid Hidbound Hmod Hpri.
    remember (Z.to_nat Ti) as nT.
    assert (H1: Ti = Z.of_nat nT).
    { rewrite HeqnT. rewrite Z2Nat.id; omega. }
    rewrite H1.
    
    rewrite count_single_task_interference_value_in_one_period_aux with (Ti:=Ti) (Ci:=Ci); auto; try omega.
    apply Z.le_min_l.
  Qed.    

  Lemma count_single_task_interference_bounded_by_workload_subtotal_aux1_aux:
    forall N ts conf toid n i Ti Ci todeadline,
      ts < ts + Z.of_nat n * Ti <= todeadline ->
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      ts mod Ti = 0 ->
      i <> toid ->
      priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true ->
      count_single_task_type_occurrence N N i ts (ts + Ti * Z.of_nat n) conf toid todeadline Sched_interference <= Ci * Z.of_nat n.
  Proof.
    intros * Htod Hts Hconf HT HC Hi Hmod Hto cHpri.
    induction n.
    {
      simpl in Htod.
      omega.
    }
    {
      assert (H1: Z.of_nat (S n) = Z.of_nat n + 1). xomega.
      assert (H2: ts + Ti * Z.of_nat (S n) = ts + Ti * Z.of_nat n + Ti). lia.
      rewrite H2 in *. clear H2.

      destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
      {
        rewrite H1 in *.
        rewrite Heq in *.
        replace (ts + Ti * 0) with ts in * by omega.
        replace (Ci * (0 + 1)) with Ci by nia.
        apply count_single_task_interference_bounded_by_budget_in_one_period; auto.
      }
      {                
        assert (H3: ts <= ts + Ti * Z.of_nat n <= ts + Ti * Z.of_nat n + Ti). nia.
        pose proof (count_single_task_type_occurrence_composition N N i _ _ _ conf toid todeadline Sched_interference H3) as H4.
        rewrite <- H4. clear H4.

        replace (Ci * Z.of_nat (S n)) with (Ci * Z.of_nat n + Ci) by nia.

        assert (H5: ts < ts + Z.of_nat n * Ti <= todeadline). nia.
        assert (H6: priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true).
        {
          unfold priEntry_lt.
          rewrite zlt_true; auto.
          rewrite H1 in Htod.         
          nia.
        }
        specialize (IHn H5 H6).
        assert (H: forall a b c d, a <= c -> b <= d -> a + b <= c + d). intros. omega.
        apply H; auto.

        apply count_single_task_interference_bounded_by_budget_in_one_period; auto; try omega.
        - rewrite Z.mul_comm.
          rewrite Z_mod_plus_full.
          auto.
      }
    }
  Qed.

  Lemma count_single_task_interference_bounded_by_workload_subtotal_aux1:
    forall N ts conf toid n i Ti Ci todeadline,
      ts + Z.of_nat n * Ti <= todeadline ->
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      ts mod Ti = 0 ->
      i <> toid ->
      priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true ->
      count_single_task_type_occurrence N N i ts (ts + Ti * Z.of_nat n) conf toid todeadline Sched_interference <= Ci * Z.of_nat n.
  Proof.
    intros.
    destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
    {
      rewrite Heq.
      replace (ts + Ti * 0) with ts by omega.
      rewrite count_single_task_type_occurance_0_when_empty_duration.
      omega.
    }
    {
      apply count_single_task_interference_bounded_by_workload_subtotal_aux1_aux; auto.
      nia.
    }
  Qed.

  Lemma count_single_task_interference_0_aux2_aux:
    forall N ts conf toid i Ti Ci todeadline nrange,
      0 <= ts ->
      0 < Ti ->
      Z.of_nat nrange <= Ti ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      ts mod Ti = 0 ->
      priEntry_lt (PriEntry (ts + Ti) i) (PriEntry todeadline toid) = false ->
      count_single_task_type_occurrence N N i ts 
                                        (ts + Z.of_nat nrange) conf toid todeadline Sched_interference = 0.
  Proof.
    intros * Hts HT Hrange Hconf Hmod Hpri.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) i); auto.
      rewrite e.

      rewrite Hconf.
      rewrite sched_type_eq_false; auto.
      
      unfold not.
      intros.
      unfold get_sched_type in H.
      destruct (zeq i toid); try congruence.
      rewrite divide_plus_equal in H; try xomega.
      rewrite mod_0_divide_mul in H; try omega.
      rewrite Hpri in H.
      destruct (zeq (ts + Ti) todeadline); congruence.
    }
  Qed.

  Lemma count_single_task_interference_0_aux2:
    forall N ts te conf toid i Ti Ci todeadline,
      0 <= ts ->
      0 < Ti ->
      ts <= te <= ts + Ti ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      ts mod Ti = 0 ->
      priEntry_lt (PriEntry (ts + Ti) i) (PriEntry todeadline toid) = false ->
      count_single_task_type_occurrence N N i ts 
                                        te conf toid todeadline Sched_interference = 0.
  Proof.
    intros.
    remember (Z.to_nat (te - ts)) as nrange.
    assert (Hte: te = ts + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; xomega.
    }
    rewrite Hte.

    apply count_single_task_interference_0_aux2_aux with (Ti:=Ti) (Ci:=Ci); auto.
    xomega.
  Qed.
    
  Lemma count_single_task_interference_bounded_by_workload_subtotal_start_aligned:
    forall N ts conf toid T i Ti Ci,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti <= T ->
      0 < Ci ->
      0 <= i < N ->      
      i <> toid ->
      ts mod Ti = 0 ->
      count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference <=
      workload_whole toid i Ti Ci ts (ts + T).
  Proof.
    intros * Hts Hconf HT HC Hibound Hi Hmod.
    unfold workload_whole.
    rewrite zlt_false; try omega.
    rewrite Hmod.
    rewrite zeq_true.
    destruct (zeq ((ts + T) mod Ti) 0) as [Heq | Hneq].
    {
      assert (H1: 0 < T / Ti).
      {
        apply div_pos; omega.
      }
      remember (Z.to_nat (T / Ti)) as n.
      assert (H2: T / Ti = Z.of_nat n).
      {
        rewrite Heqn. rewrite Z2Nat.id; omega.
      }
      assert (H3: T = Ti * Z.of_nat n).
      {
        rewrite <- H2.
        rewrite Z.mul_comm.
        rewrite mod_0_divide_mul; try omega.
        rewrite <- Z.add_mod_idemp_l in Heq; try omega.
        rewrite Hmod in Heq.
        replace (0 + T) with T in Heq by omega.
        auto.
      }
      
      destruct (zlt i toid) as [Hlt|Hge].
      {
        rewrite mod_0_divide_mul; try omega.
        replace (ts + T - ts) with T by omega.
        rewrite H2. rewrite Z.mul_comm.
        rewrite H3.
        apply count_single_task_interference_bounded_by_workload_subtotal_aux1; auto; try omega.
        - nia.
        - unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_true.
          rewrite zlt_true; auto.          
      }
      {
        replace (ts + T - Ti - ts) with ((Z.of_nat n - 1) * Ti) by nia.
        rewrite Z_div_mult_full; try omega.
        assert (H10: count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i ts (ts + T - Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i (ts + T - Ti) (ts + T) conf toid (ts + T) Sched_interference).
        {
          symmetry.
          apply count_single_task_type_occurrence_composition; try omega.
        }
        rewrite H10.
        remember (pred n) as n'.
        assert (H11: n = S n'). xomega.
        assert (H12: Z.of_nat n' = Z.of_nat n - 1). omega.
        assert (H14: T - Ti = Z.of_nat n' * Ti). nia.
        replace (ts + T - Ti) with (ts + Ti * Z.of_nat n') by nia.
        rewrite <- H12.

        assert (H: forall a b c, a <= c -> b = 0 -> a + b <= c). intros. omega.
        apply H.
        - replace (Z.of_nat n' * Ci) with (Ci * Z.of_nat n') by lia.
          apply count_single_task_interference_bounded_by_workload_subtotal_aux1; auto; try omega.
          + unfold priEntry_lt.
            rewrite zlt_true; auto.
            nia.
        - apply count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try nia.
          + rewrite Zplus_mod.
            rewrite Hmod.
            rewrite Z.mul_comm.
            rewrite Z_mod_mult.
            simpl.
            rewrite Z.mod_0_l; omega.
          + unfold priEntry_lt.
            assert (H15: ts + Ti * Z.of_nat n' + Ti = ts + T). nia.
            rewrite H15.
            rewrite zlt_false; try omega.
            rewrite zeq_true.
            rewrite zlt_false; auto.            
      }
    }
    {
      assert (H1: ts <= (ts + T) / Ti * Ti).
      {
        apply mod_not_0_adjust_to_smaller_period; omega.
      }
      assert (H2: (ts + T) / Ti * Ti <= ts + T).
      {
        apply divide_mul_le; omega.
      }
      assert (H5: ((ts + T) / Ti * Ti) mod Ti = 0).
      {
        rewrite Z_mod_mult. auto.
      }
      assert (H6: ts + T < (ts + T) / Ti * Ti + Ti).
      { apply x_lt_floor_plus; omega. }
      
      assert (H3: count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i ts ((ts + T) / Ti * Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i ((ts + T) / Ti * Ti) (ts + T) conf toid (ts + T) Sched_interference).
      {
        symmetry.
        apply count_single_task_type_occurrence_composition; try omega.
      }
      rewrite H3. clear H3.

      assert (H4: count_single_task_type_occurrence N N i ((ts + T) / Ti * Ti) 
                                                    (ts + T) conf toid (ts + T) Sched_interference = 0).
      {
        apply count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try nia.
        - unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_false; try omega.
          auto.
      }
      rewrite H4.
      rewrite Z.add_0_r.

      assert (H7: 0 <= ((ts + T) / Ti * Ti - ts) / Ti).
      {
        apply Z_div_pos; omega.
      }
      
      remember (Z.to_nat (((ts + T) / Ti * Ti - ts) / Ti)) as n.
      assert (H8: ((ts + T) / Ti * Ti - ts) / Ti = Z.of_nat n).
      { rewrite Heqn. rewrite Z2Nat.id; omega. }
      rewrite H8.
      replace (Z.of_nat n * Ci) with (Ci * Z.of_nat n) by nia.

      assert (H9: (ts + T) / Ti * Ti = ts + Ti * Z.of_nat n).
      {
        replace (Ti * Z.of_nat n) with (Z.of_nat n * Ti) by nia.
        rewrite <- H8.
        symmetry.
        rewrite mod_0_divide_mul; try omega.
        rewrite Zminus_mod.
        rewrite Hmod, H5.
        simpl.
        rewrite Z.mod_0_l; omega.
      }
      rewrite H9.

      assert (H10: ts + T - (ts + T) / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      
      assert (H11: ts + Z.of_nat n * Ti < ts + T). nia.
        
      apply count_single_task_interference_bounded_by_workload_subtotal_aux1; auto; try omega.
      - unfold priEntry_lt.
        rewrite zlt_true; try omega.
        auto.
    }
  Qed.

  Lemma count_single_task_interference_bounded_by_workload_subtotal_start_aligned_generalized:
    forall N ts te conf toid i Ti Ci,
      0 <= ts <= te ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      i <> toid ->
      ts mod Ti = 0 ->
      count_single_task_type_occurrence N N i ts te conf toid te Sched_interference <=
      workload_whole toid i Ti Ci ts te.
  Proof.
    intros * Hts Hconf HT HC Hibound Hi Hmod.
    
    destruct (zlt (te - ts) Ti) as [Htesmall | Htelarge].
    {
      unfold workload_whole.
      rewrite zlt_true; auto.
      rewrite count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try omega.
      - unfold priEntry_lt.
        rewrite zlt_false; try omega.
        rewrite zeq_false; try omega.
        auto.
    }
    {
      assert (H1: te = ts + (te - ts)). omega.
      rewrite H1.
      apply count_single_task_interference_bounded_by_workload_subtotal_start_aligned; auto; try omega.
    }
  Qed.

  Lemma count_single_task_interference_remain_aux3:
    forall N i Ti Ci ts toid todeadline conf,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      0 < Ti ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      ts mod Ti <> 0 ->
      ts / Ti * Ti + Ti < todeadline ->
      count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference <=
      workload_remain toid i Ti Ci ts todeadline (ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf)).
  Proof.
    intros * Hts Hconf HT HC Hi Hibound Hmod Hdeadline.
    assert (H1: ts / Ti * Ti < ts).
    {
      assert (ts - ts / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
    
    assert (H2: ts < ts / Ti * Ti + Ti).
    {
      apply x_lt_floor_plus; omega.
    }
      
    assert (H3: 0 <= ts / Ti * Ti).
    {
      apply Z.mul_nonneg_nonneg; try omega.
      apply Z_div_pos; omega.
    }
      
    assert (H4: (ts / Ti * Ti) mod Ti = 0).
    {
      apply Z_mod_mult.
    }
      
    assert (H5: priEntry_lt (PriEntry (ts / Ti * Ti + Ti) i) (PriEntry todeadline toid) = true).
    {
      unfold priEntry_lt.
      rewrite zlt_true; auto.
    }
    
    assert (H: count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference + count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference = count_single_task_type_occurrence N N i (ts / Ti * Ti) (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference).
    {
        apply count_single_task_type_occurrence_composition; try omega.
    }

    replace (count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference) with (count_single_task_type_occurrence N N i (ts / Ti * Ti) 
                                                                                                                                                             (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference - count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference) by omega.
    clear H.

    remember (Z.to_nat Ti) as nT.
    assert (H10: ts / Ti * Ti + Ti = ts / Ti * Ti + Z.of_nat nT).
    {
      rewrite HeqnT.
      rewrite Z2Nat.id; omega.
    }
    rewrite H10.
    assert (Hc1: count_single_task_type_occurrence N N i (ts / Ti * Ti) 
                                                   (ts / Ti * Ti + Z.of_nat nT) conf toid todeadline Sched_interference = Z.min Ci (ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Z.of_nat nT) (Z.to_nat N) conf) (ts / Ti * Ti + Z.of_nat nT) - ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti))).
    {
      apply count_single_task_interference_value_in_one_period_aux with (Ti:=Ti); auto; try omega.
    }
    rewrite Hc1. clear Hc1.

    remember (Z.to_nat (ts - ts / Ti * Ti)) as nT2.
    assert (H11: ts = ts / Ti * Ti + Z.of_nat nT2).
    {
      rewrite HeqnT2.
      rewrite Z2Nat.id; omega.
    }
    assert (Hc2: count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference = Z.min Ci (ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Z.of_nat nT2) (Z.to_nat N) conf) (ts / Ti * Ti + Z.of_nat nT2) - ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti))).
    {
      remember (ts / Ti * Ti) as k.
      rewrite H11.
      rewrite Heqk in *. clear Heqk.
      apply count_single_task_interference_value_in_one_period_aux with (Ti:=Ti); auto; try omega.
    }
    rewrite Hc2. clear Hc2.

    rewrite <- H10.
    rewrite <- H11.
    
    unfold workload_remain.
    rewrite zeq_false; auto.
    rewrite H5.

    assert (Hc3: ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf) ts = ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hc3. clear Hc3.
    
    assert (Hc4: ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf) (ts / Ti * Ti) = ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hc4. clear Hc4.

    assert (Hvt1: ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts <= ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Ti) (Z.to_nat N) conf) (ts / Ti * Ti + Ti)).
    {
      apply TAC_pool_at_t_monotonic_two_points; try omega.
    }

    destruct (zle (ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts -
                   ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)) Ci).
    {
      rewrite Z.max_r; try omega.
      assert (Hv2: Z.min Ci
     (ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts -
      ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)) = ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts -
                                                                                    ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)).
      {
        rewrite Z.min_r; omega.
      }
      rewrite Hv2. clear Hv2.
      
      assert (Z.min Ci
     (ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Ti) (Z.to_nat N) conf) (ts / Ti * Ti + Ti) -
      ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)) <= Ci).
      {
        apply Z.le_min_l.
      }
      omega.
    }
    {
      rewrite Z.min_l; try omega.
      rewrite Z.min_l; try omega.
      rewrite Z.max_l; try omega.
    }
  Qed.
        
  Lemma count_single_task_interference_bounded_by_workload_subtotal:
    forall N ts conf toid T i Ti Ci,
      0 < N ->
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti <= T ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts+T) Sched_interference <=
      workload_subtotal toid i Ti Ci ts (ts + T)
                        (ZMap.get i (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)).
  Proof.
    intros * HN Hts Hconf HTf HCf Hid Hibound.
    destruct (zeq (ts mod Ti) 0) as [Hmod|Hmod].
    {
      unfold workload_subtotal.
      rewrite workload_remain_0_start; auto.

      rewrite Z.add_0_l.
      apply count_single_task_interference_bounded_by_workload_subtotal_start_aligned; auto; try omega.
    }
    {
      unfold workload_subtotal.
      assert (H1: ts < ts / Ti * Ti + Ti).
      {
        apply x_lt_floor_plus; omega.
      }
      assert (H2: ts - ts / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      
      assert (H3: count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i (ts / Ti * Ti + Ti) (ts + T) conf toid (ts + T) Sched_interference).
      {
        symmetry.
        apply count_single_task_type_occurrence_composition; try omega.
      }
      rewrite H3. clear H3.

      assert (H4: count_single_task_type_occurrence N N i (ts / Ti * Ti + Ti) 
     (ts + T) conf toid (ts + T) Sched_interference <= workload_whole toid i Ti Ci (ts / Ti * Ti + Ti) (ts + T)).
      {
        apply count_single_task_interference_bounded_by_workload_subtotal_start_aligned_generalized; auto; try omega.
        - rewrite Z.add_comm.
          rewrite Z_mod_plus_full.
          apply Z_mod_same_full.
      }

      rewrite workload_whole_left_align; try omega.

      assert (forall a b c d, b <= d -> a <= c -> a + b <= c + d). intros. omega.
      apply H; auto.
      
      assert (H5: ts / Ti * Ti + Ti < ts + T). omega.
      apply count_single_task_interference_remain_aux3; auto; omega.
    }    
  Qed.

  Lemma count_occurrence_inteference_bounded_aux:
    forall Nnat N0 id ts T conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence2_aux Nnat N0 N0
                           ts (ts + T) conf id (ts+T) Sched_interference <=
      workload_total2 Nnat id ts (ts+T) conf (TAC_pool_at_t (ts + T) (Z.to_nat N0) conf).
  Proof.
    intros * HNnat HN0 Hts HT Hconfvalid Hlarge.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.

      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.        
        rewrite count_single_task_self_sched_others_is_0; auto; try omega.
        congruence.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H3: Tf <= T).
        {
          apply Hlarge with (i:=Z.of_nat Nnat) (Cn:=Cf); auto.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }

        assert (H: forall a b c d, a <= d -> b <= c -> a + b <= c + d). intros. omega.
        apply H; auto.

        apply count_single_task_interference_bounded_by_workload_subtotal; auto.
        - omega.
      }
    }      
  Qed.
  
  Lemma count_occurrence_interference_bounded:
    forall Nnat id ts T conf,
      0 < Z.of_nat Nnat ->
      0 <= ts ->
      1 < T ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                            ts (ts + T) conf id (ts+T) Sched_interference <=
      workload_total2 Nnat id ts (ts + T) conf
                      (TAC_pool_at_t (ts + T) Nnat conf).
  Proof.
    intros * HN Hts HT Hconfvalid Hlarge.
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    assert (H2: TAC_pool_at_t (ts + T) Nnat conf = TAC_pool_at_t (ts + T) (Z.to_nat (Z.of_nat Nnat)) conf).
    { rewrite H1. auto. }
    rewrite H2.
    
    apply count_occurrence_inteference_bounded_aux; auto; try omega.     
  Qed.
  
End Sched_interference.  


Section BUSY_SCHEDULE.

  Theorem interference_is_bounded_when_schedule_is_busy_single_period:
    forall Nnat conf id T C np,
      1 < T ->
      0 < C ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn : Z, 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= T) ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      last_idle_slot (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) 
            (Z.of_nat np * T + T) conf = None ->
      workload_overlapping_total Nnat id (Z.of_nat np * T) (Z.of_nat np * T + T) (Z.of_nat np * T + T + T) conf -
      workload_remain_total Nnat id (Z.of_nat (S np) * T) 
                            (Z.of_nat (S np) * T + T) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * T + T) Nnat conf) >=
      T - C -
      workload_total1 Nnat id (Z.of_nat np * T) 
                      (Z.of_nat np * T + T) conf (TAC_pool_at_t (Z.of_nat np * T + T) Nnat conf) -
      workload_leftover_total Nnat id (Z.of_nat np * T) (Z.of_nat np * T + T) conf.
  Proof.
    intros * Hk HT Hconfvalid Hlarge Hid Hconfid Hidle.
    assert (Hts: 0 <= Z.of_nat np * T).
    { apply Zmult_gt_0_le_0_compat; xomega. }
    assert (Htsmod: (Z.of_nat np * T) mod T = 0).
    { rewrite Z_mod_mult; auto. }
    assert (Hcount_total: count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_self + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_interference + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_leftover + count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_overlap = T).
    {
      apply sched_type_is_complete_when_schedule_is_busy with (C := C); auto.
    }
    assert (Hself: count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) (Z.of_nat np * T) 
                   (Z.of_nat np * T + T) conf id 
                   (Z.of_nat np * T + T) Sched_self <= C).
    {
      apply count_occurrence_self_bounded_by_budget_single_period; auto.
    }
    assert (Hinterference: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (Z.of_nat np * T) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_interference <=
           workload_total1 Nnat id (Z.of_nat np * T) (Z.of_nat np * T + T) conf
                           (TAC_pool_at_t (Z.of_nat np * T + T) Nnat conf)).
    {
      rewrite workload_total_eq.
      apply count_occurrence_interference_bounded; auto; omega.
    }
      
    assert (Hleftover: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (Z.of_nat np * T) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_leftover <=
                       workload_leftover_total Nnat id (Z.of_nat np * T) (Z.of_nat np * T + T) conf).
    {
      apply count_occurrence_leftover_bounded; auto.
      omega.
    }
    assert (Hoverlap: workload_overlapping_total Nnat id (Z.of_nat np * T) (Z.of_nat np * T + T) (Z.of_nat np * T + T + T) conf -
   workload_remain_total Nnat id (Z.of_nat (S np) * T) 
                         (Z.of_nat (S np) * T + T) conf (TAC_pool_at_t (Z.of_nat (S np) * T + T) Nnat conf) =
           count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) (Z.of_nat np * T) 
                   (Z.of_nat np * T + T) conf id (Z.of_nat np * T + T) Sched_overlap).
    {
      assert (H: Z.of_nat (S np) = Z.of_nat np + 1). xomega.
      rewrite H.
      replace (Z.of_nat np * T + T) with ((Z.of_nat np + 1) * T) by nia.
      rewrite workload_overlapping_remain_total_eq; auto; try xomega.
      symmetry.
      
      assert (H1: (Z.of_nat np + 1) * T = Z.of_nat np * T + T). lia.
      rewrite H1.

      rewrite count_occurrence_overlapping_bounded; auto; omega.
    }
    omega.
  Qed.    

End BUSY_SCHEDULE.


