Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import LemmaVTMap.


Open Scope Z_scope.

Section VERTICALINTERFERENCE.

  (* TAC is valid up to time t *)
  (* queue is already updated to time t *)
  (* returns 1 if task id receives interference at [t, t+1). 0 otherwise *)
  Fixpoint interference_at_t (t id: Z) (TAC: TimeContainerPool) (conf: PrioConfigPool)
           (queue: priQueue) :=
    match queue with
    | nil => 0
    | (PriEntry deadline tid) :: l =>
      if (zeq id tid) then 0 else
        match (ZMap.get tid conf) with
        | mkPrioConfigValid T C =>
          if (vt_map_not_finished t T C (ZMap.get tid TAC)) then 1 else
            interference_at_t t id TAC conf l
      end
    end.

  Lemma interference_at_t_res:
    forall t id TAC conf queue,
      interference_at_t t id TAC conf queue = 0 \/
      interference_at_t t id TAC conf queue = 1.
  Proof.
    induction queue.
    {
      simpl.
      auto.
    }
    {
      simpl.
      destruct a.
      destruct (zeq id z0) as [Heq | Hneq]; auto.
      destruct (ZMap.get z0 conf).
      destruct (vt_map_not_finished t z1 z2 (ZMap.get z0 TAC)); auto.
    }
  Qed.
      
  Lemma interference_1_highest_not_finished:
    forall t Nnat id T0 C0 sched T C TAC conf queue,
      0 <= t ->
      0 <= id < Z.of_nat Nnat ->
      priqueue_sound (Z.of_nat Nnat) t conf queue ->
      In (PriEntry (t / T0 * T0 + T0) id) queue ->
      interference_at_t t id TAC conf queue = 1 ->
      sched = highest_not_finished_map (Z.of_nat Nnat) t TAC conf queue ->
      ZMap.get sched conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      (0 <= sched < Z.of_nat Nnat /\
      priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry (t / T0 * T0 + T0) id) = true).
  Proof.
    intros *.
    induction queue.
    {
      simpl.
      omega.
    }
    {
      intros Ht Hid Hq Hidin HI Hschedval Hconfsched Hconfid.
      simpl in HI.
      destruct a as [hd hid].
      destruct (zeq id hid); try omega.
      simpl in Hschedval.
      destruct Hq as [Hqsort Hqrange].      
      destruct (ZMap.get hid conf) as [T1 C1] eqn:Hhid.
      destruct (vt_map_not_finished t T1 C1 (ZMap.get hid TAC)) as [Hnot | Hfinished].
      {
        rewrite Hschedval in *. clear Hschedval.
        assert (H1: In (PriEntry hd hid) (PriEntry hd hid :: queue)).
        { constructor; auto. }
        simpl in Hidin.
        assert (Hin': In (PriEntry (t / T0 * T0 + T0) id) queue).
        {
          destruct Hidin; auto.
          inv H.
          omega.
        }
        specialize (Hqrange _ _ _ _ H1 Hhid).
        destruct Hqrange as [Hidbound Hdval].
        split; auto.
        rewrite Hdval in *. clear Hdval.
        rewrite Hconfsched in Hhid.
        inv Hhid.
        inversion Hqsort.
        assert (H4: forall x, In x queue -> is_true (priEntry_le (PriEntry (t / T1 * T1 + T1) hid) x)).
        { apply Forall_forall; auto. }
        unfold is_true in H4.
        assert (H5: priEntry_le (PriEntry (t / T1 * T1 + T1) hid) (PriEntry (t / T0 * T0 + T0) id) = true).
        { apply H4. auto. }
        rewrite priEntry_id_diff_lt in H5; auto.
      }
      {
        apply IHqueue; auto.
        - apply priqueue_sound_shrink with (a := PriEntry hd hid).
          unfold priqueue_sound.
          split; auto.
        - inv Hidin; auto.
          inv H; omega.
      }
    }          
  Qed.
    
  Fixpoint interference_at_range (ts: Z) (range: nat) (id: Z) (Nnat: nat) (conf: PrioConfigPool) :=
    match range with
    | O => 0
    | S n => let t := ts + Z.of_nat n in
               interference_at_t t id (TAC_pool_at_t t Nnat conf) conf 
                                 (real_priqueue Nnat t conf) + interference_at_range ts n id Nnat conf
    end.

  Lemma interference_at_range_step:
    forall t nr id Nnat conf,
      0 <= t ->
      interference_at_range t (S nr) id Nnat conf =
      interference_at_range t nr id Nnat conf +
      interference_at_range (t + Z.of_nat nr) 1%nat id Nnat conf.
  Proof.
    intros * Ht.
    remember (interference_at_range (t + Z.of_nat nr) 1 id Nnat conf) as i1.
    simpl.
    assert (i1 = interference_at_t (t + Z.of_nat nr) id (TAC_pool_at_t (t + Z.of_nat nr) Nnat conf) conf
                                   (real_priqueue Nnat (t + Z.of_nat nr) conf)).
    {
      subst.
      simpl.
      replace (t + Z.of_nat nr + 0) with (t + Z.of_nat nr) by omega.
      omega.
    }
    omega.
  Qed.

  Lemma vertical_map_equal:
    forall x y t id Nnat conf,
      0 <= t ->
      t <= x ->
      t <= y ->
      ZMap.get id (TAC_pool_at_t x Nnat conf) t =
      ZMap.get id (TAC_pool_at_t y Nnat conf) t.
  Proof.
    intros.
    rewrite vertical_map_canonical; auto.
    symmetry.
    rewrite vertical_map_canonical; auto.    
  Qed.

  Section INTERFERENCE_1_SCHEDULE.

    Lemma interference_1_if_id_is_blocked_queue:
      forall id did t N conf TAC queue,
        In (PriEntry did id) queue ->
        priqueue_sound N t conf queue ->        
        id_is_blocked_queue id did t conf TAC queue ->
        interference_at_t t id TAC conf queue = 1.
    Proof.
      intros * Hin Hsound Hblock.
      induction queue.
      {
        simpl in Hin.
        omega.
      }
      {
        destruct a as [dj j].
        simpl.
        destruct (zeq id j) as [Heq | Hneq].
        {
          rewrite Heq in *.
          assert (Hdid: did = dj).
          {
            eapply priqueue_sound_equal_id_equal_entry; eauto.
            simpl.
            auto.
          }
          rewrite Hdid in *.
          exfalso.
          eapply id_is_blocked_queue_self_false; eauto.
        }
        {
          destruct (ZMap.get j conf) as [Tj Cj] eqn:Hconfj.
          destruct (vt_map_not_finished t Tj Cj (ZMap.get j TAC)) eqn:Hfinish; auto.
          apply IHqueue; auto.
          - simpl in Hin.
            destruct Hin; try congruence.
          - eapply priqueue_sound_shrink; eauto.
          - eapply id_is_blocked_queue_shrink; eauto.
        }
      }
    Qed.

    Lemma interference_1_if_id_is_blocked:
      forall id t Nnat conf,
        0 <= id < Z.of_nat Nnat ->
        id_is_blocked id t (Z.of_nat Nnat) conf (TAC_pool_at_t t Nnat conf) ->        
        interference_at_t t id (TAC_pool_at_t t Nnat conf) conf
                          (real_priqueue Nnat t conf) = 1.
    Proof.
      intros * Hid Hblock.
      destruct (ZMap.get id conf) as [T C] eqn:Hconfid.
      apply interference_1_if_id_is_blocked_queue with (did:=t/T*T+T) (N:=Z.of_nat Nnat).
      - apply real_priqueue_complete with (C:=C); auto.
      - apply real_priqueue_sound.
      - apply id_is_blocked_implies_queue with (C:=C); auto.
    Qed.    
    
  End INTERFERENCE_1_SCHEDULE.

  Section INTERFERENCE0.

    Lemma interference_0_if_id_not_blocked_queue:
      forall id did t N conf TAC queue,
        In (PriEntry did id) queue ->
        priqueue_sound N t conf queue ->        
        id_not_blocked_queue id did t conf TAC queue ->
        interference_at_t t id TAC conf queue = 0.
    Proof.
      intros * Hin Hsound Hblock.
      induction queue.
      {
        simpl.
        auto.
      }
      {
        destruct a as [dj j].
        simpl.
        destruct (zeq id j); auto.
        destruct (ZMap.get j conf) as [T C] eqn:Hconfj.

        assert (Hfinish: vt_map_not_finished t T C (ZMap.get j TAC) = false).
        {
          unfold id_not_blocked_queue in Hblock.
          assert (H1: j <> id). omega.
          assert (H2: In (PriEntry dj j) (PriEntry dj j :: queue)).
          {
            simpl.
            auto.
          }
          specialize (Hblock _ _ _ _ H1 Hconfj H2).

          assert (Hpri: priEntry_lt (PriEntry did id) (PriEntry dj j) = false).
          {
            apply priEntry_lt_neg_tf.
            rewrite <- priEntry_id_diff_lt; auto.
            destruct Hsound as [Hsort _].
            inv Hsort.
            rewrite Forall_forall in H4.
            apply H4.
            simpl in Hin.
            destruct Hin; auto; congruence.
          }
          destruct Hblock; try congruence.
        }
        rewrite Hfinish.

        apply IHqueue.
        - simpl in Hin.
          destruct Hin; auto; congruence.
        - eapply priqueue_sound_shrink; eauto.
        - eapply id_not_blocked_queue_shrink; eauto.
      }
    Qed.
    
    Lemma interference_0_if_id_not_blocked:
      forall id t Nnat conf,
        0 <= id < Z.of_nat Nnat ->
        id_not_blocked id t (Z.of_nat Nnat) conf (TAC_pool_at_t t Nnat conf) ->        
        interference_at_t t id (TAC_pool_at_t t Nnat conf) conf
                          (real_priqueue Nnat t conf) = 0.
    Proof.
      intros * Hid Hblock.
      destruct (ZMap.get id conf) as [T C] eqn:Hconfid.
      apply interference_0_if_id_not_blocked_queue with (did:=t/T*T+T) (N:=Z.of_nat Nnat).
      - apply real_priqueue_complete with (C:=C); auto.
      - apply real_priqueue_sound.
      - apply id_not_blocked_implies_queue with (C:=C); auto.
    Qed.
    
  End INTERFERENCE0.
  
  Lemma interference_and_vt_step:
    forall t id Nnat conf,
      0 <= id < Z.of_nat Nnat ->
      0 <= t ->
      (ZMap.get id (TAC_pool_at_t (t + 1) Nnat conf)) (t + 1) =
      (ZMap.get id (TAC_pool_at_t t Nnat conf)) t + 1 -
      interference_at_range t 1%nat id Nnat conf.
  Proof.
    intros * Hid Ht.

    assert (Hmono: (ZMap.get id (TAC_pool_at_t t Nnat conf)) t <=
            (ZMap.get id (TAC_pool_at_t (t + 1) Nnat conf)) (t + 1) <=
            (ZMap.get id (TAC_pool_at_t t Nnat conf)) t + 1).
    {
      remember (Z.of_nat Nnat) as N.
      assert (HNN: Nnat = Z.to_nat N).
      {
        rewrite HeqN.
        rewrite Nat2Z.id; auto.
      }
      rewrite HNN.
      apply monotonic_vt_step_bound; try omega.
    }
    
    unfold TAC_pool_at_t in *.    
    remember (Z.to_nat t) as nt.
    assert (Hnt: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    clear Heqnt.
    subst.

    assert (HSt: Z.to_nat (Z.of_nat nt + 1) = S nt).
    {
      pose proof (Nat2Z.inj_succ nt) as H1.
      replace (Z.of_nat nt + 1) with (Z.succ (Z.of_nat nt)) by omega.
      rewrite <- H1.
      rewrite Nat2Z.id; omega.
    }
    rewrite HSt in *.
    simpl. simpl in Hmono.
    repeat rewrite Z.add_0_r.
    
    destruct (zeq (ZMap.get id
     (time_map_tick (Z.of_nat nt) (TAC_pool_at_t_aux nt Nnat conf) conf
        (real_priqueue Nnat (Z.of_nat nt) conf)) (Z.of_nat nt + 1))
                  (ZMap.get id (TAC_pool_at_t_aux nt Nnat conf) (Z.of_nat nt))) as [Hstagnates | Hstep].
    {
      assert (Hblock: id_is_blocked id (Z.of_nat nt) (Z.of_nat Nnat) conf (TAC_pool_at_t (Z.of_nat nt) Nnat conf)).
      {
        apply vt_stagnates_id_is_blocked; auto; try omega.
        - unfold TAC_pool_at_t.
          rewrite Nat2Z.id.
          apply TAC_pool_at_t_aux_constant_starting_t.
        - unfold TAC_pool_at_t.
          repeat rewrite Nat2Z.id.
          auto.
      }
      
      rewrite interference_1_if_id_is_blocked; auto.
      omega.
    }
    {
      assert (Hincr: ZMap.get id
     (time_map_tick (Z.of_nat nt) (TAC_pool_at_t_aux nt Nnat conf) conf
                    (real_priqueue Nnat (Z.of_nat nt) conf)) (Z.of_nat nt + 1)  =
                     ZMap.get id (TAC_pool_at_t_aux nt Nnat conf) (Z.of_nat nt) + 1).
      omega.
      rewrite Hincr.

      assert (Hblock: id_not_blocked id (Z.of_nat nt) (Z.of_nat Nnat) conf (TAC_pool_at_t (Z.of_nat nt) Nnat conf)).
      {
        apply vt_increases_id_not_blocked; auto; try omega.
        - unfold TAC_pool_at_t.
          rewrite Nat2Z.id.
          apply TAC_pool_at_t_aux_constant_starting_t.
        - unfold TAC_pool_at_t.
          repeat rewrite Nat2Z.id.
          auto.
      }

      rewrite interference_0_if_id_not_blocked; auto.
      omega.
    }
  Qed.
  
  Lemma interference_creates_hole_in_vt:
    forall t nrange id Nnat conf,
      0 <= t ->
      0 <= id < Z.of_nat Nnat ->
      (ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange) Nnat conf)) (t + Z.of_nat nrange) -
      (ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange) Nnat conf)) t +
      interference_at_range t nrange id Nnat conf = Z.of_nat nrange.
  Proof.
    intros * Hid Ht.
    induction nrange.
    {
      simpl.
      replace (t + 0) with t by omega.
      omega.
    }
    {
      assert (HSt: t + Z.of_nat (S nrange) = (t + Z.of_nat nrange) + 1).
      {
        rewrite Nat2Z.inj_succ; omega.
      }
      rewrite HSt.      
      assert (Htnneg: 0 <= t + Z.of_nat nrange).
      {
        pose proof (Zle_0_nat nrange).
        omega.
      }
      rewrite interference_and_vt_step; auto.
      replace (ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange) Nnat conf) (t + Z.of_nat nrange)) with (Z.of_nat nrange + ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange) Nnat conf) t -
                                                                                                                          interference_at_range t nrange id Nnat conf) by omega.
      assert (Hi: interference_at_range t (S nrange) id Nnat conf =
                  interference_at_range t nrange id Nnat conf +
                  interference_at_range (t + Z.of_nat nrange) 1 id Nnat conf).
      {
        apply interference_at_range_step; auto.
      }
      assert (Hvt: ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange + 1) Nnat conf) t =
                   ZMap.get id (TAC_pool_at_t (t + Z.of_nat nrange) Nnat conf) t).
      {
        apply vertical_map_equal; auto; try omega.
      }
      omega.
    }
  Qed.
  
End VERTICALINTERFERENCE.

Section WORKLOAD.

  (* vt should be valid up to ts *)
  Function workload_remain (toid fromid T C ts te: Z) (vt: vtmap) :Z :=
    if (zeq (ts mod T) 0) then 0
      else if (priEntry_lt (PriEntry (ts / T * T + T) fromid)
                      (PriEntry te toid)) then
        Z.max 0 (C - vt ts + vt (ts / T * T))
           else 0.

  Lemma workload_remain_le_C:
    forall toid fromid T C ts te vt,
      monotonic_continuous vt ->
      0 <= ts ->
      1 < T ->
      0 < C ->
      workload_remain toid fromid T C ts te vt <= C.
  Proof.
    intros * Hvt Hts HT HC.
    unfold workload_remain.
    destruct (zeq (ts mod T) 0); try omega.
    destruct (priEntry_lt (PriEntry (ts / T * T + T) fromid) (PriEntry te toid)); try omega.
    assert (H1: ts / T * T <= ts).
    {
      apply divide_mul_le; try omega.
    }
    assert (Z.max 0 (C - vt ts + vt (ts / T * T)) < C + 1).
    {
      apply Z.max_lub_lt; try omega.
      assert (vt (ts / T * T) <= vt ts).
      {
        apply monotonic_continuous_monotonic; auto.
        apply Nfloor_le; omega.
      }
      omega.
    }
    omega.
  Qed.
    
  Lemma workload_remain_0_start:
    forall toid fromid T C ts te vt,
      ts mod T = 0 ->
      workload_remain toid fromid T C ts te vt = 0.
  Proof.
    intros * Htmod.
    unfold workload_remain.
    rewrite Htmod.
    rewrite zeq_true; auto.
  Qed.

  Lemma workload_remain_nonneg:
    forall toid fromid T C ts te vt,
      0 <= workload_remain toid fromid T C ts te vt.
  Proof.
    intros.
    unfold workload_remain.
    destruct (zeq (ts mod T) 0); try omega.
    destruct (priEntry_lt (PriEntry (ts / T * T + T) fromid) (PriEntry te toid)); try omega.
    apply Z.le_max_l.
  Qed.

  Lemma workload_remain_nondec_left_mod_pos:
    forall toid fromid T C t te vt,
      0 <= t ->
      1 < T ->
      t mod T <> 0 ->
      monotonic_continuous vt ->
      workload_remain toid fromid T C (t + 1) te vt <=
      workload_remain toid fromid T C t te vt.
  Proof.
    intros * Ht HT Hmod Hvt.
    unfold workload_remain at 1.
    destruct (zeq ((t + 1) mod T) 0) as [Hmod2 | Hmod2].
    { 
      apply workload_remain_nonneg.
    }
    {
      unfold workload_remain.
      rewrite zeq_false; auto.
      assert (H1: (t + 1) / T * T = t / T * T).
      { rewrite mod_equal; omega. }
      rewrite H1.
      destruct (priEntry_lt (PriEntry (t / T * T + T) fromid) (PriEntry te toid));
       try omega.
      assert (vt t <= vt (t + 1)).
      { apply monotonic_continuous_monotonic; auto; omega. }
      destruct (zle 0 (C - vt (t + 1) + vt (t / T * T))).
      - repeat rewrite Z.max_r; omega.
      - rewrite Z.max_l; try omega.
        apply Z.le_max_l.
    }
  Qed.
        
  Lemma workload_remain_empty_window_0:
    forall toid fromid T C t vt,
      0 < T ->
      workload_remain toid fromid T C t t vt = 0.
  Proof.
    intros * HT.
    unfold workload_remain.
    destruct (zeq (t mod T) 0) as [Hmod | Hmod]; try omega.
    unfold priEntry_lt.
    assert (H: t / T * T + T > t).
    {
      assert (H1: T > 0). omega.
      pose proof (Z_div_mod_eq t T H1) as H2.
      rewrite Z.mul_comm in H2.
      pose proof (Z.mod_pos_bound t T HT).
      omega.
    }
    rewrite zlt_false; try omega.
    rewrite zeq_false; try omega.
  Qed.
    
  Function workload_whole (toid fromid T C ts te: Z) : Z :=
    if (zlt (te - ts) T) then 0 else
      let a := if (zeq (ts mod T) 0) then ts else
                 ts / T * T + T in
      let e := if (zeq (te mod T) 0) then
                 (if (zlt fromid toid) then te / T * T else te - T)
               else te / T * T in
      (e - a) / T * C.

  Lemma workload_whole_empty_window_0:
    forall toid fromid T C t,
      0 < T ->
      workload_whole toid fromid T C t t = 0.
  Proof.
    intros * HT.
    unfold workload_whole.
    rewrite zlt_true; try omega.
  Qed.

  Lemma workload_whole_nonneg:
    forall i id T C t1 t2,
      0 <= t1 ->
      1 < T ->
      0 < C ->
      0 <= workload_whole i id T C t1 t2.
  Proof.
    intros * Ht HT HC.
    unfold workload_whole.
    destruct (zlt (t2 - t1) T); try omega.

    assert (H1: t2 / T * T - T <= (if zeq (t2 mod T) 0 then if zlt id i then t2 / T * T else t2 - T else t2 / T * T)).
    {
      destruct (zeq (t2 mod T) 0); try omega.
      rewrite mod_0_divide_mul; try omega.
      destruct (zlt id i); omega.
    }
    
    destruct (zeq (t1 mod T) 0) as [Hmod | Hmod].
    {
      apply div_mul_nonneg; try omega.
      assert (H10: (t1 + T) / T * T <= t2 / T * T).
      {
        apply div_mul_le; omega.
      }
      rewrite mod_0_divide_mul in H10; try omega.
      rewrite mod_0_preserve; omega.
    }
    {
      apply div_mul_nonneg; try omega.      
      assert (H10: (t1+T) / T * T <= t2 / T * T).
      {
        apply div_mul_le; omega.
      }
      rewrite mod_not_0_divide_mul in H10; try omega.
      destruct (zeq (t2 mod T) 0); try omega.
      {
        rewrite mod_0_divide_mul; try omega.
        assert (H11: t2 - T <= (if zlt id i then t2 else t2 - T)).
        {
          destruct (zlt id i); omega.
        }
        assert (H12: t1 / T * T + T <= t2 - T).
        {
          apply mod_0_bound_increase; try omega.
          - destruct (zeq t2 (t1 + T)); try omega.
            rewrite e0 in e.
            assert ((t1 + T) mod T <> 0).
            apply mod_not_0_preserve; try omega.
            congruence.
        }

        omega.
      }
    }
  Qed.
  
  Lemma workload_whole_nondec_left:
    forall toid fromid T C t te,
      0 <= t ->
      1 < T ->
      0 < C ->
      t + 1 <= te ->
      workload_whole toid fromid T C (t + 1) te <=
      workload_whole toid fromid T C t te.      
  Proof.
    intros * H HT HC Ht.
    unfold workload_whole.
    destruct (zlt (te - t) T) as [HT1 | HT2].
    {
      (* te - t < T *)
      rewrite zlt_true; try omega.
    }
    {
      destruct (zeq (te - t) T) as [Heq | Hneq].
      {
        (* te - t = T *)
        rewrite zlt_true; try omega.
        replace te with (t + T) by omega.
        destruct (zeq (t mod T) 0) as [Hmod0 | Hmod1].
        {          
          assert (H1: (t + T) mod T = 0).
          {
            apply mod_0_preserve; omega.
          }
          rewrite H1.
          rewrite zeq_true; auto.
          rewrite mod_0_divide_mul; try omega.          
          assert (H2: 0 <= (if zlt fromid toid then t + T else t + T - T) - t).
          {
            destruct (zlt fromid toid); omega.
          }
          apply Zmult_gt_0_le_0_compat; try omega.
          apply Z_div_pos; try omega.
        }
        {
          rewrite mod_not_0_divide_mul; try omega.          
          assert (H1: (t + T) mod T <> 0).
          {
            apply mod_not_0_preserve; try omega.
          }
          rewrite zeq_false; try omega.
          replace (t / T * T + T - (t / T * T + T)) with 0 by omega.          
          rewrite Zdiv_0_l.
          omega.
        }
      }
      {
        rewrite zlt_false; try omega.
        assert (H0: (if zeq ((t + 1) mod T) 0 then t + 1 else (t + 1) / T * T + T) <= (if zeq (te mod T) 0 then if zlt fromid toid then te / T * T else te - T else te / T * T)).
        {
          destruct (zeq (te mod T) 0) as [Hmod0 | Hmod1].
          {
            transitivity (te - T).
            - destruct (zeq ((t + 1) mod T) 0) as [Hm0 | Hm1].
              + omega.
              + rewrite <- z_divide_same; try omega.
                apply mod_0_bound_increase; try omega.
            - rewrite mod_0_divide_mul; try omega.
              destruct (zlt fromid toid); omega.
          }
          {
            destruct (zeq ((t + 1) mod T) 0) as [Hm0 | Hm1].
            - rewrite <- mod_0_divide_mul with (y := T) at 1; try omega.
              apply div_mul_le; omega.
            - rewrite <- z_divide_same; try omega.
              transitivity ((t + T) / T * T).
              + transitivity ((t / T + 1) * T).
                * rewrite Z.mul_add_distr_r.
                  omega.
                * apply Z.mul_le_mono_pos_r; try omega.
                  rewrite div_T_inc_1; try omega.
              + apply div_mul_le; omega.
          }
        }            
        assert (H1: (if zeq ((t + 1) mod T) 0 then t + 1 else (t + 1) / T * T + T) >=
                    (if zeq (t mod T) 0 then t else t / T * T + T)).
        {
          destruct (zeq (t mod T) 0) as [Hmod0 | Hmod1].
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; try omega.            
            rewrite zeq_false; try omega.
            apply mod_0_not_0; try omega.
          }
          {
            destruct (zeq ((t + 1) mod T) 0) as [Hmod3 | Hmod4].
            - rewrite mod_0_minus_1; try omega.
            - rewrite mod_equal; try omega.
          }
        }
        assert (H2: (if zeq (te mod T) 0 then if zlt fromid toid then te / T * T else te - T else te / T * T) -
                    (if zeq ((t + 1) mod T) 0 then t + 1 else (t + 1) / T * T + T) <=
               (if zeq (te mod T) 0 then if zlt fromid toid then te / T * T else te - T else te / T * T) -
               (if zeq (t mod T) 0 then t else t / T * T + T)). omega.        
        rewrite Z.mul_comm.
        replace (((if zeq (te mod T) 0 then if zlt fromid toid then te / T * T else te - T else te / T * T) -
    (if zeq (t mod T) 0 then t else t / T * T + T)) / T * C) with (C * (((if zeq (te mod T) 0 then if zlt fromid toid then te / T * T else te - T else te / T * T) -
                                                                         (if zeq (t mod T) 0 then t else t / T * T + T)) / T)) by (apply Z.mul_comm).
        apply Zmult_le_compat_l; try omega.
        apply Z.div_le_mono; try omega.
      }
    }
  Qed.

  Lemma workload_whole_left_align:
    forall toid i T C ts te,
      0 <= ts ->
      0 < T ->
      ts / T * T + T <= te ->
      ts mod T <> 0 ->
      workload_whole toid i T C ts te =
      workload_whole toid i T C (ts / T * T + T) te.
  Proof.
    intros * Hts HT Hte Hmod.
    unfold workload_whole.

    assert (H1: ts - ts / T * T >= 1).
    {
      apply z_gt_mod; try omega.
    }

    assert (H2: (if zeq (ts mod T) 0 then ts else ts / T * T + T) = ts / T * T + T).
    { rewrite zeq_false; omega. }
    rewrite H2. clear H2.

    assert (H3: (ts / T * T + T) mod T = 0).
    {
      rewrite Z.add_comm.
      rewrite Z_mod_plus_full.
      rewrite Z_mod_same_full.
      auto.
    }
    assert (H4: (if zeq ((ts / T * T + T) mod T) 0 then ts / T * T + T else (ts / T * T + T) / T * T + T) = (ts / T * T + T)).
    { rewrite H3. rewrite zeq_true; auto. }
    rewrite H4. clear H4.

    assert (H5: ts < ts / T * T + T).
    {
      apply x_lt_floor_plus; omega.
    }
    
    destruct (zlt (te - ts) T).
    {
      rewrite zlt_true; try omega.
    }
    {      
      destruct (zeq (te mod T) 0) as [Hmode | Hmode].
      {
        assert (H6: ts / T * T + T <= te - T).
        {
          apply mod_0_bound_increase; try omega.
          - destruct (zlt ts (te - T)); auto.
            assert (H: ts = te - T). omega.
            rewrite H in Hmod.
            rewrite Zminus_mod in Hmod.
            rewrite Hmode in Hmod.
            rewrite Z_mod_same_full in Hmod.
            simpl in Hmod.
            rewrite Zmod_0_l in Hmod.
            omega.
        }
        destruct (zlt (te - (ts / T * T + T)) T); try omega.
      }
      {
        destruct (zlt (te - (ts / T * T + T)) T); auto.
        {
          assert (te / T * T = ts / T * T + T).
          {
            apply mod_lower_bound; try omega.
          }
          rewrite H.
          replace (ts / T * T + T - (ts / T * T + T)) with 0 by omega.
          rewrite Zdiv_0_l.
          nia.
        }
      }
    }
  Qed.
  
  Function workload_subtotal (toid fromid T C ts te: Z) (vt: vtmap) :Z :=  
    workload_remain toid fromid T C ts te vt +
    workload_whole toid fromid T C ts te.

  Lemma workload_subtotal_empty_window_0:
    forall toid fromid T C t vt,
      0 < T ->
      workload_subtotal toid fromid T C t t vt = 0.
  Proof.
    intros.
    unfold workload_subtotal.
    rewrite workload_remain_empty_window_0; auto.
    rewrite workload_whole_empty_window_0; auto.
  Qed.
    
  Lemma workload_subtotal_nondec_left:
    forall t te toid fromid T C vt,
      0 <= t ->
      1 < T ->
      0 < C ->
      t + 1 <= te ->
      monotonic_continuous vt ->
      workload_subtotal toid fromid T C (t + 1) te vt <=
      workload_subtotal toid fromid T C t te vt.      
  Proof.
    intros * Htpos HT HC Ht Hvt.
    unfold workload_subtotal.
    destruct (zeq (t mod T) 0) as [Htmod | Htmod].
    {
      assert (Hwr: workload_remain toid fromid T C t te vt = 0).
      {
        unfold workload_remain.
        rewrite Htmod.
        rewrite zeq_true; auto.
      }
      rewrite Hwr.
      rewrite Z.add_0_l.
      assert (H1: (t + 1) mod T = 1).
      {
        rewrite Zplus_mod.
        rewrite Htmod.
        repeat rewrite Zmod_1_l; try omega.
      }
      assert (H2: (t + 1) / T * T = t).
      {
        rewrite divide_plus_equal; try omega.
        rewrite mod_0_divide_mul; omega.
      }
      unfold workload_whole.
      destruct (zlt (te - t) T) as [HTlarge | HTsmall].
      {
        rewrite zlt_true; try omega.
        unfold workload_remain.
        rewrite zeq_false; try omega.
        rewrite H2.
        assert (Hpri: priEntry_lt (PriEntry (t + T) fromid) (PriEntry te toid) = false).
        {
          unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_false; try omega.
          auto.
        }
        rewrite Hpri.
        omega.
      }
      {
        destruct (zeq te (t + T)) as [HtetplusT | Htelarger].
        {
          rewrite zlt_true; try omega.
          assert (H3: te mod T = 0).
          {
            rewrite HtetplusT.
            rewrite Zplus_mod.
            rewrite Htmod.
            rewrite Z_mod_same_full.
            auto.
          }
          rewrite H3, Htmod.
          repeat rewrite zeq_true.
          unfold workload_remain.
          rewrite zeq_false; try omega.
          unfold priEntry_lt.
          rewrite H2.
          rewrite zlt_false; try omega.
          rewrite HtetplusT.
          rewrite zeq_true.
          destruct (zlt fromid toid).
          - assert ((t + T) mod T = 0). apply mod_0_preserve; omega.
            rewrite mod_0_divide_mul; try omega.
            replace (t + T - t) with T by omega.
            rewrite Z_div_same_full; try omega.
            destruct (zle 0 (C - vt (t + 1) + vt t)).
            + unfold monotonic_continuous in Hvt.
              assert (Htt: t >= 0). omega.
              specialize (Hvt _ Htt).
              destruct Hvt as [H4 | H4].
              * rewrite H4. rewrite Z.max_r; omega.
              * rewrite H4. rewrite Z.max_r; omega.
            + rewrite Z.max_l; try omega.
          - replace (t + T - T - t) with 0 by omega.
            rewrite Zdiv_0_l.
            omega.
        }
        {
          rewrite zlt_false; try omega.
          rewrite H1, Htmod.
          assert (Hwr2: workload_remain toid fromid T C (t + 1) te vt <= C).
          {
            apply workload_remain_le_C; auto. omega.
          }
          destruct (zeq (te mod T) 0) as [Hmod1 | Hmod1].
          {
            assert (H3: t / T * T + T <= te - T).
            {
              apply mod_0_bound_increase; omega.
            }
            rewrite zeq_false; try omega.
            rewrite zeq_true.
            rewrite mod_0_divide_mul; try omega.
            rewrite H2.
            destruct (zlt fromid toid).
            {
              assert (H4: (te - t) / T * C = (te - (t + T)) / T * C + C).
              {
                replace (te - t) with (T + (te - (t + T))) by omega.
                rewrite Nadd_div_exact; try omega.
                rewrite Z_div_same_full; try omega.
                rewrite Z.mul_add_distr_r; omega.
                apply Z_mod_same_full.
              }
              omega.
            }
            {
              replace ((te - T - t) / T * C) with
                  ((te - T - (t + T)) / T * C + C); try omega.
              replace (te - T - t) with (T + (te - T - (t + T))) by omega.
              symmetry.
              rewrite mod_0_divide_mul in H3; try omega.
              rewrite Nadd_div_exact; try omega.
              rewrite Z_div_same_full; try omega.
              rewrite Z.mul_add_distr_r; omega.
              apply Z_mod_same_full.
            }
          }
          {
            rewrite H2.
            rewrite zeq_false; try omega.
            rewrite zeq_true.
            replace ((te / T * T - t) / T * C) with
                ((te / T * T - (t + T)) / T * C + C); try omega.
            symmetry.
            replace (te / T * T - t) with (T + (te / T * T - (t + T))) by omega.
            rewrite Nadd_div_exact; try omega.
            - rewrite Z_div_same_full; try omega.
              rewrite Z.mul_add_distr_r; omega.
            - assert ((t + T) / T * T <= te / T * T).
              {
                apply Z.mul_le_mono_nonneg_r; try omega.
                apply Z.div_le_mono; omega.
              }
              rewrite mod_0_divide_mul in H; try omega.
              apply mod_0_preserve; omega.
            - apply Z_mod_same_full.
          }
        }
      }
    }
    {
      (* t mod T <> 0 *)
      destruct (zeq ((t + 1) mod T) 0) as [Ht1mod | Ht1mod].
      {
        rewrite workload_remain_0_start; auto.
        rewrite Z.add_0_l.
        assert (workload_whole toid fromid T C (t + 1) te <=
                workload_whole toid fromid T C t te).
        {
          apply workload_whole_nondec_left; try omega.
        }
        assert (0 <= workload_remain toid fromid T C t te vt).
        { apply workload_remain_nonneg. }
        omega.
      }
      {
        (* (t + 1) mod T <> 0 *)
        assert (H1: (t + 1) / T * T = t / T * T).
        { rewrite mod_equal; omega. }          
        (* both decreases *)
        assert (workload_whole toid fromid T C (t + 1) te <=
                workload_whole toid fromid T C t te).
        { apply workload_whole_nondec_left; omega. }
        assert (workload_remain toid fromid T C (t + 1) te vt <=
                workload_remain toid fromid T C t te vt).
        { apply workload_remain_nondec_left_mod_pos; auto; omega. }
        omega.
      }
    }        
  Qed.
  
  (* Interference from all other tasks *)
  Fixpoint workload_remain_total (Nnat: nat) (toid ts te: Z)
           (conf: PrioConfigPool) (TAC: TimeContainerPool) : Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_remain_total n toid ts te conf TAC else
        let vt := ZMap.get (Z.of_nat n) TAC in
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid T C =>
          workload_remain toid (Z.of_nat n) T C ts te vt +
          workload_remain_total n toid ts te conf TAC
        end
    end.  

  Fixpoint workload_whole_total (Nnat: nat) (toid ts te: Z)
           (conf: PrioConfigPool) : Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_whole_total n toid ts te conf else
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid T C =>
          workload_whole toid (Z.of_nat n) T C ts te +
          workload_whole_total n toid ts te conf
        end
    end.  

  Function workload_total1 N toid ts te conf TAC :=
    workload_remain_total N toid ts te conf TAC +
    workload_whole_total N toid ts te conf.

  Fixpoint workload_total2 (Nnat: nat) (toid ts te: Z)
           (conf: PrioConfigPool) (TAC: TimeContainerPool): Z :=
    match Nnat with
    | O => 0
    | S n =>
      if (zeq toid (Z.of_nat n)) then
        workload_total2 n toid ts te conf TAC else
        match (ZMap.get (Z.of_nat n) conf) with
        | mkPrioConfigValid T C =>
          workload_subtotal toid (Z.of_nat n) T C ts te (ZMap.get (Z.of_nat n) TAC) +
          workload_total2 n toid ts te conf TAC
        end
    end.

  Lemma workload_remain_total_nonneg:
    forall Nnat toid ts te TAC conf,
      0 <= workload_remain_total Nnat toid ts te conf TAC.
  Proof.
    intros.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      simpl.
      destruct (zeq toid (Z.of_nat Nnat)); auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf).
      assert (0 <= workload_remain toid (Z.of_nat Nnat) z z0 ts te (ZMap.get (Z.of_nat Nnat) TAC)).
      {
        apply workload_remain_nonneg.
      }
      omega.
    }
  Qed.
  
  Lemma workload_remain_total_canonical_aux:
    forall Nnat Nnat0 toid ts te conf t,
      Z.of_nat Nnat <= Z.of_nat Nnat0 ->
      0 <= ts <= t ->
      config_pool_valid (Z.of_nat Nnat0) conf ->
      workload_remain_total Nnat toid ts te conf (TAC_pool_at_t t Nnat0 conf) =
      workload_remain_total Nnat toid ts te conf (TAC_pool_at_t ts Nnat0 conf).      
  Proof.
    intros * HN Ht Hconfvalid.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nnat <= Z.of_nat Nnat0). xomega.      
      rewrite (IHNnat H1).
      destruct (zeq toid (Z.of_nat Nnat)); auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:Hconf.
      assert (HT: 1 < T).
      {
        assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat Nnat0). xomega.
        specialize (Hconfvalid _ _ _ H2 Hconf).
        omega.
      }
      
      unfold workload_remain.
      destruct (zeq (ts mod T) 0); auto.
      destruct (priEntry_lt (PriEntry (ts / T * T + T) (Z.of_nat Nnat)) (PriEntry te toid)); auto.
      
      rewrite vertical_map_canonical; try omega.
      replace (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t t Nnat0 conf) (ts / T * T)) with
          (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts Nnat0 conf) (ts / T * T)); auto.

      assert (H3: 0 <= ts / T * T).
      {
        apply div_mul_nonneg; omega.
      }
      assert (H4: ts / T * T <= ts).
      {
        apply Nfloor_le; omega.
      }
      
      rewrite vertical_map_canonical; try omega.
      symmetry.
      rewrite vertical_map_canonical; try omega.      
    }
  Qed.

  Lemma workload_remain_total_canonical:
    forall Nnat toid ts te conf t,
      0 <= ts <= t ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_remain_total Nnat toid ts te conf (TAC_pool_at_t t Nnat conf) =
      workload_remain_total Nnat toid ts te conf (TAC_pool_at_t ts Nnat conf).      
  Proof.
    intros.
    apply workload_remain_total_canonical_aux; auto; omega.
  Qed.
  
  Lemma workload_remain_total_0_at_0:
    forall Nnat toid t conf TAC,
      workload_remain_total Nnat toid 0 t conf TAC = 0.
  Proof.
    intros.
    induction Nnat; auto.
    {
      simpl.
      rewrite IHNnat.
      destruct (zeq toid (Z.of_nat Nnat)); auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf); auto.
    }
  Qed.    
  
  Lemma workload_total2_empty_window_0:
    forall Nnat toid t conf TAC,
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_total2 Nnat toid t t conf TAC = 0.
  Proof.
    intro Nnat.
    induction Nnat; auto.
    intros * Hconf.
    simpl.
    assert (Hlt: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)).
    { rewrite Nat2Z.inj_succ; omega. }
    assert (Hconf0: config_pool_valid (Z.of_nat Nnat) conf).
    { eapply config_pool_valid_shrink; eauto. omega. }
    specialize (IHNnat toid t conf TAC Hconf0).
    rewrite IHNnat.
    destruct (zeq toid (Z.of_nat Nnat)); auto.
    destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:HTC.
    rewrite workload_subtotal_empty_window_0; try omega.
    specialize (Hconf _ _ _ Hlt HTC).
    omega.
  Qed.
      
  Lemma workload_total_eq:
    forall N toid ts te conf TAC,
      workload_total1 N toid ts te conf TAC = workload_total2 N toid ts te conf TAC.
  Proof.
    intros N.
    induction N.
    {
      unfold workload_total1.
      simpl.
      auto.
    }
    {
      intros.
      unfold workload_total1.
      simpl.
      destruct (zeq toid (Z.of_nat N)) as [Heq | Hneq].
      {
        apply IHN.
      }
      {
        destruct (ZMap.get (Z.of_nat N) conf).
        unfold workload_subtotal.
        rewrite <- IHN.
        unfold workload_total1.
        omega.
      }
    }
  Qed.

  Lemma add_le_distr:
    forall a b m n,
      a <= b ->
      m <= n ->
      a + m <= b + n.
  Proof.
    intros.
    omega.
  Qed.
  
  Lemma workload_total1_nondec_left:
    forall t te id Nnat conf TAC,
      0 <= t ->
      t + 1 <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i, 0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i TAC)) ->
      workload_total1 Nnat id (t + 1) te conf TAC <=
      workload_total1 Nnat id t te conf TAC.
  Proof.
    intros t te id Nnat conf TAC Ht0 Ht.
    induction Nnat.
    {
      intros.
      repeat rewrite workload_total_eq.
      simpl.
      omega.
    }
    {
      intros * Hconf HTAC.
      repeat rewrite workload_total_eq.
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)).
      { rewrite Nat2Z.inj_succ; omega. }
      assert (H2: forall i : Z,
                 0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i TAC)).
      {
        intros.
        apply HTAC; auto.
        omega.
      }
      assert (Hconf0: config_pool_valid (Z.of_nat Nnat) conf).
      {
        apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto.
        omega.
      }
      specialize (IHNnat Hconf0 H2).
      repeat rewrite workload_total_eq in IHNnat.      
      simpl.      
      destruct (zeq id (Z.of_nat Nnat)); auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:Hparam.
      assert (H3: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
      specialize (Hconf _ _ _ H3 Hparam).
      apply add_le_distr; try omega.
      apply workload_subtotal_nondec_left; auto; omega.
    }
  Qed.  

End WORKLOAD.


Section INTERFERENCE.
(*  
  Function f t :=
    interference_at_range 0 (Z.to_nat t) 2 3%nat real_prioconfig_pool +
    workload_total1 3%nat 2 t 10 real_prioconfig_pool (TAC_pool_at_t 10 3%nat real_prioconfig_pool). 
  
  Eval compute in (output_f_range 0 11 f).

(*
     = (0, 7)
       :: (1, 7)
          :: (2, 7)
             :: (3, 7) :: (4, 7) :: (5, 7) :: (6, 7) :: (7, 7) :: (8, 7) :: (9, 7) :: (10, 7) :: nil
     : list (Z * Z)
*)
*)

  Lemma workload_total2_le_history_preserve:
    forall Nnat id t te conf TAC nsched k,
      0 <= t ->
      t + 1 <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i, 0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i TAC)) ->
      workload_total2 nsched id (t + 1) te conf TAC + k <=
      workload_total2 nsched id t te conf TAC ->
      Z.of_nat nsched <= Z.of_nat Nnat ->      
      workload_total2 Nnat id (t + 1) te conf TAC + k <=
      workload_total2 Nnat id t te conf TAC.      
  Proof.
    intros * Ht Hte Hconf HTAC Hle.
    induction Nnat.
    {
      simpl.
      intros.
      assert (H1: Z.of_nat nsched = 0). omega.
      assert (H2: Z.to_nat (Z.of_nat nsched) = Z.to_nat 0). rewrite H1; auto.
      rewrite Nat2Z.id in H2.
      simpl in H2.
      rewrite H2 in Hle.
      simpl in Hle.
      omega.
    }
    {
      intros Hnsched.
      destruct (zeq (Z.of_nat nsched) (Z.of_nat (S Nnat))) as [Heq | Hneq].
      {
        assert (H1: Z.to_nat (Z.of_nat nsched) = Z.to_nat (Z.of_nat (S Nnat))).
        { rewrite Heq. auto. }
        repeat rewrite Nat2Z.id in H1.
        rewrite <- H1.
        auto.
      }
      {
        assert (H1: Z.of_nat nsched <= Z.of_nat Nnat).
        { rewrite Nat2Z.inj_succ in *. omega. }
        assert (H2: Z.of_nat (Nnat) < Z.of_nat (S Nnat)).
        { rewrite Nat2Z.inj_succ in *. omega. }        
        assert (Hconf': config_pool_valid (Z.of_nat Nnat) conf).
        {
          apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto.
          omega.
        }
        assert (HTAC': forall i : Z,
                   0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i TAC)).
        {
          intros.
          apply HTAC; omega.
        }
        specialize (IHNnat Hconf' HTAC' H1).
        simpl.
        destruct (zeq id (Z.of_nat Nnat)); auto.
        destruct (ZMap.get (Z.of_nat Nnat) conf) eqn:Hparam.
        assert (workload_subtotal id (Z.of_nat Nnat) z z0 (t + 1) te (ZMap.get (Z.of_nat Nnat) TAC) <= workload_subtotal id (Z.of_nat Nnat) z z0 t te (ZMap.get (Z.of_nat Nnat) TAC)).
        {
          assert (H3: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
          specialize (Hconf _ _ _ H3 Hparam).
          apply workload_subtotal_nondec_left; try omega.
          apply HTAC; omega.
        }
        rewrite Zplus_assoc_reverse.
        apply add_le_distr; auto.
      }
    }
  Qed.

  Lemma workload_whole_1_less_period:
    forall t te sched T C id T0 C0 conf,
      1 < T ->
      1 < T0 ->
      0 <= t <= te - 1 ->
      te mod T0 = 0 ->
      ZMap.get sched conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry (t / T0 * T0 + T0) id) = true ->
      t mod T = 0 ->
      workload_whole id sched T C (t + 1) te + C = workload_whole id sched T C t te.
  Proof.
    intros * HT HT0 Ht Htemod Hconfsched Hconfid Hpri Htmod.
    unfold workload_whole.
    assert (H1: t / T * T = t).
    { rewrite mod_0_divide_mul; omega. }
    rewrite H1 in *.
    assert (H2: t / T0 * T0 + T0 <= te).
    {
      assert (t / T0 * T0 + T0 <= te - 0 * T0).
      { apply mod_0_bound_increase_k; omega. }
      omega.
    }
    assert (H3: t + T <= t / T0 * T0 + T0).
    { eapply priEntry_lt_deadline_le; eauto. }
    assert (H4: t + T <= te). omega.
    assert (H5: (t + T) mod T = 0).
    { apply mod_0_preserve; omega. }
    assert (H6: (t + T + 1) mod T = 1).
    {
      rewrite Zplus_mod.
      rewrite mod_0_preserve; try omega.
      rewrite Zmod_1_l; try omega.
      simpl.
      rewrite Zmod_1_l; omega.
    }
    assert (H7: (t + 1) mod T = 1).
    {
      rewrite Zplus_mod.
      rewrite Htmod.
      rewrite Zmod_1_l; try omega.
      simpl.
      rewrite Zmod_1_l; omega.
    }
    
    destruct (zeq te (t + T)) as [Hte1 | Hte2].
    {
      rewrite Hte1 in *.
      rewrite zlt_true; try omega.
      rewrite zlt_false; try omega.
      rewrite H5.
      rewrite zeq_true.
      assert (H8: t / T0 * T0 + T0 = t + T). omega.
      rewrite H8 in *.
      unfold priEntry_lt in Hpri.
      rewrite zlt_false in Hpri; try omega.
      rewrite zeq_true in Hpri.
      destruct (zlt sched id); try congruence.
      rewrite Htmod.
      rewrite zeq_true.
      rewrite mod_0_divide_mul; try omega.
      replace (t + T - t) with T by omega.
      rewrite Z_div_same_full; try omega.      
    }
    {
      destruct (zeq te (t + T + 1)) as [Hte3 | Hte4].
      {
        rewrite Hte3 in *.
        rewrite zlt_false; try omega.
        rewrite H6.
        rewrite zeq_false; try omega.
        rewrite zeq_false; try omega.
        rewrite zlt_false; try omega.
        rewrite Htmod.
        rewrite zeq_true.
        repeat rewrite <- z_divide_same; try omega.
        repeat rewrite mod_0_divide_mul; try omega.
        replace (t + T - (t + T)) with 0 by omega.
        replace (t + T - t) with T by omega.
        simpl.        
        rewrite Z_div_same_full; omega.
      }
      {
        rewrite zlt_false; try omega.
        assert (H10: (if zeq ((t + 1) mod T) 0 then t + 1 else (t + 1) / T * T + T) = t + T).
        {
          rewrite zeq_false; try omega.
          rewrite <- z_divide_same; omega.
        }
        rewrite H10.
        symmetry.
        rewrite zlt_false; try omega.
        assert (H11: (if zeq (t mod T) 0 then t else t + T) = t).
        {
          rewrite Htmod.
          rewrite zeq_true; auto.
        }
        rewrite H11.        
        replace ((if zeq (te mod T) 0 then if zlt sched id then te / T * T else te - T else te / T * T) - t) with (T + ((if zeq (te mod T) 0 then if zlt sched id then te / T * T else te - T else te / T * T) - (t + T))) by omega.
        rewrite Nadd_div_exact; try omega.
        - rewrite Z_div_same_full; try omega.
          rewrite Z.mul_add_distr_r; omega.
        - destruct (zeq (te mod T) 0) as [Hmod2 | Hmod2].
          + rewrite mod_0_divide_mul; try omega.
            assert (H12: t / T * T + T <= te - T).
            { apply mod_0_bound_increase; omega. }
            rewrite mod_0_divide_mul in H12; try omega.
            destruct (zlt sched id); try omega.
          + assert (H12: (t + T) / T * T <= te / T * T).
            { apply div_mul_le; omega. }
            rewrite mod_0_divide_mul in H12; omega.
        - rewrite Z_mod_same_full; omega.
      }
    }
  Qed.

  Lemma workload_whole_remains_same:
    forall t te sched T C id,
      1 < T ->
      0 <= t <= te - 1 ->
      t mod T <> 0 ->
      workload_whole id sched T C (t + 1) te = workload_whole id sched T C t te.
  Proof.
    intros * HT Ht Htmod.
    unfold workload_whole.
    destruct (zlt (te - t) T) as [Hlt | Hge].
    {
      rewrite zlt_true; omega.
    }
    {
      destruct (zeq te (t + T)) as [Heq | Hneq].
      {
        rewrite Heq.
        rewrite zlt_true; try omega.
        assert (H1: (t + T) mod T <> 0).
        { apply mod_not_0_preserve; omega. }
        rewrite zeq_false; auto.
        rewrite zeq_false; auto.
        rewrite div_T_inc_1; try omega.
        rewrite Z.mul_add_distr_r.
        replace (t / T * T + 1 * T - (t / T * T + T)) with 0 by omega.
        simpl.
        auto.
      }
      {
        rewrite zlt_false; try omega.
        symmetry.
        remember (if zeq (te mod T) 0 then if zlt sched id then te / T * T else te - T else te / T * T) as v.
        rewrite zeq_false; auto.
        assert (HH: (if zeq ((t + 1) mod T) 0 then t + 1 else (t + 1) / T * T + T) = t / T * T + T).
        {
          destruct (zeq ((t + 1) mod T) 0) as [Ht1mod | Ht1mod].
          - rewrite mod_0_minus_1; try omega.
          - rewrite <- z_divide_same; try omega.
        }
        rewrite HH.
        auto.
      }
    }
  Qed.
        
  Lemma workload_remain_1_less_budget:
    forall t te sched T C id T0 C0 conf vt,
      monotonic_continuous vt ->
      1 < T ->
      1 < T0 ->
      0 < C ->
      0 <= t <= te - 1 ->
      te mod T0 = 0 ->
      ZMap.get sched conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry (t / T0 * T0 + T0) id) = true ->
      t mod T = 0 ->
      workload_remain id sched T C (t + 1) te vt = C - (vt (t + 1) - vt t).
  Proof.
    intros * Hvt HT HT0 HC Ht Htemod Hconfsched Hconfid Hpri Htmod.
    unfold workload_remain.
    assert (H1: (t + 1) mod T = 1).
    {
      rewrite Zplus_mod.
      rewrite Htmod.
      rewrite Zmod_1_l; try omega.
      simpl.
      rewrite Zmod_1_l; try omega.
    }
    rewrite zeq_false; try omega.
    rewrite <- z_divide_same; try omega.
    rewrite mod_0_divide_mul in *; try omega.
    assert (H2: t / T0 * T0 + T0 <= te).
    {
      assert (t / T0 * T0 + T0 <= te - 0 * T0).
      { apply mod_0_bound_increase_k; omega. }
      omega.
    }
    assert (H3: t + T <= t / T0 * T0 + T0).
    { eapply priEntry_lt_deadline_le; eauto. }
    assert (Hvalue: (if priEntry_lt (PriEntry (t + T) sched) (PriEntry te id) then Z.max 0 (C - vt (t + 1) + vt t) else 0) = Z.max 0 (C - vt (t + 1) + vt t)).
    {
      destruct (zeq te (t + T)) as [Hsync | Hregular].
      - rewrite Hsync in *.
        assert (H4: t / T0 * T0 + T0 = t + T). omega.
        rewrite H4 in *.
        rewrite Hpri; auto.
      - unfold priEntry_lt.
        rewrite zlt_true; try omega.
    }
    rewrite Hvalue.
    rewrite Z.max_r; try omega.
    assert (H5: t >= 0). omega.
    specialize (Hvt _ H5).
    destruct Hvt; omega.
  Qed.

  Lemma workload_remain_dec_at_least_1:
    forall t te sched T C id T0 C0 conf vt,
      vt t <= vt (t + 1) <= vt t + 1 ->
      vt t - vt (t / T * T) < C ->
      1 < T ->
      1 < T0 ->
      0 < C ->
      0 <= t <= te - 1 ->
      te mod T0 = 0 ->
      ZMap.get sched conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry (t / T0 * T0 + T0) id) = true ->
      t mod T <> 0 ->
      vt (t + 1) - vt t + workload_remain id sched T C (t + 1) te vt <= workload_remain id sched T C t te vt.
  Proof.
    intros * Hvt Hvt2 HT HT0 HC Ht Htemod Hconfsched Hconfid Hpri Htmod.
    unfold workload_remain.
    assert (H1: (if zeq (t mod T) 0
    then 0
    else
     if priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry te id)
     then Z.max 0 (C - vt t + vt (t / T * T))
     else 0) = (if priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry te id)
     then Z.max 0 (C - vt t + vt (t / T * T))
     else 0)).
    { rewrite zeq_false; auto. }
    rewrite H1.
    assert (H2: t / T0 * T0 + T0 <= te).
    {
      assert (t / T0 * T0 + T0 <= te - 0 * T0).
      { apply mod_0_bound_increase_k; omega. }
      omega.
    }
    assert (H3: t / T * T + T <= t / T0 * T0 + T0).
    { eapply priEntry_lt_deadline_le; eauto. }
    assert (Hpri2: priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry te id) = true).
    {
      destruct (zeq te (t / T * T + T)) as [Heq | Hneq].
      {
        rewrite Heq in *.
        replace (t / T0 * T0 + T0) with (t / T * T + T) in Hpri by omega.
        auto.
      }
      {
        unfold priEntry_lt.
        rewrite zlt_true; try omega.
        auto.
      }
    }
    rewrite Hpri2.    
    destruct (zeq ((t + 1) mod T) 0) as [Ht1mod | Ht1mod].
    {
      (* This is where <= is mandatory *)
      rewrite Z.max_r; try omega.      
    }
    {
      rewrite <- z_divide_same; try omega.
      rewrite Hpri2.
      rewrite Z.max_r; try omega.
      rewrite Z.max_r; try omega.
    }
  Qed.            
        
  Lemma sched_task_workload_subtotal_step_by_1:
    forall Nnat t te sched T C id T0 C0 conf,
      config_pool_valid (Z.of_nat Nnat) conf ->
      te mod T0 = 0 ->
      0 <= t <= te - 1 ->
      sched = highest_not_finished_map (Z.of_nat Nnat) t 
               (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) ->
      ZMap.get sched conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      0 <= id < Z.of_nat Nnat ->
      0 <= sched < Z.of_nat Nnat ->
      priEntry_lt (PriEntry (t / T * T + T) sched) (PriEntry (t / T0 * T0 + T0) id) = true ->
      1 + workload_subtotal id sched T C (t + 1) te (ZMap.get sched (TAC_pool_at_t te Nnat conf)) <=
      workload_subtotal id sched T C t te (ZMap.get sched (TAC_pool_at_t te Nnat conf)).
  Proof.
    intros * Hvalid Hmod Ht Hschedval Hconfsched Hconfid Hid Hsched Hpri.
    unfold workload_subtotal.
    assert (HT: 1 < T).
    {
      specialize (Hvalid _ _ _ Hsched Hconfsched).
      omega.
    }
    assert (HT0: 1 < T0).
    {
      specialize (Hvalid _ _ _ Hid Hconfid).
      omega.
    }
    assert (HC: 0 < C).
    {
      specialize (Hvalid _ _ _ Hsched Hconfsched).
      omega.
    }
    assert (Hvt: monotonic_continuous (ZMap.get sched (TAC_pool_at_t te Nnat conf))).
    { apply TAC_pool_all_monotonic_continuous; auto; omega. }
    destruct (zeq (t mod T) 0) as [Hnewperiod | Hregular].
    {
      assert (H1 : workload_remain id sched T C t te (ZMap.get sched (TAC_pool_at_t te Nnat conf)) = 0).
      {
        unfold workload_remain.
        rewrite Hnewperiod.
        rewrite zeq_true; auto.
      }
      rewrite H1.
      assert (H2: workload_whole id sched T C (t + 1) te + C = workload_whole id sched T C t te).
      { eapply workload_whole_1_less_period; eauto; try omega. }
      erewrite workload_remain_1_less_budget; eauto; try omega.
      assert (H3: ZMap.get sched (TAC_pool_at_t te Nnat conf) (t + 1) =
                  ZMap.get sched (TAC_pool_at_t te Nnat conf) t + 1).
      {
        rewrite vertical_map_canonical; try omega.
        symmetry.
        rewrite vertical_map_canonical; try omega.
        symmetry.
        apply vt_increase_for_the_scheduled_task; omega.
      }
      omega.
    }
    {
      rewrite workload_whole_remains_same; try omega.
      remember (ZMap.get sched (TAC_pool_at_t te Nnat conf)) as vt.
      assert (H1: vt (t + 1) = vt t + 1).
      {
        rewrite Heqvt.
        rewrite vertical_map_canonical; try omega.
        symmetry.        
        rewrite vertical_map_canonical; try omega.
        symmetry.
        apply vt_increase_for_the_scheduled_task; auto; omega.
      }
      assert (H2: vt t - vt (t / T * T) < C).
      {
        rewrite Heqvt.
        rewrite vertical_map_canonical; try omega.
        assert (H20: 0 <= t / T * T <= t).
        { apply Nfloor_le; omega. }
        assert (H21: ZMap.get sched (TAC_pool_at_t te Nnat conf) (t / T * T) =
                     ZMap.get sched (TAC_pool_at_t t Nnat conf) (t / T * T)).
        {          
          rewrite vertical_map_canonical; try omega.
          symmetry.
          rewrite vertical_map_canonical; try omega.          
        }
        rewrite H21.
        assert (H22: vt_map_not_finished t T C (ZMap.get sched (TAC_pool_at_t t Nnat conf)) = true).
        {
          eapply highest_not_finished_map_condition_met; eauto.
        }
        unfold vt_map_not_finished in H22.
        destruct (zlt
          (ZMap.get sched (TAC_pool_at_t t Nnat conf) t -
           ZMap.get sched (TAC_pool_at_t t Nnat conf) (t / T * T)) C); try omega.
        discriminate H22.
      }
      assert (H3: vt (t + 1) - vt t + workload_remain id sched T C (t + 1) te vt <= workload_remain id sched T C t te vt).
      {
        eapply workload_remain_dec_at_least_1; eauto; omega.
      }
      omega.
    }
  Qed.
  
  Lemma workload_complements_interference_step:
    forall t te id T0 C0 Nnat conf,
      config_pool_valid (Z.of_nat Nnat) conf ->
      0 <= t <= te - 1 ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      te mod T0 = 0 ->
      interference_at_t t id (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) +
      workload_total1 Nnat id (t + 1) te conf (TAC_pool_at_t te Nnat conf) <=
      workload_total1 Nnat id t te conf (TAC_pool_at_t te Nnat conf).
  Proof.
    intros * Hconf Ht Hid Hconfid Htemod.
    pose proof (interference_at_t_res t id (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf)) as HI.
    destruct HI as [HI0 | HI1].
    {
      rewrite HI0.
      eapply workload_total1_nondec_left with (TAC := TAC_pool_at_t te Nnat conf); eauto; try omega.
      intros.
      eapply TAC_pool_all_monotonic_continuous; eauto.
      omega.
    }
    {
      rewrite HI1.
      repeat rewrite workload_total_eq.
      remember (highest_not_finished_map (Z.of_nat Nnat) t (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf)) as sched.
      destruct (ZMap.get sched conf) as [T C] eqn:Hschedconf.
      assert (HHH: 0 <= sched < Z.of_nat Nnat /\ priEntry_lt (PriEntry (t / T * T + T) sched)
                                                           (PriEntry (t / T0 * T0 + T0) id) = true).
      {
        apply interference_1_highest_not_finished with (C0 := C0) (C := C) (TAC := TAC_pool_at_t t Nnat conf) (conf := conf) (queue := real_priqueue Nnat t conf); auto; try omega.
        - apply real_priqueue_sound.
        - apply real_priqueue_complete with (C := C0); auto.
      }
      destruct HHH as [Hsched Hpri].
      assert (Hidneq: id <> sched).
      {
        unfold not.
        intros Heq.
        rewrite Heq in *.
        rewrite Hconfid in Hschedconf.
        inv Hschedconf.
        unfold priEntry_lt in Hpri.
        rewrite zlt_false in Hpri; try omega.
        rewrite zeq_true in Hpri.
        rewrite zlt_false in Hpri; try omega.
        congruence.
      }
      assert (Hschedeq: Z.of_nat (Z.to_nat sched) = sched).
      { rewrite Z2Nat.id; omega. }
      assert (Hw: 1 + workload_subtotal id sched T C (t+1) te (ZMap.get sched (TAC_pool_at_t te Nnat conf)) <= workload_subtotal id sched T C t te (ZMap.get sched (TAC_pool_at_t te Nnat conf))).
      {
        eapply sched_task_workload_subtotal_step_by_1; eauto.
      }
      assert (Hle1: workload_total2 (S (Z.to_nat sched)) id (t + 1) te conf (TAC_pool_at_t te Nnat conf) + 1 <= workload_total2 (S (Z.to_nat sched)) id t te conf (TAC_pool_at_t te Nnat conf)).
      {
        simpl.
        rewrite Hschedeq.
        repeat rewrite zeq_false; auto.
        rewrite Hschedconf.
        assert (workload_total2 (Z.to_nat sched) id (t + 1) te conf (TAC_pool_at_t te Nnat conf) <= workload_total2 (Z.to_nat sched) id t te conf (TAC_pool_at_t te Nnat conf)).
        {
          repeat rewrite <- workload_total_eq.
          apply workload_total1_nondec_left; auto; try omega.
          - rewrite Hschedeq.
            apply config_pool_valid_shrink with (N2 := Z.of_nat Nnat); auto; omega.
          - intros.
            apply TAC_pool_all_monotonic_continuous; auto; omega. 
        }
        omega.
      }
      rewrite Z.add_comm.
      apply workload_total2_le_history_preserve with (nsched := S (Z.to_nat sched)); auto; try omega.
      - intros.
        apply TAC_pool_all_monotonic_continuous; auto; omega.
      - rewrite <- Z2Nat.inj_succ; try omega.
        rewrite Z2Nat.id; try omega.
    }
  Qed.
  
  Lemma workload_complements_interference:
    forall ts te id Tid Cid Nnat conf,
      config_pool_valid (Z.of_nat Nnat) conf ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid Tid Cid ->
      0 <= ts ->
      ts mod Tid = 0 ->
      te mod Tid = 0 ->
      forall nrange,
        0 <= Z.of_nat nrange <= te - ts ->
        interference_at_range ts nrange id Nnat conf +
        workload_total1 Nnat id (ts + Z.of_nat nrange) te conf (TAC_pool_at_t te Nnat conf) <=
        workload_total1 Nnat id ts te conf (TAC_pool_at_t te Nnat conf).
  Proof.
    intros * Hconfvalid Hidbound Hconfid Hnneg Hmod1 Hmod2.
    intros nrange.
    induction nrange.
    {      
      simpl.
      rewrite Z.add_0_r.
      omega.
    }
    {
      intros HSnrange.
      assert (Hnrange: 0 <= Z.of_nat nrange <= te - ts).
      { rewrite Nat2Z.inj_succ in HSnrange. omega. }
      specialize (IHnrange Hnrange).
      assert (H1 : ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1).
      { rewrite Nat2Z.inj_succ. omega. }
      rewrite H1.
      simpl.
      transitivity (interference_at_range ts nrange id Nnat conf +
                    workload_total1 Nnat id (ts + Z.of_nat nrange) te conf (TAC_pool_at_t te Nnat conf));
        try omega.

      remember (ts + Z.of_nat nrange) as t.
      assert (interference_at_t t id (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) +
   workload_total1 Nnat id (t + 1) te conf (TAC_pool_at_t te Nnat conf) <=
              workload_total1 Nnat id t te conf (TAC_pool_at_t te Nnat conf)).
      {
        apply workload_complements_interference_step with (T0 := Tid) (C0 := Cid); auto.
        rewrite Nat2Z.inj_succ in *; omega.
      }
      omega.
    }
  Qed.

  Lemma interference_bounded_by_workload:
    forall ts te id Tid Cid Nnat conf,
      config_pool_valid (Z.of_nat Nnat) conf ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid Tid Cid ->
      ts mod Tid = 0 ->
      te mod Tid = 0 ->
      0 <= ts ->
      ts <= te ->
      interference_at_range ts (Z.to_nat (te - ts)) id Nnat conf <=
      workload_total1 Nnat id ts te conf (TAC_pool_at_t te Nnat conf).
  Proof.
    intros * Hvalid Hidbound Hconf H1 H2 H3 H4.
    pose proof (workload_complements_interference _ _ _ _ _ Nnat conf
               Hvalid Hidbound Hconf H3 H1 H2) as H.
    assert (Ht: 0 <= Z.of_nat (Z.to_nat (te - ts)) <= te - ts).
    { rewrite Z2Nat.id; omega. }
    specialize (H _ Ht).
    replace (ts + Z.of_nat (Z.to_nat (te - ts))) with te in H.
    assert (Hw: workload_total1 Nnat id te te conf (TAC_pool_at_t te Nnat conf) = 0).
    {
      rewrite workload_total_eq.
      rewrite workload_total2_empty_window_0; auto.
    }
    omega.
    rewrite Z2Nat.id; omega.
  Qed.
    
End INTERFERENCE.


Section VTBOUND.

  Lemma vtbound_by_workload:
    forall i T C (N: nat) (conf: PrioConfigPool),
      0 <= i < Z.of_nat N ->
      config_pool_valid (Z.of_nat N) conf ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      (forall k, 0 <= k ->
            T - workload_total1 N i (k * T) (k * T + T) conf
                                (TAC_pool_at_t (k * T + T) N conf) <=
            (ZMap.get i (TAC_pool_at_t (k * T + T) N conf)) (k * T + T) -
            (ZMap.get i (TAC_pool_at_t (k * T + T) N conf)) (k * T)).
  Proof.
    intros * Hi Hvalid Hconfi.
    intros k Hk.

    assert (HT: 1 < T).
    {
      specialize (Hvalid _ _ _ Hi Hconfi).
      omega.
    }
    remember (Z.to_nat T) as nT.
    assert (HTT: T = Z.of_nat nT).
    {
      rewrite HeqnT.
      rewrite Z2Nat.id; omega.
    }
      
    assert (Hvt: ZMap.get i (TAC_pool_at_t (k * T + T) N conf) (k * T + T) -
                 ZMap.get i (TAC_pool_at_t (k * T + T) N conf) (k * T) +
           interference_at_range (k*T) nT i N conf = T).
    {
      rewrite HTT.
      apply interference_creates_hole_in_vt; auto.
      nia.
    }

    replace (ZMap.get i (TAC_pool_at_t (k * T + T) N conf) (k * T + T) -
             ZMap.get i (TAC_pool_at_t (k * T + T) N conf) (k * T)) with
        (T - interference_at_range (k * T) nT i N conf) by omega.

    assert (H: forall a b, b <= a -> T - a <= T - b). intros. omega.
    apply H.

    assert (H1: nT = Z.to_nat ((k * T + T) - k*T)).
    {
      replace (k * T + T - k * T) with T by omega.
      auto.
    }
    rewrite H1.

    assert (H2: 0 <= k * T). nia.
    assert (H3: (k * T) mod T = 0). apply Z_mod_mult.
    assert (H4: (k * T + T) mod T = 0). apply mod_0_preserve; try omega.
    
    apply interference_bounded_by_workload with (Tid:=T) (Cid:=C); auto; omega.
  Qed.

  Definition workload_schedulable N conf :=
    forall i T C,
      0 <= i < N ->      
      ZMap.get i conf = mkPrioConfigValid T C ->      
      (forall p, 0 <= p ->
            workload_total1 (Z.to_nat N) i (p * T) (p * T + T) conf
              (TAC_pool_at_t (p * T + T) (Z.to_nat N) conf) <= T - C).

  Lemma workload_schedulable_implies_system_schedulable:
    forall N conf,
      config_pool_valid N conf ->
      workload_schedulable N conf ->
      system_schedulable N conf.
  Proof.
    intros * Hconf Hworkload.
    unfold system_schedulable.
    intros * Ht.
    unfold schedulable_upto.
    intros * Hp.
    destruct (ZMap.get p conf) as [T C] eqn:Hconfp.
    unfold schedulable_per_level_upto.
    intros * Hk1 Hk2.

    assert (HT: 1 < T).
    {
      specialize (Hconf _ _ _ Hp Hconfp).
      omega.
    }
    
    rewrite vertical_map_canonical; try nia.
    assert (H1: ZMap.get p (TAC_pool_at_t t (Z.to_nat N) conf) (k * T) =
           ZMap.get p (TAC_pool_at_t (k * T + T) (Z.to_nat N) conf) (k * T)).
    {
      rewrite vertical_map_canonical; try nia.
      symmetry.
      rewrite vertical_map_canonical; try nia.      
    }
    rewrite H1.

    assert (HNN: Z.of_nat (Z.to_nat N) = N).
    {
      rewrite Z2Nat.id; omega.
    }
    
    transitivity (T - workload_total1 (Z.to_nat N) p (k * T) (k * T + T) conf
                                      (TAC_pool_at_t (k * T + T) (Z.to_nat N) conf)).
    - unfold workload_schedulable in Hworkload.
      specialize (Hworkload _ _ _ Hp Hconfp _ Hk1).
      omega.
    - apply vtbound_by_workload with (C:=C); auto.
      + rewrite HNN. auto.
      + rewrite HNN. auto.
  Qed.
      
End VTBOUND.


