Require Import ZArith.
Require Import Omega.
Require Import Psatz.
Require Import Coqlib.
Open Scope Z.

Lemma Nadd_div_exact : forall a b c,
    0 <= a ->
    0 <= b ->
    0 < c ->
    a mod c = 0 ->
    (a + b) / c = a / c + b / c.
Proof.
  intros * Ha Hb Hc Hmod.
  assert (H1: a = c * (a / c)).
  {
    apply Z_div_exact_full_2; omega.
  }
  assert (H2: a + b = (a / c) * c + b). lia.
  rewrite H2.
  rewrite Z_div_plus_full_l; omega.
Qed.

Definition monotonic_continuous (f: Z -> Z) : Prop :=
  forall x, x >= 0 -> f (x + 1) = f x \/ f (x + 1) = f x + 1.

Lemma monotonic_continuous_monotonic:
  forall f x y,
    monotonic_continuous f ->
    0 <= x ->
    x <= y ->
    f x <= f y.
Proof.
  intros * Hf Hx Hy.
  remember (Z.to_nat (y - x)) as range.
  assert (y = x + Z.of_nat range).
  {
    rewrite Heqrange.
    rewrite Z2Nat.id; omega.
  }
  rewrite H in *.
  clear Heqrange H.
  induction range.
  {
    rewrite Z.add_0_r.
    omega.
  }
  {
    assert (H: x <= x + Z.of_nat range).
    { pose proof (Zle_0_nat range). omega. }
    transitivity (f (x + Z.of_nat range)).
    - apply IHrange; auto.
    - unfold monotonic_continuous in Hf.
      assert (H1: x + Z.of_nat (S range) = (x + Z.of_nat range) + 1).
      { rewrite Nat2Z.inj_succ; omega. }
      rewrite H1.
      assert (H2: x + Z.of_nat range >= 0). omega.
      destruct (Hf _ H2); omega.
  }
Qed.    

Lemma mod_increase:
  forall x y,
    x >= 0 ->          
    y > 1 ->
    (x + 1) mod y <> 0 ->
    (x + 1) mod y = x mod y + 1.
Proof.
  intros.        
  rewrite Zplus_mod in *.        
  assert (1 mod y = 1). 
  {
    rewrite Z.mod_small.
    reflexivity.
    omega.
  }
  rewrite H2 in *.
  assert (0 <= x mod y + 1 < y).
  {
    split.
    + assert (0 <= x mod y).
      eapply Z.mod_pos_bound; omega.
      omega.
    + assert (0 <= x mod y < y).
      eapply Z.mod_pos_bound; omega.
      destruct (zeq (x mod y + 1) y) eqn:Hmod5.
    - rewrite e in H1.
      replace (y mod y) with 0 in H1.
      omega.
      rewrite Z_mod_same_full.
      reflexivity.
    - omega.
  }
  erewrite Zmod_small.
  reflexivity.
  omega.
Qed.

Lemma mod_increase_bound:
  forall x y,
    0 <= x ->          
    1 < y ->
    (x + 1) mod y <> 0 ->
    0 <= x mod y < y - 1.
Proof.
  intros.
  assert (H10: (x + 1) mod y = x mod y + 1).
  {
    apply mod_increase; omega.
  }
  assert (H11: 0 <= (x + 1) mod y < y).
  {
    apply Z_mod_lt; omega.
  }
  assert (H12: 0 <= x mod y < y).
  {
    apply Z_mod_lt; omega.
  }
  omega.
Qed.
        
Lemma z_divide_same:
  forall t y,
    t >= 0 ->
    y > 1 ->
    (t + 1) mod y <> 0 ->
    t / y = (t + 1) / y.
Proof.
  intros.
  assert (t + 1 = t + 1). reflexivity.
  assert (y <> 0). omega.
  rewrite (Z.div_mod t y) in H2 at 1; [ | auto ].
  rewrite (Z.div_mod (t + 1) y) in H2; [ | auto ].        
  assert ((t + 1) mod y = t mod y + 1).
  {
    eapply mod_increase; try omega.
  }
  rewrite H4 in *.
  clear H4.
  assert (y * (t / y) = y * ((t + 1) / y)). omega.
  eapply Z.mul_cancel_l; eauto.
Qed.

Lemma Zdiv_incr_exact : forall a b,
    0 <= a ->
    0 < b ->
    ((a+1) mod b = 0 <-> (a+1) / b = a / b + 1).
Proof.
  intros * Ha Hb.
  split; intro H1.
  {
    assert (H2: a + 1 = b * ((a + 1) / b)).
    {
      apply Z_div_exact_full_2; omega.
    }
    assert (H3: (a + 1) / b >= 0).
    {
      apply Z_div_ge0; omega.
    }
    remember ((a + 1) / b) as k.
    replace a with ((k - 1) * b + (b - 1)) by lia.
    rewrite Z_div_plus_full_l; try omega.
    assert (Hbb: (b - 1) / b = 0).
    {
      apply Z.div_small.
      omega.
    }
    rewrite Hbb.
    omega.
  }
  {
    destruct (zeq ((a + 1) mod b) 0); auto.
    destruct (zeq b 1).
    {
      rewrite e.
      apply Zmod_1_r.
    }
    {
      rewrite <- z_divide_same in H1; try omega.
    }
  }
Qed.


Lemma z_gt_mod:
  forall x y,
    x >= 0 ->
    y > 0 ->
    x mod y <> 0 ->
    x - x / y * y >= 1.
Proof.
  intros.
  destruct (zeq y 1) eqn:Hy.
  {
    rewrite e in H1.
    rewrite Zmod_1_r in H1.
    omega.
  }
  {
    rewrite (Z.div_mod x y) at 1; [ | omega ].
    rewrite Z.mul_comm.
    assert (0 < y). omega.
    pose proof (Z.mod_pos_bound x y H2).
    omega.
  }
Qed.

Lemma mul_1: 
  forall x,
    x * 1 = x.
Proof.
  intros.
  omega.
Qed.

    Lemma mod_lower_bound:
      forall x y b,
        0 <= x ->
        x <= y < x + b ->
        0 < b ->
        x mod b = 0 ->
        y / b * b = x.
    Proof.
      intros.
      replace y with (x + (y - x)) by omega.
      rewrite Nadd_div_exact; try omega.
      rewrite Z.mul_add_distr_r.
      assert (y - x < b).
      omega.
      replace ((y - x) / b) with 0.
      assert (H4: b <> 0). omega.
      rewrite (Z.div_mod x b H4) at 2.
      rewrite H2.
      rewrite Z.mul_comm.
      omega.
      symmetry.
      apply Zdiv.Zdiv_small.
      omega.
    Qed.

Lemma Nfloor_le : forall a b,
    0 <= a ->
    0 < b ->
    0 <= a / b * b <= a.
Proof.
  intros * Ha Hb.
  split.
  {
    apply Zmult_gt_0_le_0_compat; try omega.
    apply Z_div_pos; omega.
  }
  {
    destruct (zeq (a mod b) 0).
    {
      rewrite mod_lower_bound with (x := a); try omega.
    }
    {
      assert (a - a / b * b >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
  }
Qed.    

        Lemma mod_equal:
          forall x y,
            y > 1 ->
            0 <= x ->
            (x + 1) mod y <> 0 ->
            (x + 1) / y * y = x / y * y.
        Proof.
          intros.
          destruct (zeq (x mod y) 0).
          + assert (0 <= 1). omega.
            assert (0 < y). omega.
            rewrite (Nadd_div_exact _ _ _ H0 H2 H3 e).
            rewrite Z.mul_add_distr_r.
            assert (1 / y = 0).
            {
              apply Zdiv_small.
              omega.
            }
            rewrite H4.
            replace (0 * y) with 0; omega.
          + assert (x + 1 = x + 1). reflexivity.            
            assert (y <> 0). omega.
            rewrite (Z.div_mod (x + 1) y H3) in H2 at 1.
            rewrite Z.mul_comm in H2.
            rewrite (Z.div_mod x y H3) in H2 at 3.
            assert (y * (x / y) = x / y * y). rewrite Z.mul_comm. reflexivity.
            rewrite H4 in H2.
            assert ((x + 1) mod y = x mod y + 1).
            {
              rewrite Zplus_mod.
              assert (1 < y). omega.
              rewrite (Zmod_1_l y H5).
              rewrite Z.mod_small.
              reflexivity.
              split.
              assert (0 <= x mod y). apply Z.mod_pos_bound. omega.
              omega.
              assert (x mod y < y). apply Z.mod_pos_bound. omega.
              destruct (zeq (x mod y + 1) y).
              - rewrite Zplus_mod in H1.
                rewrite (Zmod_1_l y H5) in H1.
                rewrite e in H1.
                rewrite Z_mod_same_full in H1.
                omega.
              - omega.
            }
            omega.
        Qed.

Lemma divide_plus_equal:
  forall x y z,
    0 <= x ->
    0 <= z < y ->
    x mod y = 0 ->
    (x + z) / y = x / y.
Proof.
  intros * Hx Hz Hmod.
  rewrite Nadd_div_exact; try solve [ omega | auto].
  replace (z / y) with 0.
  omega.
  rewrite Zdiv_small; omega.
Qed.

Lemma mod_0_divide_mul:
  forall x y,
    0 <= x ->
    0 < y ->
    x mod y = 0 ->
    x / y * y = x.
Proof.
  intros.
  assert (H2: y <> 0). omega.
  pose proof (Z.div_mod x y H2) as H3.
  rewrite Z.mul_comm in H3.
  omega.
Qed.

Lemma divide_mul_le:
  forall x y,
    0 <= x ->
    0 < y ->
    x / y * y <= x.
Proof.
  intros.
  assert (H1: y <> 0). omega.
  pose proof (Zmod_eq_full x y H1).
  assert (Hpos: y > 0). omega.
  pose proof (Z_mod_lt x y Hpos).
  omega.
Qed.

  Lemma mod_0_preserve:
    forall a b,
      0 < b ->
      a mod b = 0 ->
      (a + b) mod b = 0.
  Proof.
    intros * Hb Hmod.
    rewrite Zplus_mod.
    rewrite Hmod.
    rewrite Z_mod_same_full.
    simpl.
    apply Zmod_0_l.
  Qed.

  Lemma mod_not_0_divide:
    forall a b,
      0 <= a ->
      0 < b ->
      a mod b <> 0 ->
      (a + b) / b = a / b + 1.
  Proof.
    intros.
    rewrite Z.add_comm.
    erewrite Nadd_div_exact; try omega.
    rewrite Z_div_same_full; try omega.
    apply Z_mod_same_full.
  Qed.

  Lemma mod_not_0_divide_mul:
    forall a b,
      0 <= a ->
      0 < b ->
      a mod b <> 0 ->
      (a + b) / b * b = a / b * b + b.
  Proof.
    intros.
    rewrite mod_not_0_divide; try omega.
    rewrite Z.mul_add_distr_r.
    omega.
  Qed.
      
  Lemma mod_not_0_preserve:
    forall a b,
      0 <= a ->
      0 < b ->
      a mod b <> 0 ->
      (a + b) mod b <> 0.
  Proof.
    intros * Ha Hb Hmod.
    assert (H1: b <> 0). omega.
    pose proof (Z.div_mod (a + b) _ H1) as H2.
    rewrite Z.mul_comm in H2.
    rewrite mod_not_0_divide_mul in H2; try omega.
    pose proof (Zmod_eq_full a b H1).
    omega.
  Qed.

  Lemma mod_0_not_0:
    forall a b c,
      0 < b < c ->
      a mod c = 0 ->
      (a + b) mod c <> 0.
  Proof.
    intros.
    rewrite Zplus_mod.
    rewrite H0.
    rewrite Z.add_0_l.
    rewrite Zmod_mod.
    rewrite Z.mod_small; try omega.
  Qed.

  Lemma mod_0_minus_1:
    forall a b,
      0 <= a ->      
      1 < b ->
      (a + 1) mod b = 0 ->
      a / b * b = a + 1 - b.
  Proof.
    intros * Ha Hb Hmod.
    assert (H1: (a + 1) / b = a / b + 1).
    {
      apply Zdiv_incr_exact; try omega.
    }
    assert (H2: (a + 1) / b * b = (a / b + 1) * b). rewrite H1. auto.
    rewrite mod_0_divide_mul in H2; try omega.
    rewrite Z.mul_add_distr_r in H2.
    omega.
  Qed.

  Lemma mod_0_minus_1_2:
    forall a b,
      0 <= a ->      
      1 < b ->
      (a + 1) mod b = 0 ->
      a mod b = b - 1.
  Proof.
    intros * Ha Hb Hmod.
    assert (a / b * b = a + 1 - b).
    {
      apply mod_0_minus_1; omega.
    }
    replace a with (a / b * b + (b - 1)) by omega.
    rewrite Zplus_mod.
    rewrite Z_mod_mult; try omega.
    rewrite Z.add_0_l.
    rewrite Zmod_mod.
    rewrite Zmod_small; try omega.
  Qed.
  
  Lemma div_less_1:
    forall a b c,
      0 <= a ->
      a < b ->
      0 < c ->
      b mod c = 0 ->
      a / c < b / c.
  Proof.
    intros * Ha Hb Hc Hmod.
    assert (H1: a / c <= b / c).
    { apply Z.div_le_mono; try omega. }
    destruct (zeq (a / c) (b / c)) as [Heq | Hneq]; try omega.
    exfalso. clear H1.
    assert (Heqa : let (q1, r1) := Z.div_eucl a c in a = c * q1 + r1).
    { apply Z.div_eucl_eq. lia. }
    assert (Heqb : let (q2, r2) := Z.div_eucl b c in b = c * q2 + r2).
    { apply Z.div_eucl_eq. lia. }
    assert (Hcc: c > 0). omega.
    assert (Hr := Z_mod_lt a c Hcc).    
    unfold Z.modulo, Z.div in *.
    destruct (Z.div_eucl a c) as [q1 c1].
    destruct (Z.div_eucl b c) as [q2 c2].    
    nia.
  Qed.

  Lemma mod_0_bound_increase_k:
    forall a b c k,
      0 <= k ->
      0 <= a ->
      1 < c ->
      a < b - k * c ->
      b mod c = 0 ->
      a / c * c + c <= b - k * c.
  Proof.
    intros * Hk Ha Hc Hb Hmod.
    assert (H1: (b - k * c) mod c = 0).
    {
      rewrite Zminus_mod.
      rewrite Hmod.
      rewrite Z_mod_mult.
      simpl.
      rewrite Zmod_0_l.
      auto.
    }      
    assert (H2: a / c < (b - k * c) / c).
    {
      apply div_less_1; try omega.
    }
    assert (H3: a / c <= (b - k * c) / c - 1). omega.
    assert (H4: a/ c * c <= ((b - k * c) / c - 1) * c). nia.
    assert (H5: a/ c * c <= (b - k * c) / c * c - c). nia.
    replace ((b - k * c) / c * c) with (b - k * c) in H5.
    - omega.
    - rewrite mod_0_divide_mul; omega.
  Qed.

  Lemma mod_0_bound_increase:
    forall a b c,
      0 <= a ->
      1 < c ->
      a < b - c ->
      b mod c = 0 ->
      a / c * c + c <= b - c.
  Proof.
    intros * Ha Hc Hb Hmod.
    assert (H1: 1 <= 1). omega.
    assert (H2: a < b - 1 * c). omega.
    assert (a / c * c + c <= b - 1 * c).
    {
      apply mod_0_bound_increase_k; omega.
    }
    omega.
  Qed.
  
  Lemma div_T_inc_1:
    forall a b,
      0 <= a ->
      0 < b ->
      (a + b) / b = a / b + 1.
  Proof.
    intros * Ha Hb.
    replace (a + b) with (a + 1 * b) by omega.
    rewrite Z_div_plus_full; try omega.
  Qed.

  Lemma div_mul_le:
    forall a b c d,
      a <= b ->
      0 < c ->
      0 < d ->
      a / c * d <= b / c * d.
  Proof.
    intros.
    apply Z.mul_le_mono_pos_r; try omega.
    apply Z_div_le; try omega.
  Qed.

  Lemma x_le_ceil:
    forall a b,
      0 <= a ->
      0 < b ->
      a <= (a + b) / b * b.
  Proof.
    intros.
    assert (Hr: 0 <= a mod b < b). apply Z_mod_lt; omega.
    assert (Heq : let (q, r) := Z.div_eucl a b in a = b * q + r).
    { apply Z.div_eucl_eq. lia. }
    unfold Z.modulo in Hr.
    rewrite div_T_inc_1; try omega.
    unfold Z.div.
    destruct (Z.div_eucl a b) as [q r].    
    lia.
  Qed.

  Lemma x_lt_ceil:
    forall a b,
      0 <= a ->
      0 < b ->
      a < (a + b) / b * b.
  Proof.
    intros.
    assert (Hr: 0 <= a mod b < b). apply Z_mod_lt; omega.
    assert (Heq : let (q, r) := Z.div_eucl a b in a = b * q + r).
    { apply Z.div_eucl_eq. lia. }
    unfold Z.modulo in Hr.
    rewrite div_T_inc_1; try omega.
    unfold Z.div.
    destruct (Z.div_eucl a b) as [q r].    
    lia.
  Qed.

  Lemma x_lt_floor_plus:
    forall a b,
      0 <= a ->
      0 < b ->
      a < a / b * b + b.
  Proof.
    intros * Ha Hb.
    pose proof (x_lt_ceil _ _ Ha Hb).
    rewrite div_T_inc_1 in H; try omega.
    nia.
  Qed. 
  
  Lemma mod_not_0_adjust_to_smaller_period:
    forall ts T Tf,
      0 <= ts ->
      (* ts mod T = 0 -> *)
      1 < Tf <= T ->
      ts <= (ts + T) / Tf * Tf.
  Proof.
    intros * Hts HT.
    replace (ts + T) with ((ts + T - Tf) + Tf) by omega.
    transitivity (ts + T - Tf); try omega.
    apply x_le_ceil; omega.
  Qed.

  Lemma a_between_floor_and_self_mod_not_0:
    forall a b c,
      0 <= c ->
      0 < b ->
      c / b * b < a <= c ->
      a mod b <> 0.
  Proof.
    intros.
    assert (Hr: 0 <= c mod b < b). apply Z_mod_lt; omega.
    assert (Heq : let (q, r) := Z.div_eucl c b in c = b * q + r).
    { apply Z.div_eucl_eq. lia. }
    
    unfold Z.modulo in Hr.
    unfold Z.div in H1.
    destruct (Z.div_eucl c b) as [q r].
    remember (a - b * q) as d.
    assert (0 < d <= r). nia.
    replace a with (b * q + d) by omega.

    rewrite Zplus_mod.
    rewrite Z.mul_comm.
    rewrite Z_mod_mult.
    assert (d mod b = d).
    rewrite Z.mod_small; try omega.
    rewrite H3.
    simpl.
    rewrite H3.
    omega.
  Qed.

  Lemma div_pos:
    forall a b,
      0 < a <= b ->
      0 < b / a.
  Proof.
    intros.
    assert (Hr: 0 <= b mod a < a). apply Z_mod_lt; omega.
    assert (Heq : let (q, r) := Z.div_eucl b a in b = a * q + r).
    { apply Z.div_eucl_eq. lia. }
    unfold Z.div.
    unfold Z.modulo in Hr.
    destruct (Z.div_eucl b a) as [q r].
    nia.
  Qed.

  Lemma div_mul_nonneg:
    forall a b c,
      0 <= a ->
      0 < b ->
      0 < c ->
      0 <= a / b * c.
  Proof.
    intros.
    apply Zmult_gt_0_le_0_compat; try omega.
    apply Z_div_pos; omega.
  Qed.

  Lemma div_mul_round_up:
    forall a b c,
      0 <= a ->
      0 < b ->
      c mod b = 0 ->
      c - b < a < c ->
      a / b * b + b = c.
  Proof.
    intros.
    assert (Hr: 0 <= a mod b < b). apply Z_mod_lt; omega.
    assert (Heq : let (q, r) := Z.div_eucl a b in a = b * q + r).
    { apply Z.div_eucl_eq. lia. }
    assert (Hr': 0 <= c mod b < b). apply Z_mod_lt; omega.
    assert (Heq' : let (q', r') := Z.div_eucl c b in c = b * q' + r').
    { apply Z.div_eucl_eq. lia. }
    unfold Z.div.
    unfold Z.modulo in *.
    destruct (Z.div_eucl a b) as [q r].
    destruct (Z.div_eucl c b) as [q' r'].        
    destruct (zeq q' (q + 1)); nia.
  Qed.

  Lemma within_boundary_mod_not_0:
    forall a b c,
      0 <> b ->
      c mod b = 0 ->
      c - b < a < c ->
      a mod b <> 0.
  Proof.
    intros.
    assert (Heq : let (q, r) := Z.div_eucl a b in a = b * q + r).
    { apply Z.div_eucl_eq. lia. }
    assert (Heq' : let (q', r') := Z.div_eucl c b in c = b * q' + r').
    { apply Z.div_eucl_eq. lia. }
    unfold Z.div.
    unfold Z.modulo in *.
    destruct (Z.div_eucl a b) as [q r].
    destruct (Z.div_eucl c b) as [q' r'].    
    destruct (zeq q' (q + 1)); nia.
  Qed.
  
  Lemma divide_mul_plus_equal:
    forall x y z,
      0 <= x ->
      0 <= z < y ->
      x mod y = 0 ->
      (x + z) / y * y = x.
  Proof.
    intros.
    rewrite divide_plus_equal; try omega.
    rewrite mod_0_divide_mul; omega.
  Qed.

  Lemma divide_mul_plus_equal2:
    forall x y z,
      0 <= x ->
      x <= z < x + y ->
      x mod y = 0 ->
      z / y * y = x.
  Proof.
    intros.
    replace z with (x + (z - x)) by omega.
    apply divide_mul_plus_equal; omega.
  Qed.
  

  Lemma mod_0_divide_mul_assoc:
    forall a b c d,
      c <> 0 ->
      a mod c = 0 ->
      b mod c = 0 ->
      a / c * d + b / c * d = (a + b) / c * d.
  Proof.
    intros.
    rewrite <- Z.mul_add_distr_r.
    assert (a / c + b / c = (a + b) / c).
    {
      assert (Heq : let (q, r) := Z.div_eucl a c in a = c * q + r).
      { apply Z.div_eucl_eq. lia. }
      assert (Heq' : let (q', r') := Z.div_eucl b c in b = c * q' + r').
      { apply Z.div_eucl_eq. lia. }
      assert (Heq'' : let (q'', r'') := Z.div_eucl (a + b) c in a + b = c * q'' + r'').
      { apply Z.div_eucl_eq. lia. }

      assert ((a + b) mod c = 0).
      {
        rewrite Zplus_mod.
        rewrite H0, H1.
        simpl.
        rewrite Zmod_0_l.
        auto.
      }
      unfold Z.div.
      unfold Z.modulo in *.
      destruct (Z.div_eucl a c) as [q r].
      destruct (Z.div_eucl b c) as [q' r'].
      destruct (Z.div_eucl (a+b) c) as [q'' r''].       
      nia.
    }
    rewrite H2.
    auto.
  Qed.

  Lemma period_enlarge_le:
    forall a k T C,
      0 <= a ->
      0 < k ->
      0 < T ->
      0 <= C ->
      a / (k*T) * (k*C) <= a / T * C.
  Proof.
    intros * Ha Hk HT HC.
    assert (Hr: 0 <= a mod (k*T) < (k*T)). apply Z_mod_lt; try nia.    
    assert (Heq : let (q, r) := Z.div_eucl a (k*T) in a = (k*T) * q + r).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.div at 1.
    unfold Z.modulo in Hr.
    destruct (Z.div_eucl a (k*T)) as [q r].
    replace (q * (k*C)) with ((q*k) * C) by lia.
    apply Zmult_le_compat_r; try omega.    
    
    rewrite Heq.
    replace (k*T*q) with (q*k*T) by lia.
    rewrite Z_div_plus_full_l; try omega.
    assert (0 <= r / T).
    {
      apply Z_div_pos; omega.
    }
    omega.
  Qed.

  Lemma period_enlarge_le_le:
    forall a b k T C,
      0 <= a <= b ->
      0 < k ->
      0 < T ->
      0 < C ->
      a / (k*T) * (k*C) <= b / T * C.
  Proof.
    intros * Ha Hk HT HC.
    transitivity (a / T * C).
    - apply period_enlarge_le; omega.
    - apply div_mul_le; try omega.
  Qed.
    
  Lemma off_by_one_period_le:
    forall a b T,
      0 < T ->
      0 <= a ->
      a + T <= b ->
      a < b / T * T.
  Proof.
    intros * HT Ha Hb.
    
    assert (H1: a < (a + T) / T * T).
    {
      apply x_lt_ceil; omega.
    }

    assert (H2: (a + T) / T * T <= b / T * T).
    {
      apply div_mul_le; omega.
    }
    omega.
  Qed.

  Lemma period_enlarge_roundup_le:
    forall a k T C,
      0 <= a ->
      0 < k ->
      0 < T ->
      0 <= C ->
      a / T * C + C <= a / (k*T) * (k*C) + k*C .
  Proof.
    intros * Ha Hk HT HC.
    assert (Hr: 0 <= a mod (k*T) < (k*T)). apply Z_mod_lt; try nia.    
    assert (Heq : let (q, r) := Z.div_eucl a (k*T) in a = (k*T) * q + r).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.div at 2.
    unfold Z.modulo in Hr.
    destruct (Z.div_eucl a (k*T)) as [q r].
    replace (q * (k*C) + k * C) with ((q*k + k) * C) by lia.
    replace (a / T * C + C) with ((a / T + 1) * C) by lia.
    apply Zmult_le_compat_r; try omega.    
    
    rewrite Heq.
    replace (k*T*q) with (q*k*T) by lia.
    rewrite Z_div_plus_full_l; try omega.
    assert (r / T < k).
    {
      assert (Hr': 0 <= r mod T < T). apply Z_mod_lt; try nia.    
      assert (Heq' : let (q', r') := Z.div_eucl r T in r = T * q' + r').
      { apply Z.div_eucl_eq. nia. }
      unfold Z.div at 1.
      unfold Z.modulo in Hr'.
      destruct (Z.div_eucl r T) as [q' r'].
      nia.
    }
    
    nia.
  Qed.

  Lemma mod_not_0_large:
    forall k t T,
      0 < k ->
      0 < T ->
      0 <= t ->
      t mod T = 0 ->
      t mod (k*T) <> 0 ->
      t / (k*T) * (k*T) <= t - T.
  Proof.
    intros * Hk HT Ht Hmod1 Hmod2.
    destruct (zeq k 1).
    {
      rewrite e in *.
      replace (1 * T) with T in Hmod2 by omega.
      omega.
    }
    {
      remember (t / T) as p.
      assert (Hp: 0 <= p).
      {
        rewrite Heqp.
        apply Z.div_pos; omega.
      }
      assert (Hpp: t = p * T).
      {
        rewrite Heqp.
        rewrite mod_0_divide_mul; omega.
      }
      rewrite Hpp in *.
      clear Heqp Hpp Hmod1.

      rewrite Zmult_mod_distr_r in Hmod2.
      assert (Hmod3: p mod k <> 0). nia.
      rewrite Zdiv_mult_cancel_r; try omega.

      replace (p / k * (k * T)) with ((p / k * k) * T) by lia.
      replace (p * T - T) with ((p - 1) * T) by lia.
      apply Zmult_le_compat_r; try omega.

      assert (p - p / k * k >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
  Qed.
      
  Lemma roundup_le_rounddown_off_by_period:
    forall t1 t2 T,
      1 < T ->
      0 <= t1 ->
      t1 + T <= t2 ->
      (if zeq (t1 mod T) 0 then t1 else t1 / T * T + T) <=
      (if zeq (t2 mod T) 0 then (t2 - T) else t2 / T * T).
  Proof.
    intros * HT Ht1 Ht2.
    assert (H1: t1 < t2 / T * T).
    {
      apply off_by_one_period_le; try nia.
    }
    
    destruct (zeq (t1 mod T) 0) as [Hmod1| Hmod1].
    {
      destruct (zeq (t2 mod T) 0); try omega.
    }
    {
      destruct (zeq (t2 mod T) 0) as [Hmod2 | Hmod2].
      {
        apply mod_0_bound_increase; try omega.
        - destruct (zeq t2 (t1 + T)); try omega.
          replace t2 with (t1 + 1 * T) in Hmod2 by omega.
          rewrite Z_mod_plus_full in Hmod2.
          omega.
      }
      {
        transitivity ((t1 + T) / T * T).
        - rewrite div_T_inc_1; try omega.
          nia.
        - apply div_mul_le; omega.
      }
    }
  Qed.

  Lemma period_le_hyper_period:
    forall T H,
      0 < T ->
      0 < H ->
      H mod T = 0 ->
      T <= H.
  Proof.
    intros * HT HH Hmod.
    assert (Heq : let (q, r) := Z.div_eucl H T in H = T * q + r).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.modulo in Hmod.
    destruct (Z.div_eucl H T) as [q r].
    rewrite Hmod in Heq.
    assert (0 < q). nia.
    nia.
  Qed.

  Lemma budget_valid_after_hyper:
    forall C T H,
      0 < C < T ->
      1 < T ->
      1 < H ->
      H mod T = 0 ->
      0 < H / T * C < H.
  Proof.
    intros * HC HT HH Hmod.
    assert (Heq : let (q, r) := Z.div_eucl H T in H = T * q + r).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.modulo in Hmod.
    unfold Z.div.
    destruct (Z.div_eucl H T) as [q r].
    rewrite Hmod in Heq.
    assert (0 < q). nia.
    nia.
  Qed.

  Lemma div_sub_mod_0:
    forall a b c,
      0 < c ->
      0 <= b <= a ->
      b mod c = 0 ->
      a / c - b / c = (a - b) / c.
  Proof.
    intros * Hc Ha Hb.
    remember (a - b) as d.
    assert (Hd: 0 <= d). omega.
    replace a with (b + d) by omega.
    destruct Ha as [Ha1 Ha2].
    clear Ha2. clear Heqd.

    assert (Heq : let (q, r) := Z.div_eucl b c in b = c * q + r).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.modulo in Hb.
    unfold Z.div at 2.
    destruct (Z.div_eucl b c) as [q r].    
    subst.
    rewrite Z.add_0_r.
    
    rewrite Z.mul_comm.
    rewrite Z_div_plus_full_l; try omega.
  Qed.

  Lemma divide_same_mod_k:
    forall t k T,
      0 <= t ->
      0 < k ->
      0 < T ->
      T mod k = 0 ->
      t / k * k / T = t / T.
  Proof.
    intros * Ht Hk HT Hmod.
    assert (HTk: k <= T).
    {
      apply period_le_hyper_period; auto.
    }

    assert (HTT: T / k * k = T).
    {
      rewrite mod_0_divide_mul; omega.
    }
    assert (H1: 0 <= T / k).
    {
      apply Z_div_pos; omega.
    }
    
    rewrite <- HTT.
    rewrite Zdiv_mult_cancel_r; try omega.
    rewrite Zdiv.Zdiv_Zdiv; try omega.
    rewrite Z.mul_comm.
    auto.
  Qed.
    
  Lemma zlt_k_times:
    forall x y k c,
      0 < k ->
      x mod k = 0 ->
      x / k - y / k < c ->
      x - y < k * c.
  Proof.
    intros * Hk Hmodx Hlt.
    assert (Heq : let (q1, r1) := Z.div_eucl x k in x = k * q1 + r1).
    { apply Z.div_eucl_eq. nia. }
    assert (Heq' : let (q2, r2) := Z.div_eucl y k in y = k * q2 + r2).
    { apply Z.div_eucl_eq. nia. }
    assert (Hr': 0 <= y mod k < k). apply Z_mod_lt; try nia.    
    unfold Z.modulo in *.
    unfold Z.div in *.

    destruct (Z.div_eucl x k) as [q1 r1].
    destruct (Z.div_eucl y k) as [q2 r2].            
    rewrite Hmodx in *.
    nia.
  Qed.

  Lemma zge_k_times:
    forall x y k c,
      0 < k ->
      y mod k = 0 ->
      x / k - y / k >= c ->
      x - y >= k * c.
  Proof.
    intros * Hk Hmodx Hlt.
    assert (Heq : let (q1, r1) := Z.div_eucl x k in x = k * q1 + r1).
    { apply Z.div_eucl_eq. nia. }
    assert (Heq' : let (q2, r2) := Z.div_eucl y k in y = k * q2 + r2).
    { apply Z.div_eucl_eq. nia. }
    assert (Hr': 0 <= x mod k < k). apply Z_mod_lt; try nia.    
    unfold Z.modulo in *.
    unfold Z.div in *.

    destruct (Z.div_eucl x k) as [q1 r1].
    destruct (Z.div_eucl y k) as [q2 r2].            
    rewrite Hmodx in *.
    nia.
  Qed.
  
  Lemma zlt_incr_less_k:
    forall a b k d,
      0 < k ->
      b mod k = 0 ->
      a mod k = 0 ->
      a < b ->
      d < k ->
      a + d < b.
  Proof.
    intros * Hk Hmodb Hmoda Ha Hd.
    assert (Heq : let (q1, r1) := Z.div_eucl a k in a = k * q1 + r1).
    { apply Z.div_eucl_eq. nia. }
    assert (Heq' : let (q2, r2) := Z.div_eucl b k in b = k * q2 + r2).
    { apply Z.div_eucl_eq. nia. }
    unfold Z.modulo in *.

    destruct (Z.div_eucl a k) as [q1 r1].
    destruct (Z.div_eucl b k) as [q2 r2].            
    rewrite Hmoda in *.
    rewrite Hmodb in *.
    assert (q1 < q2). nia.
    nia.
  Qed.

  Lemma monotonic_continuous_range:
    forall x1 x2 f,
      monotonic_continuous f ->
      0 <= x1 <= x2 ->
      f x2 - f x1 <= x2 - x1.
  Proof.
    intros * Hf Hx.
    remember (Z.to_nat (x2 - x1)) as nrange.
    assert (Hx2: x2 = x1 + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hx2.
    replace (x1 + Z.of_nat nrange - x1) with (Z.of_nat nrange) by omega.
    destruct Hx as [Hx1 _].
    clear Hx2.
    clear Heqnrange.

    induction nrange.
    {
      simpl.
      rewrite Z.add_0_r.
      omega.
    }
    {
      replace (Z.of_nat (S nrange)) with (Z.of_nat nrange + 1) by xomega.
      replace (x1 + (Z.of_nat nrange + 1)) with (x1 + Z.of_nat nrange + 1) by omega.
      assert (Ht: x1 + Z.of_nat nrange >= 0). omega.
      specialize (Hf _ Ht).
      destruct Hf; omega.
    }
  Qed.

Lemma divide_range:
  forall x y M,
    0 <= x < M ->
    1 < y ->
    0 <= x / y < M.
Proof.
  intros.
  assert (0 <= x / y <= M).
  {
    eapply Zdiv_interval_2; try omega.
  }
  assert (x / y <> M).
  {
    destruct (zeq x 0).
    - rewrite e. rewrite Zdiv_0_l. omega.
    - assert (x / y < x).
      eapply Z_div_lt; try omega.
      omega.
  }
  omega.
Qed.
  
